@@ -17,23 +17,18 @@
 
 package org.apache.celeborn.plugin.flink;
 
+import java.time.Duration;
+
+import org.apache.flink.configuration.Configuration;
+import org.apache.flink.configuration.NettyShuffleEnvironmentOptions;
 import org.apache.flink.runtime.io.network.api.writer.ResultPartitionWriter;
 import org.apache.flink.runtime.io.network.partition.consumer.IndexedInputGate;
 import org.apache.flink.runtime.shuffle.ShuffleEnvironment;
 import org.apache.flink.runtime.shuffle.ShuffleEnvironmentContext;
-import org.apache.flink.runtime.shuffle.ShuffleMaster;
-import org.apache.flink.runtime.shuffle.ShuffleMasterContext;
-import org.apache.flink.runtime.shuffle.ShuffleServiceFactory;
 
-public class RemoteShuffleServiceFactory extends AbstractRemoteShuffleServiceFactory
-    implements ShuffleServiceFactory<
-        RemoteShuffleDescriptor, ResultPartitionWriter, IndexedInputGate> {
+import org.apache.celeborn.plugin.flink.netty.NettyShuffleEnvironmentWrapper;
 
-  @Override
-  public ShuffleMaster<RemoteShuffleDescriptor> createShuffleMaster(
-      ShuffleMasterContext shuffleMasterContext) {
-    return new RemoteShuffleMaster(shuffleMasterContext, new SimpleResultPartitionAdapter());
-  }
+public class RemoteShuffleServiceFactory extends AbstractRemoteShuffleServiceFactory {
 
   @Override
   public ShuffleEnvironment<ResultPartitionWriter, IndexedInputGate> createShuffleEnvironment(
@@ -55,6 +50,14 @@ public class RemoteShuffleServiceFactory extends AbstractRemoteShuffleServiceFac
         parameters.resultPartitionManager,
         resultPartitionFactory,
         inputGateFactory,
-        parameters.celebornConf);
+        parameters.celebornConf,
+        new NettyShuffleEnvironmentWrapper(nettyShuffleServiceFactory, shuffleEnvironmentContext));
+  }
+
+  @Override
+  public Duration getRequestSegmentsTimeout(Configuration configuration) {
+    return Duration.ofMillis(
+        configuration.get(
+            NettyShuffleEnvironmentOptions.NETWORK_EXCLUSIVE_BUFFERS_REQUEST_TIMEOUT_MILLISECONDS));
   }
 }
