@@ -19,10 +19,12 @@ package org.apache.celeborn.plugin.flink;
 
 import java.net.InetAddress;
 import java.net.UnknownHostException;
+import java.time.Duration;
 import java.util.Collection;
 import java.util.HashMap;
 import java.util.Map;
 import java.util.Random;
+import java.util.Set;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.ExecutionException;
 
@@ -33,11 +35,22 @@ import org.apache.flink.configuration.ExecutionOptions;
 import org.apache.flink.configuration.MemorySize;
 import org.apache.flink.runtime.clusterframework.types.ResourceID;
 import org.apache.flink.runtime.executiongraph.ExecutionAttemptID;
+import org.apache.flink.runtime.executiongraph.ExecutionGraphID;
+import org.apache.flink.runtime.io.network.NettyShuffleServiceFactory;
 import org.apache.flink.runtime.io.network.partition.ResultPartitionID;
 import org.apache.flink.runtime.io.network.partition.ResultPartitionType;
 import org.apache.flink.runtime.jobgraph.IntermediateDataSetID;
 import org.apache.flink.runtime.jobgraph.IntermediateResultPartitionID;
-import org.apache.flink.runtime.shuffle.*;
+import org.apache.flink.runtime.jobgraph.JobVertexID;
+import org.apache.flink.runtime.scheduler.strategy.ExecutionVertexID;
+import org.apache.flink.runtime.shuffle.JobShuffleContext;
+import org.apache.flink.runtime.shuffle.NettyShuffleDescriptor;
+import org.apache.flink.runtime.shuffle.PartitionDescriptor;
+import org.apache.flink.runtime.shuffle.PartitionWithMetrics;
+import org.apache.flink.runtime.shuffle.ProducerDescriptor;
+import org.apache.flink.runtime.shuffle.ShuffleDescriptor;
+import org.apache.flink.runtime.shuffle.ShuffleMasterContext;
+import org.apache.flink.runtime.shuffle.TaskInputsOutputsDescriptor;
 import org.junit.After;
 import org.junit.Assert;
 import org.junit.Before;
@@ -46,7 +59,9 @@ import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
 import org.apache.celeborn.common.CelebornConf;
+import org.apache.celeborn.common.protocol.FallbackPolicy;
 import org.apache.celeborn.common.util.Utils$;
+import org.apache.celeborn.plugin.flink.fallback.ForceFallbackPolicy;
 import org.apache.celeborn.plugin.flink.utils.FlinkUtils;
 
 public class RemoteShuffleMasterSuiteJ {
@@ -58,9 +73,10 @@ public class RemoteShuffleMasterSuiteJ {
   @Before
   public void setUp() {
     configuration = new Configuration();
-    int startPort = Utils$.MODULE$.selectRandomPort(1024, 65535);
-    configuration.setInteger("celeborn.master.port", startPort);
+    int startPort = Utils$.MODULE$.selectRandomInt(1024, 65535);
+    configuration.setString("celeborn.master.port", String.valueOf(startPort));
     configuration.setString("celeborn.master.endpoints", "localhost:" + startPort);
+    configuration.setString("celeborn.client.application.heartbeatInterval", "30s");
     remoteShuffleMaster = createShuffleMaster(configuration);
   }
 
@@ -81,6 +97,19 @@ public class RemoteShuffleMasterSuiteJ {
     remoteShuffleMaster.registerJob(jobShuffleContext);
   }
 
+  @Test
+  public void testRegisterJobWithForceFallbackPolicy() {
+    configuration.setString(
+        CelebornConf.FLINK_SHUFFLE_FALLBACK_POLICY().key(), FallbackPolicy.ALWAYS.name());
+    remoteShuffleMaster = createShuffleMaster(configuration, new NettyShuffleServiceFactory());
+    JobID jobID = JobID.generate();
+    JobShuffleContext jobShuffleContext = createJobShuffleContext(jobID);
+    remoteShuffleMaster.registerJob(jobShuffleContext);
+    Assert.assertTrue(remoteShuffleMaster.jobFallbackPolicies().containsKey(jobID));
+    remoteShuffleMaster.unregisterJob(jobShuffleContext.getJobId());
+    Assert.assertTrue(remoteShuffleMaster.jobFallbackPolicies().isEmpty());
+  }
+
   @Test
   public void testRegisterPartitionWithProducer()
       throws UnknownHostException, ExecutionException, InterruptedException {
@@ -92,9 +121,11 @@ public class RemoteShuffleMasterSuiteJ {
     PartitionDescriptor partitionDescriptor = createPartitionDescriptor(intermediateDataSetID, 0);
     ProducerDescriptor producerDescriptor = createProducerDescriptor();
     RemoteShuffleDescriptor remoteShuffleDescriptor =
-        remoteShuffleMaster
-            .registerPartitionWithProducer(jobID, partitionDescriptor, producerDescriptor)
-            .get();
+        (RemoteShuffleDescriptor)
+            remoteShuffleMaster
+                .registerPartitionWithProducer(jobID, partitionDescriptor, producerDescriptor)
+                .get();
+    Assert.assertEquals(1, remoteShuffleMaster.lifecycleManager().shuffleCount().sum());
     ShuffleResource shuffleResource = remoteShuffleDescriptor.getShuffleResource();
     ShuffleResourceDescriptor mapPartitionShuffleDescriptor =
         shuffleResource.getMapPartitionShuffleDescriptor();
@@ -108,9 +139,11 @@ public class RemoteShuffleMasterSuiteJ {
     // use same dataset id
     partitionDescriptor = createPartitionDescriptor(intermediateDataSetID, 1);
     remoteShuffleDescriptor =
-        remoteShuffleMaster
-            .registerPartitionWithProducer(jobID, partitionDescriptor, producerDescriptor)
-            .get();
+        (RemoteShuffleDescriptor)
+            remoteShuffleMaster
+                .registerPartitionWithProducer(jobID, partitionDescriptor, producerDescriptor)
+                .get();
+    Assert.assertEquals(2, remoteShuffleMaster.lifecycleManager().shuffleCount().sum());
     mapPartitionShuffleDescriptor =
         remoteShuffleDescriptor.getShuffleResource().getMapPartitionShuffleDescriptor();
     Assert.assertEquals(0, mapPartitionShuffleDescriptor.getShuffleId());
@@ -119,9 +152,11 @@ public class RemoteShuffleMasterSuiteJ {
     // use another attemptId
     producerDescriptor = createProducerDescriptor();
     remoteShuffleDescriptor =
-        remoteShuffleMaster
-            .registerPartitionWithProducer(jobID, partitionDescriptor, producerDescriptor)
-            .get();
+        (RemoteShuffleDescriptor)
+            remoteShuffleMaster
+                .registerPartitionWithProducer(jobID, partitionDescriptor, producerDescriptor)
+                .get();
+    Assert.assertEquals(3, remoteShuffleMaster.lifecycleManager().shuffleCount().sum());
     mapPartitionShuffleDescriptor =
         remoteShuffleDescriptor.getShuffleResource().getMapPartitionShuffleDescriptor();
     Assert.assertEquals(0, mapPartitionShuffleDescriptor.getShuffleId());
@@ -129,6 +164,32 @@ public class RemoteShuffleMasterSuiteJ {
     Assert.assertEquals(1, mapPartitionShuffleDescriptor.getMapId());
   }
 
+  @Test
+  public void testRegisterPartitionWithProducerForForceFallbackPolicy()
+      throws UnknownHostException, ExecutionException, InterruptedException {
+    configuration.setString(
+        CelebornConf.FLINK_SHUFFLE_FALLBACK_POLICY().key(), FallbackPolicy.ALWAYS.name());
+    remoteShuffleMaster = createShuffleMaster(configuration, new NettyShuffleServiceFactory());
+    JobID jobID = JobID.generate();
+    JobShuffleContext jobShuffleContext = createJobShuffleContext(jobID);
+    remoteShuffleMaster.registerJob(jobShuffleContext);
+
+    IntermediateDataSetID intermediateDataSetID = new IntermediateDataSetID();
+    PartitionDescriptor partitionDescriptor = createPartitionDescriptor(intermediateDataSetID, 0);
+    ProducerDescriptor producerDescriptor = createProducerDescriptor();
+    ShuffleDescriptor shuffleDescriptor =
+        remoteShuffleMaster
+            .registerPartitionWithProducer(jobID, partitionDescriptor, producerDescriptor)
+            .get();
+    Assert.assertTrue(shuffleDescriptor instanceof NettyShuffleDescriptor);
+    Assert.assertEquals(1, remoteShuffleMaster.lifecycleManager().shuffleCount().sum());
+    Map<String, Long> shuffleFallbackCounts =
+        remoteShuffleMaster.lifecycleManager().shuffleFallbackCounts();
+    Assert.assertEquals(1, shuffleFallbackCounts.size());
+    Assert.assertEquals(
+        1L, shuffleFallbackCounts.get(ForceFallbackPolicy.class.getName()).longValue());
+  }
+
   @Test
   public void testRegisterMultipleJobs()
       throws UnknownHostException, ExecutionException, InterruptedException {
@@ -144,15 +205,17 @@ public class RemoteShuffleMasterSuiteJ {
     PartitionDescriptor partitionDescriptor = createPartitionDescriptor(intermediateDataSetID, 0);
     ProducerDescriptor producerDescriptor = createProducerDescriptor();
     RemoteShuffleDescriptor remoteShuffleDescriptor1 =
-        remoteShuffleMaster
-            .registerPartitionWithProducer(jobID1, partitionDescriptor, producerDescriptor)
-            .get();
+        (RemoteShuffleDescriptor)
+            remoteShuffleMaster
+                .registerPartitionWithProducer(jobID1, partitionDescriptor, producerDescriptor)
+                .get();
 
     // use same datasetId but different jobId
     RemoteShuffleDescriptor remoteShuffleDescriptor2 =
-        remoteShuffleMaster
-            .registerPartitionWithProducer(jobID2, partitionDescriptor, producerDescriptor)
-            .get();
+        (RemoteShuffleDescriptor)
+            remoteShuffleMaster
+                .registerPartitionWithProducer(jobID2, partitionDescriptor, producerDescriptor)
+                .get();
 
     Assert.assertEquals(
         remoteShuffleDescriptor1
@@ -172,6 +235,8 @@ public class RemoteShuffleMasterSuiteJ {
   public void testShuffleMemoryAnnouncing() {
     Map<IntermediateDataSetID, Integer> numberOfInputGateChannels = new HashMap<>();
     Map<IntermediateDataSetID, Integer> numbersOfResultSubpartitions = new HashMap<>();
+    Map<IntermediateDataSetID, Integer> subPartitionNums = new HashMap<>();
+    Map<IntermediateDataSetID, ResultPartitionType> inputPartitionTypes = new HashMap<>();
     Map<IntermediateDataSetID, ResultPartitionType> resultPartitionTypes = new HashMap<>();
     IntermediateDataSetID inputDataSetID0 = new IntermediateDataSetID();
     IntermediateDataSetID inputDataSetID1 = new IntermediateDataSetID();
@@ -181,23 +246,36 @@ public class RemoteShuffleMasterSuiteJ {
     Random random = new Random();
     numberOfInputGateChannels.put(inputDataSetID0, random.nextInt(1000));
     numberOfInputGateChannels.put(inputDataSetID1, random.nextInt(1000));
-    numbersOfResultSubpartitions.put(outputDataSetID0, random.nextInt(1000));
-    numbersOfResultSubpartitions.put(outputDataSetID1, random.nextInt(1000));
-    numbersOfResultSubpartitions.put(outputDataSetID2, random.nextInt(1000));
+    int subPartitionNum1 = random.nextInt(1000);
+    int subPartitionNum2 = random.nextInt(1000);
+    int subPartitionNum3 = random.nextInt(1000);
+    numbersOfResultSubpartitions.put(outputDataSetID0, subPartitionNum1);
+    numbersOfResultSubpartitions.put(outputDataSetID1, subPartitionNum2);
+    numbersOfResultSubpartitions.put(outputDataSetID2, subPartitionNum3);
+    subPartitionNums.put(outputDataSetID0, subPartitionNum1);
+    subPartitionNums.put(outputDataSetID1, subPartitionNum2);
+    subPartitionNums.put(outputDataSetID2, subPartitionNum3);
+    inputPartitionTypes.put(inputDataSetID0, ResultPartitionType.BLOCKING);
+    inputPartitionTypes.put(inputDataSetID1, ResultPartitionType.BLOCKING);
     resultPartitionTypes.put(outputDataSetID0, ResultPartitionType.BLOCKING);
     resultPartitionTypes.put(outputDataSetID1, ResultPartitionType.BLOCKING);
     resultPartitionTypes.put(outputDataSetID2, ResultPartitionType.BLOCKING);
     MemorySize calculated =
         remoteShuffleMaster.computeShuffleMemorySizeForTask(
             TaskInputsOutputsDescriptor.from(
-                numberOfInputGateChannels, numbersOfResultSubpartitions, resultPartitionTypes));
+                3,
+                numberOfInputGateChannels,
+                numbersOfResultSubpartitions,
+                subPartitionNums,
+                inputPartitionTypes,
+                resultPartitionTypes));
 
-    CelebornConf conf = FlinkUtils.toCelebornConf(configuration);
+    CelebornConf celebornConf = FlinkUtils.toCelebornConf(configuration);
 
-    long numBytesPerGate = conf.clientFlinkMemoryPerInputGate();
+    long numBytesPerGate = celebornConf.clientFlinkMemoryPerInputGate();
     long expectedInput = 2 * numBytesPerGate;
 
-    long numBytesPerResultPartition = conf.clientFlinkMemoryPerResultPartition();
+    long numBytesPerResultPartition = celebornConf.clientFlinkMemoryPerResultPartition();
     long expectedOutput = 3 * numBytesPerResultPartition;
     MemorySize expected = new MemorySize(expectedInput + expectedOutput);
 
@@ -218,6 +296,14 @@ public class RemoteShuffleMasterSuiteJ {
                     .set(
                         ExecutionOptions.BATCH_SHUFFLE_MODE,
                         BatchShuffleMode.ALL_EXCHANGES_PIPELINED)));
+    Configuration configuration = new Configuration();
+    configuration.setString(CelebornConf.CLIENT_PUSH_REPLICATE_ENABLED().key(), "true");
+    Assert.assertThrows(
+        String.format(
+            "Flink does not support replicate shuffle data. Please check the config %s.",
+            CelebornConf.CLIENT_PUSH_REPLICATE_ENABLED().key()),
+        IllegalArgumentException.class,
+        () -> createShuffleMaster(configuration));
   }
 
   @After
@@ -232,6 +318,11 @@ public class RemoteShuffleMasterSuiteJ {
   }
 
   public RemoteShuffleMaster createShuffleMaster(Configuration configuration) {
+    return createShuffleMaster(configuration, null);
+  }
+
+  public RemoteShuffleMaster createShuffleMaster(
+      Configuration configuration, NettyShuffleServiceFactory nettyShuffleServiceFactory) {
     remoteShuffleMaster =
         new RemoteShuffleMaster(
             new ShuffleMasterContext() {
@@ -245,7 +336,7 @@ public class RemoteShuffleMasterSuiteJ {
                 System.exit(-1);
               }
             },
-            new SimpleResultPartitionAdapter());
+            nettyShuffleServiceFactory);
 
     return remoteShuffleMaster;
   }
@@ -262,6 +353,17 @@ public class RemoteShuffleMasterSuiteJ {
           Collection<ResultPartitionID> collection) {
         return CompletableFuture.completedFuture(null);
       }
+
+      @Override
+      public CompletableFuture<Collection<PartitionWithMetrics>> getPartitionWithMetrics(
+          Duration duration, Set<ResultPartitionID> set) {
+        return CompletableFuture.completedFuture(null);
+      }
+
+      @Override
+      public void notifyPartitionRecoveryStarted() {
+        // no-op
+      }
     };
   }
 
@@ -269,17 +371,24 @@ public class RemoteShuffleMasterSuiteJ {
       IntermediateDataSetID intermediateDataSetId, int partitionNum) {
     IntermediateResultPartitionID intermediateResultPartitionId =
         new IntermediateResultPartitionID(intermediateDataSetId, partitionNum);
+    // Flink 1.19.0
+    // [FLINK-33743][runtime] Align watermark at subpartition granularity
     return new PartitionDescriptor(
         intermediateDataSetId,
         10,
         intermediateResultPartitionId,
         ResultPartitionType.BLOCKING,
         5,
-        1);
+        1,
+        false,
+        false,
+        false);
   }
 
   public ProducerDescriptor createProducerDescriptor() throws UnknownHostException {
-    ExecutionAttemptID executionAttemptId = new ExecutionAttemptID();
+    ExecutionAttemptID executionAttemptId =
+        new ExecutionAttemptID(
+            new ExecutionGraphID(), new ExecutionVertexID(new JobVertexID(0, 0), 0), 0);
     return new ProducerDescriptor(
         ResourceID.generate(), executionAttemptId, InetAddress.getLocalHost(), 100);
   }
