@@ -19,9 +19,15 @@ package org.apache.celeborn.client;
 
 import java.io.IOException;
 import java.util.ArrayList;
+import java.util.Map;
+import java.util.Optional;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.atomic.LongAdder;
+import java.util.function.BiFunction;
 
+import scala.Tuple2;
+
+import org.apache.commons.lang3.tuple.Pair;
 import org.apache.hadoop.fs.FileSystem;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -34,9 +40,12 @@ import org.apache.celeborn.common.identity.UserIdentifier;
 import org.apache.celeborn.common.network.client.TransportClientFactory;
 import org.apache.celeborn.common.protocol.PartitionLocation;
 import org.apache.celeborn.common.protocol.PbStreamHandler;
+import org.apache.celeborn.common.protocol.StorageInfo;
+import org.apache.celeborn.common.protocol.message.ControlMessages;
 import org.apache.celeborn.common.rpc.RpcEndpointRef;
 import org.apache.celeborn.common.util.CelebornHadoopUtils;
 import org.apache.celeborn.common.util.ExceptionMaker;
+import org.apache.celeborn.common.write.LocationPushFailedBatches;
 import org.apache.celeborn.common.write.PushState;
 
 /**
@@ -47,15 +56,19 @@ public abstract class ShuffleClient {
   private static Logger logger = LoggerFactory.getLogger(ShuffleClient.class);
   private static volatile ShuffleClient _instance;
   private static volatile boolean initialized = false;
-  private static volatile FileSystem hdfsFs;
+  private static volatile Map<StorageInfo.Type, FileSystem> hadoopFs;
   private static LongAdder totalReadCounter = new LongAdder();
   private static LongAdder localShuffleReadCounter = new LongAdder();
 
+  private static volatile Optional<
+          BiFunction<Integer, byte[], ControlMessages.GetReducerFileGroupResponse>>
+      deserializeReducerFileGroupResponseFunction = Optional.empty();
+
   // for testing
   public static void reset() {
     _instance = null;
     initialized = false;
-    hdfsFs = null;
+    hadoopFs = null;
   }
 
   protected ShuffleClient() {}
@@ -101,19 +114,19 @@ public abstract class ShuffleClient {
     return _instance;
   }
 
-  public static FileSystem getHdfsFs(CelebornConf conf) {
-    if (null == hdfsFs) {
+  public static Map<StorageInfo.Type, FileSystem> getHadoopFs(CelebornConf conf) {
+    if (null == hadoopFs) {
       synchronized (ShuffleClient.class) {
-        if (null == hdfsFs) {
+        if (null == hadoopFs) {
           try {
-            hdfsFs = CelebornHadoopUtils.getHadoopFS(conf);
+            hadoopFs = CelebornHadoopUtils.getHadoopFS(conf);
           } catch (Exception e) {
-            logger.error("Celeborn initialize HDFS failed.", e);
+            logger.error("Celeborn initialize DFS failed.", e);
           }
         }
       }
     }
-    return hdfsFs;
+    return hadoopFs;
   }
 
   public static void incrementLocalReadCounter() {
@@ -204,6 +217,8 @@ public abstract class ShuffleClient {
   public abstract ShuffleClientImpl.ReduceFileGroups updateFileGroup(int shuffleId, int partitionId)
       throws CelebornIOException;
 
+  public abstract boolean isShuffleStageEnd(int shuffleId) throws Exception;
+
   // Reduce side read partition which is deduplicated by mapperId+mapperAttemptNum+batchId, batchId
   // is a self-incrementing variable hidden in the implementation when sending data.
   /**
@@ -222,6 +237,7 @@ public abstract class ShuffleClient {
       int shuffleId,
       int partitionId,
       int attemptNumber,
+      long taskId,
       int startMapIndex,
       int endMapIndex,
       MetricsCallback metricsCallback)
@@ -231,13 +247,17 @@ public abstract class ShuffleClient {
         shuffleId,
         partitionId,
         attemptNumber,
+        taskId,
         startMapIndex,
         endMapIndex,
         null,
         null,
         null,
         null,
-        metricsCallback);
+        null,
+        null,
+        metricsCallback,
+        true);
   }
 
   public abstract CelebornInputStream readPartition(
@@ -245,13 +265,17 @@ public abstract class ShuffleClient {
       int appShuffleId,
       int partitionId,
       int attemptNumber,
+      long taskId,
       int startMapIndex,
       int endMapIndex,
       ExceptionMaker exceptionMaker,
       ArrayList<PartitionLocation> locations,
       ArrayList<PbStreamHandler> streamHandlers,
+      Map<String, LocationPushFailedBatches> failedBatchSetMap,
+      Map<String, Pair<Integer, Integer>> chunksRange,
       int[] mapAttempts,
-      MetricsCallback metricsCallback)
+      MetricsCallback metricsCallback,
+      boolean needDecompress)
       throws IOException;
 
   public abstract boolean cleanupShuffle(int shuffleId);
@@ -266,7 +290,7 @@ public abstract class ShuffleClient {
 
   public abstract PushState getPushState(String mapKey);
 
-  public abstract int getShuffleId(
+  public abstract Tuple2<Integer, Boolean> getShuffleId(
       int appShuffleId, String appShuffleIdentifier, boolean isWriter, boolean isBarrierStage);
 
   /**
@@ -274,7 +298,7 @@ public abstract class ShuffleClient {
    * cleanup for spark app. It must be a sync call and make sure the cleanup is done, otherwise,
    * incorrect shuffle data can be fetched in re-run tasks
    */
-  public abstract boolean reportShuffleFetchFailure(int appShuffleId, int shuffleId);
+  public abstract boolean reportShuffleFetchFailure(int appShuffleId, int shuffleId, long taskId);
 
   /**
    * Report barrier task failure. When any barrier task fails, all (shuffle) output for that stage
@@ -285,4 +309,21 @@ public abstract class ShuffleClient {
   public abstract TransportClientFactory getDataClientFactory();
 
   public abstract void excludeFailedFetchLocation(String hostAndFetchPort, Exception e);
+
+  public static void registerDeserializeReducerFileGroupResponseFunction(
+      BiFunction<Integer, byte[], ControlMessages.GetReducerFileGroupResponse> function) {
+    if (!deserializeReducerFileGroupResponseFunction.isPresent()) {
+      deserializeReducerFileGroupResponseFunction = Optional.ofNullable(function);
+    }
+  }
+
+  public static ControlMessages.GetReducerFileGroupResponse deserializeReducerFileGroupResponse(
+      int shuffleId, byte[] bytes) {
+    if (!deserializeReducerFileGroupResponseFunction.isPresent()) {
+      // Should never happen
+      logger.warn("DeserializeReducerFileGroupResponseFunction is not registered.");
+      return null;
+    }
+    return deserializeReducerFileGroupResponseFunction.get().apply(shuffleId, bytes);
+  }
 }
