@@ -31,6 +31,7 @@ import scala.reflect.ClassTag$;
 
 import com.google.common.annotations.VisibleForTesting;
 import com.google.common.collect.Lists;
+import com.google.protobuf.InvalidProtocolBufferException;
 import io.netty.buffer.CompositeByteBuf;
 import io.netty.buffer.Unpooled;
 import org.apache.commons.lang3.StringUtils;
@@ -42,17 +43,19 @@ import org.apache.celeborn.client.compress.Compressor;
 import org.apache.celeborn.client.read.CelebornInputStream;
 import org.apache.celeborn.client.read.MetricsCallback;
 import org.apache.celeborn.common.CelebornConf;
+import org.apache.celeborn.common.exception.CelebornBroadcastException;
 import org.apache.celeborn.common.exception.CelebornIOException;
 import org.apache.celeborn.common.exception.CelebornRuntimeException;
 import org.apache.celeborn.common.identity.UserIdentifier;
+import org.apache.celeborn.common.metrics.source.Role;
 import org.apache.celeborn.common.network.TransportContext;
 import org.apache.celeborn.common.network.buffer.NettyManagedBuffer;
 import org.apache.celeborn.common.network.client.RpcResponseCallback;
 import org.apache.celeborn.common.network.client.TransportClient;
 import org.apache.celeborn.common.network.client.TransportClientBootstrap;
 import org.apache.celeborn.common.network.client.TransportClientFactory;
-import org.apache.celeborn.common.network.protocol.PushData;
-import org.apache.celeborn.common.network.protocol.PushMergedData;
+import org.apache.celeborn.common.network.protocol.*;
+import org.apache.celeborn.common.network.protocol.SerdeVersion;
 import org.apache.celeborn.common.network.sasl.SaslClientBootstrap;
 import org.apache.celeborn.common.network.sasl.SaslCredentials;
 import org.apache.celeborn.common.network.server.BaseMessageHandler;
@@ -66,6 +69,7 @@ import org.apache.celeborn.common.rpc.RpcEnv;
 import org.apache.celeborn.common.unsafe.Platform;
 import org.apache.celeborn.common.util.*;
 import org.apache.celeborn.common.write.DataBatches;
+import org.apache.celeborn.common.write.LocationPushFailedBatches;
 import org.apache.celeborn.common.write.PushState;
 
 public class ShuffleClientImpl extends ShuffleClient {
@@ -100,7 +104,7 @@ public class ShuffleClientImpl extends ShuffleClient {
   protected byte[] extension;
 
   // key: appShuffleIdentifier, value: shuffleId
-  protected Map<String, Integer> shuffleIdCache = JavaUtils.newConcurrentHashMap();
+  protected Map<String, Tuple2<Integer, Boolean>> shuffleIdCache = JavaUtils.newConcurrentHashMap();
 
   // key: shuffleId, value: (partitionId, PartitionLocation)
   final Map<Integer, ConcurrentHashMap<Integer, PartitionLocation>> reducePartitionMap =
@@ -132,6 +136,7 @@ public class ShuffleClientImpl extends ShuffleClient {
   private final boolean authEnabled;
   private final TransportConf dataTransportConf;
 
+  @SuppressWarnings("ThreadLocalUsage")
   private final ThreadLocal<Compressor> compressorThreadLocal =
       new ThreadLocal<Compressor>() {
         @Override
@@ -142,30 +147,37 @@ public class ShuffleClientImpl extends ShuffleClient {
 
   private final ReviveManager reviveManager;
 
+  private final boolean dataPushFailureTrackingEnabled;
+
   public static class ReduceFileGroups {
     public Map<Integer, Set<PartitionLocation>> partitionGroups;
+    public Map<String, LocationPushFailedBatches> pushFailedBatches;
     public int[] mapAttempts;
     public Set<Integer> partitionIds;
 
     ReduceFileGroups(
         Map<Integer, Set<PartitionLocation>> partitionGroups,
         int[] mapAttempts,
-        Set<Integer> partitionIds) {
+        Set<Integer> partitionIds,
+        Map<String, LocationPushFailedBatches> pushFailedBatches) {
       this.partitionGroups = partitionGroups;
       this.mapAttempts = mapAttempts;
       this.partitionIds = partitionIds;
+      this.pushFailedBatches = pushFailedBatches;
     }
 
     public ReduceFileGroups() {
       this.partitionGroups = null;
       this.mapAttempts = null;
       this.partitionIds = null;
+      this.pushFailedBatches = null;
     }
 
     public void update(ReduceFileGroups fileGroups) {
       partitionGroups = fileGroups.partitionGroups;
       mapAttempts = fileGroups.mapAttempts;
       partitionIds = fileGroups.partitionIds;
+      pushFailedBatches = fileGroups.pushFailedBatches;
     }
   }
 
@@ -173,6 +185,8 @@ public class ShuffleClientImpl extends ShuffleClient {
   protected final Map<Integer, Tuple3<ReduceFileGroups, String, Exception>> reduceFileGroupsMap =
       JavaUtils.newConcurrentHashMap();
 
+  private final TransportMessagesHelper messagesHelper = new TransportMessagesHelper();
+
   public ShuffleClientImpl(String appUniqueId, CelebornConf conf, UserIdentifier userIdentifier) {
     super();
     this.appUniqueId = appUniqueId;
@@ -195,6 +209,7 @@ public class ShuffleClientImpl extends ShuffleClient {
       pushDataTimeout = conf.pushDataTimeoutMs();
     }
     authEnabled = conf.authEnabledOnClient();
+    dataPushFailureTrackingEnabled = conf.clientAdaptiveOptimizeSkewedPartitionReadEnabled();
 
     // init rpc env
     rpcEnv =
@@ -204,11 +219,11 @@ public class ShuffleClientImpl extends ShuffleClient {
             Utils.localHostName(conf),
             0,
             conf,
+            Role.CLIENT(),
             scala.None$.empty());
 
     String module = TransportModuleConstants.DATA_MODULE;
-    dataTransportConf =
-        Utils.fromCelebornConf(conf, module, conf.getInt("celeborn." + module + ".io.threads", 8));
+    dataTransportConf = Utils.fromCelebornConf(conf, module, conf.networkIoThreads(module));
     initDataClientFactoryIfNeeded();
     int pushDataRetryThreads = conf.clientPushRetryThreads();
     pushDataRetryPool =
@@ -320,7 +335,7 @@ public class ShuffleClientImpl extends ShuffleClient {
                   + ", revive status "
                   + request.reviveStatus
                   + "("
-                  + Utils.toStatusCode(request.reviveStatus)
+                  + StatusCode.fromValue(request.reviveStatus)
                   + ")"
                   + ", old location: "
                   + request.loc));
@@ -363,8 +378,12 @@ public class ShuffleClientImpl extends ShuffleClient {
             batchId,
             newLoc,
             e);
-        pushDataRpcResponseCallback.onFailure(
-            new CelebornIOException(StatusCode.PUSH_DATA_CREATE_CONNECTION_FAIL_PRIMARY, e));
+        if (e instanceof InterruptedException) {
+          pushDataRpcResponseCallback.onFailure(e);
+        } else {
+          pushDataRpcResponseCallback.onFailure(
+              new CelebornIOException(StatusCode.PUSH_DATA_CREATE_CONNECTION_FAIL_PRIMARY, e));
+        }
       }
     }
   }
@@ -439,7 +458,7 @@ public class ShuffleClientImpl extends ShuffleClient {
                             + " then revive but "
                             + request.reviveStatus
                             + "("
-                            + Utils.toStatusCode(request.reviveStatus)
+                            + StatusCode.fromValue(request.reviveStatus)
                             + ")")));
             return;
           }
@@ -475,7 +494,7 @@ public class ShuffleClientImpl extends ShuffleClient {
                         + " then revive but "
                         + request.reviveStatus
                         + "("
-                        + Utils.toStatusCode(request.reviveStatus)
+                        + StatusCode.fromValue(request.reviveStatus)
                         + ")")));
         return;
       }
@@ -541,6 +560,17 @@ public class ShuffleClientImpl extends ShuffleClient {
   @Override
   public PartitionLocation registerMapPartitionTask(
       int shuffleId, int numMappers, int mapId, int attemptId, int partitionId) throws IOException {
+    return registerMapPartitionTask(shuffleId, numMappers, mapId, attemptId, partitionId, false);
+  }
+
+  public PartitionLocation registerMapPartitionTask(
+      int shuffleId,
+      int numMappers,
+      int mapId,
+      int attemptId,
+      int partitionId,
+      boolean isSegmentGranularityVisible)
+      throws IOException {
     logger.info(
         "Register MapPartition task for shuffle {} map {} attempt {} partition {} with {} mapper.",
         shuffleId,
@@ -556,7 +586,12 @@ public class ShuffleClientImpl extends ShuffleClient {
             () ->
                 lifecycleManagerRef.askSync(
                     RegisterMapPartitionTask$.MODULE$.apply(
-                        shuffleId, numMappers, mapId, attemptId, partitionId),
+                        shuffleId,
+                        numMappers,
+                        mapId,
+                        attemptId,
+                        partitionId,
+                        isSegmentGranularityVisible),
                     conf.clientRpcRegisterShuffleAskTimeout(),
                     ClassTag$.MODULE$.apply(PbRegisterShuffleResponse.class)));
 
@@ -591,7 +626,7 @@ public class ShuffleClientImpl extends ShuffleClient {
   }
 
   @Override
-  public int getShuffleId(
+  public Tuple2<Integer, Boolean> getShuffleId(
       int appShuffleId, String appShuffleIdentifier, boolean isWriter, boolean isBarrierStage) {
     return shuffleIdCache.computeIfAbsent(
         appShuffleIdentifier,
@@ -608,16 +643,18 @@ public class ShuffleClientImpl extends ShuffleClient {
                   pbGetShuffleId,
                   conf.clientRpcRegisterShuffleAskTimeout(),
                   ClassTag$.MODULE$.apply(PbGetShuffleIdResponse.class));
-          return pbGetShuffleIdResponse.getShuffleId();
+          return Tuple2.apply(
+              pbGetShuffleIdResponse.getShuffleId(), pbGetShuffleIdResponse.getSuccess());
         });
   }
 
   @Override
-  public boolean reportShuffleFetchFailure(int appShuffleId, int shuffleId) {
+  public boolean reportShuffleFetchFailure(int appShuffleId, int shuffleId, long taskId) {
     PbReportShuffleFetchFailure pbReportShuffleFetchFailure =
         PbReportShuffleFetchFailure.newBuilder()
             .setAppShuffleId(appShuffleId)
             .setShuffleId(shuffleId)
+            .setTaskId(taskId)
             .build();
     PbReportShuffleFetchFailureResponse pbReportShuffleFetchFailureResponse =
         lifecycleManagerRef.askSync(
@@ -627,6 +664,7 @@ public class ShuffleClientImpl extends ShuffleClient {
     return pbReportShuffleFetchFailureResponse.getSuccess();
   }
 
+  @Override
   public boolean reportBarrierTaskFailure(int appShuffleId, String appShuffleIdentifier) {
     PbReportBarrierStageAttemptFailure pbReportBarrierStageAttemptFailure =
         PbReportBarrierStageAttemptFailure.newBuilder()
@@ -652,7 +690,7 @@ public class ShuffleClientImpl extends ShuffleClient {
     while (numRetries > 0) {
       try {
         PbRegisterShuffleResponse response = callable.call();
-        StatusCode respStatus = Utils.toStatusCode(response.getStatus());
+        StatusCode respStatus = StatusCode.fromValue(response.getStatus());
         if (StatusCode.SUCCESS.equals(respStatus)) {
           ConcurrentHashMap<Integer, PartitionLocation> result = JavaUtils.newConcurrentHashMap();
           Tuple2<List<PartitionLocation>, List<PartitionLocation>> locations =
@@ -686,6 +724,9 @@ public class ShuffleClientImpl extends ShuffleClient {
               numRetries - 1);
         }
       } catch (Exception e) {
+        if (e instanceof InterruptedException) {
+          Thread.currentThread().interrupt();
+        }
         logger.error(
             "Exception raised while registering shuffle {} with {} mapper and {} partitions.",
             shuffleId,
@@ -775,6 +816,7 @@ public class ShuffleClientImpl extends ShuffleClient {
         case PUSH_DATA_TIMEOUT_REPLICA:
           pushExcludedWorkers.add(oldLocation.getPeer().hostAndPushPort());
           break;
+        default: // fall out
       }
     }
   }
@@ -883,6 +925,9 @@ public class ShuffleClientImpl extends ShuffleClient {
           partitionIds,
           epochs,
           e);
+      if (e instanceof InterruptedException) {
+        Thread.currentThread().interrupt();
+      }
       return null;
     }
   }
@@ -1076,6 +1121,10 @@ public class ShuffleClientImpl extends ShuffleClient {
                       attemptId,
                       partitionId,
                       nextBatchId);
+                  if (dataPushFailureTrackingEnabled && pushReplicateEnabled) {
+                    pushState.recordFailedBatch(
+                        latest.getUniqueId(), mapId, attemptId, nextBatchId);
+                  }
                   ReviveRequest reviveRequest =
                       new ReviveRequest(
                           shuffleId,
@@ -1142,9 +1191,17 @@ public class ShuffleClientImpl extends ShuffleClient {
 
             @Override
             public void onFailure(Throwable e) {
+              if (dataPushFailureTrackingEnabled) {
+                pushState.recordFailedBatch(latest.getUniqueId(), mapId, attemptId, nextBatchId);
+              }
               if (pushState.exception.get() != null) {
                 return;
               }
+              if (e instanceof InterruptedException) {
+                Thread.currentThread().interrupt();
+                callback.onFailure(e);
+                return;
+              }
               StatusCode cause = getPushDataFailCause(e.getMessage());
               if (remainReviveTimes <= 0) {
                 if (e instanceof CelebornIOException) {
@@ -1226,8 +1283,12 @@ public class ShuffleClientImpl extends ShuffleClient {
             nextBatchId,
             loc,
             e);
-        wrappedCallback.onFailure(
-            new CelebornIOException(StatusCode.PUSH_DATA_CREATE_CONNECTION_FAIL_PRIMARY, e));
+        if (e instanceof InterruptedException) {
+          wrappedCallback.onFailure(e);
+        } else {
+          wrappedCallback.onFailure(
+              new CelebornIOException(StatusCode.PUSH_DATA_CREATE_CONNECTION_FAIL_PRIMARY, e));
+        }
       }
     } else {
       // add batch data
@@ -1427,9 +1488,65 @@ public class ShuffleClientImpl extends ShuffleClient {
         new RpcResponseCallback() {
           @Override
           public void onSuccess(ByteBuffer response) {
-            if (response.remaining() > 0) {
-              byte reason = response.get();
-              if (reason == StatusCode.HARD_SPLIT.getValue()) {
+            byte reason = response.get();
+            if (reason == StatusCode.HARD_SPLIT.getValue()
+                || reason == StatusCode.SOFT_SPLIT.getValue()) {
+              ArrayList<DataBatches.DataBatch> batchesNeedResubmit;
+              if (response.remaining() > 0) {
+                batchesNeedResubmit = new ArrayList<>();
+                PbPushMergedDataSplitPartitionInfo partitionInfo;
+                try {
+                  partitionInfo = TransportMessage.fromByteBuffer(response).getParsedPayload();
+                } catch (CelebornIOException | InvalidProtocolBufferException e) {
+                  callback.onFailure(
+                      new CelebornIOException("parse pushMergedData response failed", e));
+                  return;
+                }
+                List<Integer> splitPartitionIndexes = partitionInfo.getSplitPartitionIndexesList();
+                List<Integer> statusCodeList = partitionInfo.getStatusCodesList();
+                StringBuilder dataBatchReviveInfos = new StringBuilder();
+                for (int i = 0; i < splitPartitionIndexes.size(); i++) {
+                  int partitionIndex = splitPartitionIndexes.get(i);
+                  int batchId = batches.get(partitionIndex).batchId;
+                  dataBatchReviveInfos.append(
+                      String.format(
+                          "(batchId=%d, partitionId=%d, cause=%s)",
+                          batchId,
+                          partitionIds[partitionIndex],
+                          StatusCode.fromValue(statusCodeList.get(i).byteValue())));
+                  if (statusCodeList.get(i) == StatusCode.SOFT_SPLIT.getValue()) {
+                    PartitionLocation loc = batches.get(partitionIndex).loc;
+                    if (!newerPartitionLocationExists(
+                        reducePartitionMap.get(shuffleId), loc.getId(), loc.getEpoch(), false)) {
+                      ReviveRequest reviveRequest =
+                          new ReviveRequest(
+                              shuffleId,
+                              mapId,
+                              attemptId,
+                              loc.getId(),
+                              loc.getEpoch(),
+                              loc,
+                              StatusCode.SOFT_SPLIT);
+                      reviveManager.addRequest(reviveRequest);
+                    }
+                  } else {
+                    batchesNeedResubmit.add(batches.get(partitionIndex));
+                  }
+                }
+                logger.info(
+                    "Push merged data to {} partial success required for shuffle {} map {} attempt {} groupedBatch {}. split batches {}.",
+                    addressPair,
+                    shuffleId,
+                    mapId,
+                    attemptId,
+                    groupedBatchId,
+                    dataBatchReviveInfos);
+              } else {
+                // Workers that do not incorporate changes from [CELEBORN-1721]
+                // will respond with a status of HARD_SPLIT,
+                // but will not include a PbPushMergedDataSplitPartitionInfo.
+                // For backward compatibility, all batches must be resubmitted.
+                batchesNeedResubmit = batches;
                 logger.info(
                     "Push merged data to {} hard split required for shuffle {} map {} attempt {} partition {} groupedBatch {} batch {}.",
                     addressPair,
@@ -1439,10 +1556,20 @@ public class ShuffleClientImpl extends ShuffleClient {
                     Arrays.toString(partitionIds),
                     groupedBatchId,
                     Arrays.toString(batchIds));
-
+              }
+              if (batchesNeedResubmit.isEmpty()) {
+                pushState.onSuccess(hostPort);
+                callback.onSuccess(ByteBuffer.wrap(new byte[] {StatusCode.SOFT_SPLIT.getValue()}));
+              } else {
+                if (dataPushFailureTrackingEnabled && pushReplicateEnabled) {
+                  for (DataBatches.DataBatch resubmitBatch : batchesNeedResubmit) {
+                    pushState.recordFailedBatch(
+                        resubmitBatch.loc.getUniqueId(), mapId, attemptId, resubmitBatch.batchId);
+                  }
+                }
                 ReviveRequest[] requests =
                     addAndGetReviveRequests(
-                        shuffleId, mapId, attemptId, batches, StatusCode.HARD_SPLIT);
+                        shuffleId, mapId, attemptId, batchesNeedResubmit, StatusCode.HARD_SPLIT);
                 pushDataRetryPool.submit(
                     () ->
                         submitRetryPushMergedData(
@@ -1450,7 +1577,7 @@ public class ShuffleClientImpl extends ShuffleClient {
                             shuffleId,
                             mapId,
                             attemptId,
-                            batches,
+                            batchesNeedResubmit,
                             StatusCode.HARD_SPLIT,
                             groupedBatchId,
                             requests,
@@ -1459,47 +1586,55 @@ public class ShuffleClientImpl extends ShuffleClient {
                                 + conf.clientRpcRequestPartitionLocationAskTimeout()
                                     .duration()
                                     .toMillis()));
-              } else if (reason == StatusCode.PUSH_DATA_SUCCESS_PRIMARY_CONGESTED.getValue()) {
-                logger.debug(
-                    "Push merged data to {} primary congestion required for shuffle {} map {} attempt {} partition {} groupedBatch {} batch {}.",
-                    addressPair,
-                    shuffleId,
-                    mapId,
-                    attemptId,
-                    Arrays.toString(partitionIds),
-                    groupedBatchId,
-                    Arrays.toString(batchIds));
-                pushState.onCongestControl(hostPort);
-                callback.onSuccess(response);
-              } else if (reason == StatusCode.PUSH_DATA_SUCCESS_REPLICA_CONGESTED.getValue()) {
-                logger.debug(
-                    "Push merged data to {} replica congestion required for shuffle {} map {} attempt {} partition {} groupedBatch {} batch {}.",
-                    addressPair,
-                    shuffleId,
-                    mapId,
-                    attemptId,
-                    Arrays.toString(partitionIds),
-                    groupedBatchId,
-                    Arrays.toString(batchIds));
-                pushState.onCongestControl(hostPort);
-                callback.onSuccess(response);
-              } else {
-                // StageEnd.
-                response.rewind();
-                pushState.onSuccess(hostPort);
-                callback.onSuccess(response);
               }
-            } else {
-              pushState.onSuccess(hostPort);
+            } else if (reason == StatusCode.PUSH_DATA_SUCCESS_PRIMARY_CONGESTED.getValue()) {
+              logger.debug(
+                  "Push merged data to {} primary congestion required for shuffle {} map {} attempt {} partition {} groupedBatch {} batch {}.",
+                  addressPair,
+                  shuffleId,
+                  mapId,
+                  attemptId,
+                  Arrays.toString(partitionIds),
+                  groupedBatchId,
+                  Arrays.toString(batchIds));
+              pushState.onCongestControl(hostPort);
               callback.onSuccess(response);
+            } else if (reason == StatusCode.PUSH_DATA_SUCCESS_REPLICA_CONGESTED.getValue()) {
+              logger.debug(
+                  "Push merged data to {} replica congestion required for shuffle {} map {} attempt {} partition {} groupedBatch {} batch {}.",
+                  addressPair,
+                  shuffleId,
+                  mapId,
+                  attemptId,
+                  Arrays.toString(partitionIds),
+                  groupedBatchId,
+                  Arrays.toString(batchIds));
+              pushState.onCongestControl(hostPort);
+              callback.onSuccess(response);
+            } else if (reason == StatusCode.MAP_ENDED.getValue()) {
+              pushState.onSuccess(hostPort);
+              callback.onSuccess(ByteBuffer.wrap(new byte[] {StatusCode.MAP_ENDED.getValue()}));
+            } else { // success
+              pushState.onSuccess(hostPort);
+              callback.onSuccess(ByteBuffer.wrap(new byte[] {StatusCode.SUCCESS.getValue()}));
             }
           }
 
           @Override
           public void onFailure(Throwable e) {
+            if (dataPushFailureTrackingEnabled) {
+              for (int i = 0; i < numBatches; i++) {
+                pushState.recordFailedBatch(partitionUniqueIds[i], mapId, attemptId, batchIds[i]);
+              }
+            }
             if (pushState.exception.get() != null) {
               return;
             }
+            if (e instanceof InterruptedException) {
+              Thread.currentThread().interrupt();
+              callback.onFailure(e);
+              return;
+            }
             StatusCode cause = getPushDataFailCause(e.getMessage());
             if (remainReviveTimes <= 0) {
               if (e instanceof CelebornIOException) {
@@ -1580,8 +1715,12 @@ public class ShuffleClientImpl extends ShuffleClient {
           Arrays.toString(batchIds),
           addressPair,
           e);
-      wrappedCallback.onFailure(
-          new CelebornIOException(StatusCode.PUSH_DATA_CREATE_CONNECTION_FAIL_PRIMARY, e));
+      if (e instanceof InterruptedException) {
+        wrappedCallback.onFailure(e);
+      } else {
+        wrappedCallback.onFailure(
+            new CelebornIOException(StatusCode.PUSH_DATA_CREATE_CONNECTION_FAIL_PRIMARY, e));
+      }
     }
   }
 
@@ -1608,7 +1747,13 @@ public class ShuffleClientImpl extends ShuffleClient {
 
       MapperEndResponse response =
           lifecycleManagerRef.askSync(
-              new MapperEnd(shuffleId, mapId, attemptId, numMappers, partitionId),
+              new MapperEnd(
+                  shuffleId,
+                  mapId,
+                  attemptId,
+                  numMappers,
+                  partitionId,
+                  pushState.getFailedBatches()),
               rpcMaxRetries,
               rpcRetryWait,
               ClassTag$.MODULE$.apply(MapperEndResponse.class));
@@ -1643,7 +1788,8 @@ public class ShuffleClientImpl extends ShuffleClient {
     return true;
   }
 
-  protected Tuple3<ReduceFileGroups, String, Exception> loadFileGroupInternal(int shuffleId) {
+  protected Tuple3<ReduceFileGroups, String, Exception> loadFileGroupInternal(
+      int shuffleId, boolean isSegmentGranularityVisible) {
     long getReducerFileGroupStartTime = System.nanoTime();
     String exceptionMsg = null;
     Exception exception = null;
@@ -1653,7 +1799,9 @@ public class ShuffleClientImpl extends ShuffleClient {
       return Tuple3.apply(null, exceptionMsg, exception);
     }
     try {
-      GetReducerFileGroup getReducerFileGroup = new GetReducerFileGroup(shuffleId);
+      GetReducerFileGroup getReducerFileGroup =
+          new GetReducerFileGroup(shuffleId, isSegmentGranularityVisible, SerdeVersion.V1);
+
       GetReducerFileGroupResponse response =
           lifecycleManagerRef.askSync(
               getReducerFileGroup,
@@ -1663,6 +1811,14 @@ public class ShuffleClientImpl extends ShuffleClient {
               ClassTag$.MODULE$.apply(GetReducerFileGroupResponse.class));
       switch (response.status()) {
         case SUCCESS:
+          if (response.broadcast() != null && response.broadcast().length > 0) {
+            response =
+                ShuffleClient.deserializeReducerFileGroupResponse(shuffleId, response.broadcast());
+            if (response == null) {
+              throw new CelebornBroadcastException(
+                  "Failed to get GetReducerFileGroupResponse broadcast for shuffle: " + shuffleId);
+            }
+          }
           logger.info(
               "Shuffle {} request reducer file group success using {} ms, result partition size {}.",
               shuffleId,
@@ -1670,7 +1826,10 @@ public class ShuffleClientImpl extends ShuffleClient {
               response.fileGroup().size());
           return Tuple3.apply(
               new ReduceFileGroups(
-                  response.fileGroup(), response.attempts(), response.partitionIds()),
+                  response.fileGroup(),
+                  response.attempts(),
+                  response.partitionIds(),
+                  response.pushFailedBatches()),
               null,
               null);
         case SHUFFLE_NOT_REGISTERED:
@@ -1679,7 +1838,10 @@ public class ShuffleClientImpl extends ShuffleClient {
           // return empty result
           return Tuple3.apply(
               new ReduceFileGroups(
-                  response.fileGroup(), response.attempts(), response.partitionIds()),
+                  response.fileGroup(),
+                  response.attempts(),
+                  response.partitionIds(),
+                  response.pushFailedBatches()),
               null,
               null);
         case STAGE_END_TIME_OUT:
@@ -1703,14 +1865,37 @@ public class ShuffleClientImpl extends ShuffleClient {
     return Tuple3.apply(null, exceptionMsg, exception);
   }
 
+  @Override
   public ReduceFileGroups updateFileGroup(int shuffleId, int partitionId)
       throws CelebornIOException {
+    return updateFileGroup(shuffleId, partitionId, false);
+  }
+
+  @Override
+  public boolean isShuffleStageEnd(int shuffleId) throws Exception {
+    if (null != lifecycleManagerRef) {
+      PbGetStageEnd request = PbGetStageEnd.newBuilder().setShuffleId(shuffleId).build();
+      PbGetStageEndResponse response =
+          lifecycleManagerRef.askSync(
+              request,
+              rpcMaxRetries,
+              rpcRetryWait,
+              ClassTag$.MODULE$.apply(PbGetStageEndResponse.class));
+      return response.getStageEnd();
+    } else {
+      throw new RuntimeException("Driver endpoint is null!");
+    }
+  }
+
+  public ReduceFileGroups updateFileGroup(
+      int shuffleId, int partitionId, boolean isSegmentGranularityVisible)
+      throws CelebornIOException {
     Tuple3<ReduceFileGroups, String, Exception> fileGroupTuple =
         reduceFileGroupsMap.compute(
             shuffleId,
             (id, existsTuple) -> {
               if (existsTuple == null || existsTuple._1() == null) {
-                return loadFileGroupInternal(shuffleId);
+                return loadFileGroupInternal(shuffleId, isSegmentGranularityVisible);
               } else {
                 return existsTuple;
               }
@@ -1738,13 +1923,17 @@ public class ShuffleClientImpl extends ShuffleClient {
       int appShuffleId,
       int partitionId,
       int attemptNumber,
+      long taskId,
       int startMapIndex,
       int endMapIndex,
       ExceptionMaker exceptionMaker,
       ArrayList<PartitionLocation> locations,
       ArrayList<PbStreamHandler> streamHandlers,
+      Map<String, LocationPushFailedBatches> failedBatchSetMap,
+      Map<String, Pair<Integer, Integer>> chunksRange,
       int[] mapAttempts,
-      MetricsCallback metricsCallback)
+      MetricsCallback metricsCallback,
+      boolean needDecompress)
       throws IOException {
     if (shuffleId == Utils$.MODULE$.UNKNOWN_APP_SHUFFLE_ID()) {
       logger.warn("Shuffle data is empty for shuffle {}: UNKNOWN_APP_SHUFFLE_ID.", shuffleId);
@@ -1755,7 +1944,7 @@ public class ShuffleClientImpl extends ShuffleClient {
     // CelebornShuffleReader, which means `updateFileGroup` is already called and
     // batch open stream has been tried
     if (mapAttempts == null) {
-      ReduceFileGroups fileGroups = updateFileGroup(shuffleId, partitionId);
+      ReduceFileGroups fileGroups = updateFileGroup(shuffleId, partitionId, false);
       mapAttempts = fileGroups.mapAttempts;
       if (fileGroups.partitionGroups.containsKey(partitionId)) {
         locations = new ArrayList(fileGroups.partitionGroups.get(partitionId));
@@ -1775,7 +1964,10 @@ public class ShuffleClientImpl extends ShuffleClient {
           locations,
           streamHandlers,
           mapAttempts,
+          failedBatchSetMap,
+          chunksRange,
           attemptNumber,
+          taskId,
           startMapIndex,
           endMapIndex,
           fetchExcludedWorkers,
@@ -1784,7 +1976,8 @@ public class ShuffleClientImpl extends ShuffleClient {
           shuffleId,
           partitionId,
           exceptionMaker,
-          metricsCallback);
+          metricsCallback,
+          needDecompress);
     }
   }
 
@@ -1817,6 +2010,7 @@ public class ShuffleClientImpl extends ShuffleClient {
     shuffleIdCache.clear();
     pushExcludedWorkers.clear();
     fetchExcludedWorkers.clear();
+    messagesHelper.close();
     logger.warn("Shuffle client has been shutdown!");
   }
 
@@ -1899,10 +2093,12 @@ public class ShuffleClientImpl extends ShuffleClient {
   }
 
   @VisibleForTesting
+  @Override
   public TransportClientFactory getDataClientFactory() {
     return dataClientFactory;
   }
 
+  @Override
   public void excludeFailedFetchLocation(String hostAndFetchPort, Exception e) {
     if (pushReplicateEnabled
         && fetchExcludeWorkerOnFailureEnabled
