@@ -22,7 +22,6 @@ import java.io.InputStream;
 import java.util.*;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.TimeUnit;
-import java.util.concurrent.TimeoutException;
 import java.util.concurrent.atomic.LongAdder;
 
 import scala.Tuple2;
@@ -31,12 +30,15 @@ import com.github.luben.zstd.ZstdException;
 import com.google.common.util.concurrent.Uninterruptibles;
 import io.netty.buffer.ByteBuf;
 import net.jpountz.lz4.LZ4Exception;
+import org.apache.commons.lang3.tuple.Pair;
 import org.roaringbitmap.RoaringBitmap;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
+import org.apache.celeborn.client.ClientUtils;
 import org.apache.celeborn.client.ShuffleClient;
 import org.apache.celeborn.client.compress.Decompressor;
+import org.apache.celeborn.client.read.checkpoint.PartitionReaderCheckpointMetadata;
 import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.common.exception.CelebornIOException;
 import org.apache.celeborn.common.network.client.TransportClient;
@@ -46,6 +48,7 @@ import org.apache.celeborn.common.protocol.*;
 import org.apache.celeborn.common.unsafe.Platform;
 import org.apache.celeborn.common.util.ExceptionMaker;
 import org.apache.celeborn.common.util.Utils;
+import org.apache.celeborn.common.write.LocationPushFailedBatches;
 
 public abstract class CelebornInputStream extends InputStream {
   private static final Logger logger = LoggerFactory.getLogger(CelebornInputStream.class);
@@ -57,7 +60,10 @@ public abstract class CelebornInputStream extends InputStream {
       ArrayList<PartitionLocation> locations,
       ArrayList<PbStreamHandler> streamHandlers,
       int[] attempts,
+      Map<String, LocationPushFailedBatches> failedBatchSetMap,
+      Map<String, Pair<Integer, Integer>> chunksRange,
       int attemptNumber,
+      long taskId,
       int startMapIndex,
       int endMapIndex,
       ConcurrentHashMap<String, Long> fetchExcludedWorkers,
@@ -66,28 +72,62 @@ public abstract class CelebornInputStream extends InputStream {
       int shuffleId,
       int partitionId,
       ExceptionMaker exceptionMaker,
-      MetricsCallback metricsCallback)
+      MetricsCallback metricsCallback,
+      boolean needDecompress)
       throws IOException {
-    if (locations == null || locations.size() == 0) {
+    if (locations == null || locations.isEmpty()) {
       return emptyInputStream;
     } else {
-      return new CelebornInputStreamImpl(
-          conf,
-          clientFactory,
-          shuffleKey,
-          locations,
-          streamHandlers,
-          attempts,
-          attemptNumber,
-          startMapIndex,
-          endMapIndex,
-          fetchExcludedWorkers,
-          shuffleClient,
-          appShuffleId,
-          shuffleId,
-          partitionId,
-          exceptionMaker,
-          metricsCallback);
+      // if startMapIndex > endMapIndex, means partition is skew partition and read by Celeborn
+      // implementation.
+      // locations will split to sub-partitions with startMapIndex size.
+      boolean readSkewPartitionWithoutMapRange =
+          ClientUtils.readSkewPartitionWithoutMapRange(conf, startMapIndex, endMapIndex);
+      if (readSkewPartitionWithoutMapRange) {
+        return new CelebornInputStreamImpl(
+            conf,
+            clientFactory,
+            shuffleKey,
+            locations,
+            streamHandlers,
+            attempts,
+            failedBatchSetMap,
+            attemptNumber,
+            taskId,
+            chunksRange,
+            fetchExcludedWorkers,
+            shuffleClient,
+            appShuffleId,
+            shuffleId,
+            partitionId,
+            exceptionMaker,
+            true,
+            metricsCallback,
+            needDecompress);
+      } else {
+        return new CelebornInputStreamImpl(
+            conf,
+            clientFactory,
+            shuffleKey,
+            locations,
+            streamHandlers,
+            attempts,
+            failedBatchSetMap,
+            attemptNumber,
+            taskId,
+            startMapIndex,
+            endMapIndex,
+            /*partitionLocationToChunkRange = */ null,
+            fetchExcludedWorkers,
+            shuffleClient,
+            appShuffleId,
+            shuffleId,
+            partitionId,
+            exceptionMaker,
+            false,
+            metricsCallback,
+            needDecompress);
+      }
     }
   }
 
@@ -132,11 +172,15 @@ public abstract class CelebornInputStream extends InputStream {
     private ArrayList<PbStreamHandler> streamHandlers;
     private int[] attempts;
     private final int attemptNumber;
+    private final long taskId;
     private final int startMapIndex;
     private final int endMapIndex;
+    private final Map<String, Pair<Integer, Integer>> partitionLocationToChunkRange;
 
     private Map<Integer, Set<Integer>> batchesRead = new HashMap<>();
 
+    private final Map<String, LocationPushFailedBatches> failedBatches;
+
     private byte[] compressedBuf;
     private byte[] rawDataBuf;
     private Decompressor decompressor;
@@ -161,7 +205,7 @@ public abstract class CelebornInputStream extends InputStream {
     private final boolean enabledReadLocalShuffle;
     private final String localHostAddress;
 
-    private boolean shuffleCompressionEnabled;
+    private boolean shouldDecompress;
     private long fetchExcludedWorkerExpireTimeout;
     private ConcurrentHashMap<String, Long> fetchExcludedWorkers;
 
@@ -173,6 +217,8 @@ public abstract class CelebornInputStream extends InputStream {
     private ExceptionMaker exceptionMaker;
     private boolean closed = false;
 
+    private final boolean readSkewPartitionWithoutMapRange;
+
     CelebornInputStreamImpl(
         CelebornConf conf,
         TransportClientFactory clientFactory,
@@ -180,16 +226,66 @@ public abstract class CelebornInputStream extends InputStream {
         ArrayList<PartitionLocation> locations,
         ArrayList<PbStreamHandler> streamHandlers,
         int[] attempts,
+        Map<String, LocationPushFailedBatches> failedBatchSet,
         int attemptNumber,
-        int startMapIndex,
-        int endMapIndex,
+        long taskId,
+        Map<String, Pair<Integer, Integer>> partitionLocationToChunkRange,
         ConcurrentHashMap<String, Long> fetchExcludedWorkers,
         ShuffleClient shuffleClient,
         int appShuffleId,
         int shuffleId,
         int partitionId,
         ExceptionMaker exceptionMaker,
-        MetricsCallback metricsCallback)
+        boolean splitSkewPartitionWithoutMapRange,
+        MetricsCallback metricsCallback,
+        boolean needDecompress)
+        throws IOException {
+      this(
+          conf,
+          clientFactory,
+          shuffleKey,
+          locations,
+          streamHandlers,
+          attempts,
+          failedBatchSet,
+          attemptNumber,
+          taskId,
+          /*startMapIndex = */ -1,
+          /*endMapIndex = */ -1,
+          partitionLocationToChunkRange,
+          fetchExcludedWorkers,
+          shuffleClient,
+          appShuffleId,
+          shuffleId,
+          partitionId,
+          exceptionMaker,
+          splitSkewPartitionWithoutMapRange,
+          metricsCallback,
+          needDecompress);
+    }
+
+    CelebornInputStreamImpl(
+        CelebornConf conf,
+        TransportClientFactory clientFactory,
+        String shuffleKey,
+        ArrayList<PartitionLocation> locations,
+        ArrayList<PbStreamHandler> streamHandlers,
+        int[] attempts,
+        Map<String, LocationPushFailedBatches> failedBatchSet,
+        int attemptNumber,
+        long taskId,
+        int startMapIndex,
+        int endMapIndex,
+        Map<String, Pair<Integer, Integer>> partitionLocationToChunkRange,
+        ConcurrentHashMap<String, Long> fetchExcludedWorkers,
+        ShuffleClient shuffleClient,
+        int appShuffleId,
+        int shuffleId,
+        int partitionId,
+        ExceptionMaker exceptionMaker,
+        boolean readSkewPartitionWithoutMapRange,
+        MetricsCallback metricsCallback,
+        boolean needDecompress)
         throws IOException {
       this.conf = conf;
       this.clientFactory = clientFactory;
@@ -200,14 +296,18 @@ public abstract class CelebornInputStream extends InputStream {
       }
       this.attempts = attempts;
       this.attemptNumber = attemptNumber;
+      this.taskId = taskId;
       this.startMapIndex = startMapIndex;
       this.endMapIndex = endMapIndex;
+      this.partitionLocationToChunkRange = partitionLocationToChunkRange;
       this.rangeReadFilter = conf.shuffleRangeReadFilterEnabled();
       this.enabledReadLocalShuffle = conf.enableReadLocalShuffleFile();
       this.localHostAddress = Utils.localHostName(conf);
-      this.shuffleCompressionEnabled =
-          !conf.shuffleCompressionCodec().equals(CompressionCodec.NONE);
+      this.shouldDecompress =
+          !conf.shuffleCompressionCodec().equals(CompressionCodec.NONE) && needDecompress;
       this.fetchExcludedWorkerExpireTimeout = conf.clientFetchExcludedWorkerExpireTimeout();
+      this.failedBatches = failedBatchSet;
+      this.readSkewPartitionWithoutMapRange = readSkewPartitionWithoutMapRange;
       this.fetchExcludedWorkers = fetchExcludedWorkers;
 
       if (conf.clientPushReplicateEnabled()) {
@@ -256,7 +356,11 @@ public abstract class CelebornInputStream extends InputStream {
         return null;
       }
       PartitionLocation currentLocation = locations.get(fileIndex);
-      while (skipLocation(startMapIndex, endMapIndex, currentLocation)) {
+      // if pushShuffleFailureTrackingEnabled is true, should not skip location
+      while ((readSkewPartitionWithoutMapRange
+              && !partitionLocationToChunkRange.containsKey(currentLocation.getUniqueId()))
+          || (!readSkewPartitionWithoutMapRange
+              && skipLocation(startMapIndex, endMapIndex, currentLocation))) {
         skipCount.increment();
         fileIndex++;
         if (fileIndex == locationCount) {
@@ -318,25 +422,16 @@ public abstract class CelebornInputStream extends InputStream {
       }
     }
 
-    private boolean isCriticalCause(Exception e) {
-      boolean rpcTimeout =
-          e instanceof IOException
-              && e.getCause() != null
-              && e.getCause() instanceof TimeoutException;
-      boolean connectException =
-          e instanceof CelebornIOException
-              && e.getMessage() != null
-              && (e.getMessage().startsWith("Connecting to")
-                  || e.getMessage().startsWith("Failed to"));
-      boolean fetchChunkTimeout =
-          e instanceof CelebornIOException
-              && e.getCause() != null
-              && e.getCause() instanceof IOException;
-      return connectException || rpcTimeout || fetchChunkTimeout;
+    private PartitionReader createReaderWithRetry(
+        PartitionLocation location, PbStreamHandler pbStreamHandler) throws IOException {
+      return createReaderWithRetry(location, pbStreamHandler, Optional.empty());
     }
 
     private PartitionReader createReaderWithRetry(
-        PartitionLocation location, PbStreamHandler pbStreamHandler) throws IOException {
+        PartitionLocation location,
+        PbStreamHandler pbStreamHandler,
+        Optional<PartitionReaderCheckpointMetadata> checkpointMetadata)
+        throws IOException {
       Exception lastException = null;
       while (fetchChunkRetryCnt < fetchChunkMaxRetry) {
         try {
@@ -344,12 +439,19 @@ public abstract class CelebornInputStream extends InputStream {
           if (isExcluded(location)) {
             throw new CelebornIOException("Fetch data from excluded worker! " + location);
           }
-          return createReader(location, pbStreamHandler, fetchChunkRetryCnt, fetchChunkMaxRetry);
+          PartitionReader reader =
+              createReader(
+                  location,
+                  pbStreamHandler,
+                  fetchChunkRetryCnt,
+                  fetchChunkMaxRetry,
+                  checkpointMetadata);
+          return reader;
         } catch (Exception e) {
           lastException = e;
           shuffleClient.excludeFailedFetchLocation(location.hostAndFetchPort(), e);
           fetchChunkRetryCnt++;
-          if (location.hasPeer()) {
+          if (location.hasPeer() && !readSkewPartitionWithoutMapRange) {
             // fetchChunkRetryCnt % 2 == 0 means both replicas have been tried,
             // so sleep before next try.
             if (fetchChunkRetryCnt % 2 == 0) {
@@ -420,7 +522,7 @@ public abstract class CelebornInputStream extends InputStream {
                     + currentReader.getLocation(),
                 e);
           } else {
-            if (currentReader.getLocation().hasPeer()) {
+            if (currentReader.getLocation().hasPeer() && !readSkewPartitionWithoutMapRange) {
               logger.warn(
                   "Fetch chunk failed {}/{} times for location {}, change to peer",
                   fetchChunkRetryCnt,
@@ -432,6 +534,8 @@ public abstract class CelebornInputStream extends InputStream {
               if (fetchChunkRetryCnt % 2 == 0) {
                 Uninterruptibles.sleepUninterruptibly(retryWaitMs, TimeUnit.MILLISECONDS);
               }
+              // We must not use checkpoint for peer location since chunkIds don't always match
+              // across peers
               currentReader = createReaderWithRetry(currentReader.getLocation().getPeer(), null);
             } else {
               logger.warn(
@@ -441,7 +545,14 @@ public abstract class CelebornInputStream extends InputStream {
                   currentReader.getLocation(),
                   e);
               Uninterruptibles.sleepUninterruptibly(retryWaitMs, TimeUnit.MILLISECONDS);
-              currentReader = createReaderWithRetry(currentReader.getLocation(), null);
+              // When reading from the same host again, it is possible to skip already read data
+              // chunks,
+              // improving read performance during retries.
+              currentReader =
+                  createReaderWithRetry(
+                      currentReader.getLocation(),
+                      null,
+                      currentReader.getPartitionReaderCheckpointMetadata());
             }
           }
         }
@@ -453,10 +564,20 @@ public abstract class CelebornInputStream extends InputStream {
         PartitionLocation location,
         PbStreamHandler pbStreamHandler,
         int fetchChunkRetryCnt,
-        int fetchChunkMaxRetry)
+        int fetchChunkMaxRetry,
+        Optional<PartitionReaderCheckpointMetadata> checkpointMetadata)
         throws IOException, InterruptedException {
 
       StorageInfo storageInfo = location.getStorageInfo();
+
+      int startChunkIndex = -1;
+      int endChunkIndex = -1;
+      if (partitionLocationToChunkRange != null) {
+        Pair<Integer, Integer> chunkRange =
+            partitionLocationToChunkRange.get(location.getUniqueId());
+        startChunkIndex = chunkRange.getLeft();
+        endChunkIndex = chunkRange.getRight();
+      }
       switch (storageInfo.getType()) {
         case HDD:
         case SSD:
@@ -474,7 +595,9 @@ public abstract class CelebornInputStream extends InputStream {
                 clientFactory,
                 startMapIndex,
                 endMapIndex,
-                callback);
+                callback,
+                startChunkIndex,
+                endChunkIndex);
           } else {
             return new WorkerPartitionReader(
                 conf,
@@ -486,8 +609,13 @@ public abstract class CelebornInputStream extends InputStream {
                 endMapIndex,
                 fetchChunkRetryCnt,
                 fetchChunkMaxRetry,
-                callback);
+                callback,
+                startChunkIndex,
+                endChunkIndex,
+                checkpointMetadata);
           }
+        case S3:
+        case OSS:
         case HDFS:
           return new DfsPartitionReader(
               conf,
@@ -497,7 +625,10 @@ public abstract class CelebornInputStream extends InputStream {
               clientFactory,
               startMapIndex,
               endMapIndex,
-              callback);
+              callback,
+              startChunkIndex,
+              endChunkIndex,
+              checkpointMetadata);
         default:
           throw new CelebornIOException(
               String.format("Unknown storage info %s to read location %s", storageInfo, location));
@@ -612,7 +743,7 @@ public abstract class CelebornInputStream extends InputStream {
     private void init() {
       int bufferSize = conf.clientFetchBufferSize();
 
-      if (shuffleCompressionEnabled) {
+      if (shouldDecompress) {
         int headerLen = Decompressor.getCompressionHeaderLength(conf);
         bufferSize += headerLen;
         compressedBuf = new byte[bufferSize];
@@ -633,6 +764,7 @@ public abstract class CelebornInputStream extends InputStream {
           return false;
         }
 
+        LocationPushFailedBatches failedBatch = new LocationPushFailedBatches();
         boolean hasData = false;
         while (currentChunk.isReadable() || moveToNextChunk()) {
           currentChunk.readBytes(sizeBuf);
@@ -641,7 +773,7 @@ public abstract class CelebornInputStream extends InputStream {
           int batchId = Platform.getInt(sizeBuf, Platform.BYTE_ARRAY_OFFSET + 8);
           int size = Platform.getInt(sizeBuf, Platform.BYTE_ARRAY_OFFSET + 12);
 
-          if (shuffleCompressionEnabled) {
+          if (shouldDecompress) {
             if (size > compressedBuf.length) {
               compressedBuf = new byte[size];
             }
@@ -657,6 +789,20 @@ public abstract class CelebornInputStream extends InputStream {
 
           // de-duplicate
           if (attemptId == attempts[mapId]) {
+            if (readSkewPartitionWithoutMapRange) {
+              LocationPushFailedBatches locationPushFailedBatches =
+                  this.failedBatches.get(currentReader.getLocation().getUniqueId());
+              if (null != locationPushFailedBatches) {
+                if (locationPushFailedBatches.contains(mapId, attemptId, batchId)) {
+                  logger.warn(
+                      "Skip duplicated batch: mapId={}, attemptId={}, batchId={}",
+                      mapId,
+                      attemptId,
+                      batchId);
+                  continue;
+                }
+              }
+            }
             if (!batchesRead.containsKey(mapId)) {
               Set<Integer> batchSet = new HashSet<>();
               batchesRead.put(mapId, batchSet);
@@ -665,7 +811,7 @@ public abstract class CelebornInputStream extends InputStream {
             if (!batchSet.contains(batchId)) {
               batchSet.add(batchId);
               callback.incBytesRead(BATCH_HEADER_SIZE + size);
-              if (shuffleCompressionEnabled) {
+              if (shouldDecompress) {
                 // decompress data
                 int originalLength = decompressor.getOriginalLen(compressedBuf);
                 if (rawDataBuf.length < originalLength) {
@@ -679,6 +825,7 @@ public abstract class CelebornInputStream extends InputStream {
               hasData = true;
               break;
             } else {
+              callback.incDuplicateBytesRead(BATCH_HEADER_SIZE + size);
               logger.debug(
                   "Skip duplicated batch: mapId {}, attemptId {}, batchId {}.",
                   mapId,
@@ -704,9 +851,9 @@ public abstract class CelebornInputStream extends InputStream {
           ioe = new IOException(e);
         }
         if (exceptionMaker != null) {
-          if (shuffleClient.reportShuffleFetchFailure(appShuffleId, shuffleId)) {
+          if (shuffleClient.reportShuffleFetchFailure(appShuffleId, shuffleId, taskId)) {
             /*
-             * [[ExceptionMaker.makeException]], for spark applications with celeborn.client.spark.fetch.throwsFetchFailure enabled will result in creating
+             * [[ExceptionMaker.makeException]], for spark applications with celeborn.client.spark.stageRerun.enabled enabled will result in creating
              * a FetchFailedException; and that will make the TaskContext as failed with shuffle fetch issues - see SPARK-19276 for more.
              * Given this, Celeborn can wrap the FetchFailedException with our CelebornIOException
              */
