@@ -30,18 +30,20 @@ import scala.concurrent.duration.Duration
 
 import org.apache.celeborn.client.{ShuffleCommittedInfo, WorkerStatusTracker}
 import org.apache.celeborn.client.CommitManager.CommittedPartitionInfo
-import org.apache.celeborn.client.LifecycleManager.{ShuffleFailedWorkers, ShuffleFileGroups}
+import org.apache.celeborn.client.LifecycleManager.{ShuffleFailedWorkers, ShuffleFileGroups, ShufflePushFailedBatches}
 import org.apache.celeborn.common.CelebornConf
 import org.apache.celeborn.common.internal.Logging
 import org.apache.celeborn.common.meta.{ShufflePartitionLocationInfo, WorkerInfo}
+import org.apache.celeborn.common.network.protocol.SerdeVersion
 import org.apache.celeborn.common.protocol.{PartitionLocation, PartitionType}
 import org.apache.celeborn.common.protocol.message.ControlMessages.{CommitFiles, CommitFilesResponse}
 import org.apache.celeborn.common.protocol.message.StatusCode
-import org.apache.celeborn.common.rpc.{RpcCallContext, RpcEndpointRef}
-import org.apache.celeborn.common.util.{CollectionUtils, JavaUtils, ThreadUtils, Utils}
+import org.apache.celeborn.common.rpc.RpcCallContext
+import org.apache.celeborn.common.util.{CollectionUtils, JavaUtils, Utils}
 // Can Remove this if celeborn don't support scala211 in future
 import org.apache.celeborn.common.util.FunctionConverter._
 import org.apache.celeborn.common.util.ThreadUtils.awaitResult
+import org.apache.celeborn.common.write.LocationPushFailedBatches
 
 case class CommitFilesParam(
     worker: WorkerInfo,
@@ -74,6 +76,7 @@ abstract class CommitHandler(
   private val totalWritten = new LongAdder
   private val fileCount = new LongAdder
   protected val reducerFileGroupsMap = new ShuffleFileGroups
+  protected val shufflePushFailedBatches = new ShufflePushFailedBatches
 
   val ec = ExecutionContext.fromExecutor(sharedRpcPool)
 
@@ -82,6 +85,8 @@ abstract class CommitHandler(
 
   def getPartitionType(): PartitionType
 
+  def getShuffleFailedBatches(): ShufflePushFailedBatches = this.shufflePushFailedBatches
+
   def isStageEnd(shuffleId: Int): Boolean = false
 
   def isStageEndOrInProcess(shuffleId: Int): Boolean = false
@@ -174,10 +179,23 @@ abstract class CommitHandler(
    * partitions are complete by the time the method is called, as downstream tasks may start early before all tasks
    * are completed.So map partition may need refresh reducer file group if needed.
    */
-  def handleGetReducerFileGroup(context: RpcCallContext, shuffleId: Int): Unit
+  def handleGetReducerFileGroup(
+      context: RpcCallContext,
+      shuffleId: Int,
+      serdeVersion: SerdeVersion): Unit
+
+  /**
+   * Only Reduce partition mode supports get stage end.
+   */
+  def handleGetStageEnd(context: RpcCallContext, shuffleId: Int): Unit = {
+    throw new UnsupportedOperationException(
+      "Failed when do handleGetStageEnd Operation, MapPartition shuffleType don't " +
+        "support stage end")
+  }
 
   def removeExpiredShuffle(shuffleId: Int): Unit = {
     reducerFileGroupsMap.remove(shuffleId)
+    shufflePushFailedBatches.remove(shuffleId)
   }
 
   /**
@@ -197,12 +215,23 @@ abstract class CommitHandler(
       attemptId: Int,
       numMappers: Int,
       partitionId: Int,
+      pushFailedBatches: util.Map[String, LocationPushFailedBatches],
       recordWorkerFailure: ShuffleFailedWorkers => Unit): (Boolean, Boolean)
 
-  def registerShuffle(shuffleId: Int, numMappers: Int): Unit = {
+  def registerShuffle(
+      shuffleId: Int,
+      numMappers: Int,
+      isSegmentGranularityVisible: Boolean): Unit = {
+    // TODO: if isSegmentGranularityVisible is set to true, it is necessary to handle the pending
+    //  get partition request of downstream reduce task here, in scenarios which support
+    //  downstream task start early before the upstream task, e.g. flink hybrid shuffle.
     reducerFileGroupsMap.put(shuffleId, JavaUtils.newConcurrentHashMap())
   }
 
+  def isSegmentGranularityVisible(shuffleId: Int): Boolean = {
+    false
+  }
+
   def doParallelCommitFiles(
       shuffleId: Int,
       shuffleCommittedInfo: ShuffleCommittedInfo,
@@ -292,7 +321,7 @@ abstract class CommitHandler(
     val futureSeq = Future.sequence(outFutures)(cbf, ec)
     awaitResult(futureSeq, Duration.Inf)
 
-    val timeout = conf.rpcAskTimeout.duration.toMillis
+    val timeout = clientRpcCommitFilesAskTimeout.duration.toMillis
     var remainingTime = timeout * maxRetries
     val delta = 50
     while (remainingTime >= 0 && !futures.isEmpty) {
@@ -301,29 +330,38 @@ abstract class CommitHandler(
       while (iter.hasNext) {
         val status = iter.next()
         val worker = status.workerInfo
+        val shuffleKey = Utils.makeShuffleKey(appUniqueId, shuffleId)
         if (status.future.isCompleted) {
           status.future.value.get match {
             case scala.util.Success(res) =>
               res.status match {
                 case StatusCode.SUCCESS | StatusCode.PARTIAL_SUCCESS | StatusCode.SHUFFLE_NOT_REGISTERED | StatusCode.REQUEST_FAILED | StatusCode.WORKER_EXCLUDED | StatusCode.COMMIT_FILE_EXCEPTION =>
-                  logInfo(s"Request commitFiles return ${res.status} for " +
-                    s"${Utils.makeShuffleKey(appUniqueId, shuffleId)}")
-                  if (res.status != StatusCode.SUCCESS && res.status != StatusCode.WORKER_EXCLUDED) {
-                    commitFilesFailedWorkers.put(worker, (res.status, System.currentTimeMillis()))
+                  if (res.status == StatusCode.SUCCESS) {
+                    logDebug(s"Request commitFiles return ${res.status} for " +
+                      s"$shuffleKey from worker ${worker.readableAddress()}")
+                  } else {
+                    logWarning(s"Request commitFiles return ${res.status} for " +
+                      s"$shuffleKey from worker ${worker.readableAddress()}")
+                    if (res.status != StatusCode.WORKER_EXCLUDED) {
+                      commitFilesFailedWorkers.put(worker, (res.status, System.currentTimeMillis()))
+                    }
                   }
                   processResponse(res, worker)
                   iter.remove()
                 case StatusCode.COMMIT_FILES_MOCK_FAILURE =>
                   if (status.retriedTimes < maxRetries) {
                     logError(s"Request commitFiles return ${res.status} for " +
-                      s"${Utils.makeShuffleKey(appUniqueId, shuffleId)} for ${status.retriedTimes}/$maxRetries, will retry")
+                      s"$shuffleKey for ${status.retriedTimes}/$maxRetries, will retry")
                     retryCommitFiles(status, currentTime)
                   } else {
                     logError(
                       s"Request commitFiles return ${StatusCode.COMMIT_FILES_MOCK_FAILURE} for " +
-                        s"${Utils.makeShuffleKey(appUniqueId, shuffleId)} for ${status.retriedTimes}/$maxRetries, will not retry")
+                        s"$shuffleKey for ${status.retriedTimes}/$maxRetries, will not retry")
+                    commitFilesFailedWorkers.put(
+                      worker,
+                      (StatusCode.COMMIT_FILES_MOCK_FAILURE, System.currentTimeMillis()))
                     val res = createFailResponse(status)
-                    processResponse(res, status.workerInfo)
+                    processResponse(res, worker)
                     iter.remove()
                   }
                 case _ =>
@@ -331,18 +369,20 @@ abstract class CommitHandler(
               }
 
             case scala.util.Failure(e) =>
-              val worker = status.workerInfo
               if (status.retriedTimes < maxRetries) {
                 logError(
-                  s"Ask worker($worker) CommitFiles for $shuffleId failed" +
+                  s"Ask worker(${worker.readableAddress()}) CommitFiles for $shuffleKey failed" +
                     s" (attempt ${status.retriedTimes}/$maxRetries), will retry.",
                   e)
                 retryCommitFiles(status, currentTime)
               } else {
                 logError(
-                  s"Ask worker($worker) CommitFiles for $shuffleId failed" +
+                  s"Ask worker(${worker.readableAddress()}) CommitFiles for $shuffleKey failed" +
                     s" (attempt ${status.retriedTimes}/$maxRetries), will not retry.",
                   e)
+                commitFilesFailedWorkers.put(
+                  worker,
+                  (StatusCode.WORKER_UNRESPONSIVE, System.currentTimeMillis()))
                 val res = createFailResponse(status)
                 processResponse(res, status.workerInfo)
                 iter.remove()
@@ -351,13 +391,16 @@ abstract class CommitHandler(
         } else if (currentTime - status.startTime > timeout) {
           if (status.retriedTimes < maxRetries) {
             logError(
-              s"Ask worker($worker) CommitFiles for $shuffleId failed because of Timeout" +
+              s"Ask worker(${worker.readableAddress()}) CommitFiles for $shuffleKey failed because of Timeout" +
                 s" (attempt ${status.retriedTimes}/$maxRetries), will retry.")
             retryCommitFiles(status, currentTime)
           } else {
             logError(
-              s"Ask worker($worker) CommitFiles for $shuffleId failed because of Timeout" +
+              s"Ask worker(${worker.readableAddress()}) CommitFiles for $shuffleKey failed because of Timeout" +
                 s" (attempt ${status.retriedTimes}/$maxRetries), will not retry.")
+            commitFilesFailedWorkers.put(
+              worker,
+              (StatusCode.WORKER_UNRESPONSIVE, System.currentTimeMillis()))
           }
         }
       }
@@ -372,7 +415,7 @@ abstract class CommitHandler(
     while (iter.hasNext) {
       val status = iter.next()
       logError(
-        s"Ask worker(${status.workerInfo}) CommitFiles for $shuffleId timed out")
+        s"Ask worker(${status.workerInfo.readableAddress()}) CommitFiles for $shuffleId timed out")
       val res = createFailResponse(status)
       processResponse(res, status.workerInfo)
       iter.remove()
@@ -381,7 +424,7 @@ abstract class CommitHandler(
 
   def parallelCommitFiles(
       shuffleId: Int,
-      allocatedWorkers: util.Map[WorkerInfo, ShufflePartitionLocationInfo],
+      allocatedWorkers: util.Map[String, ShufflePartitionLocationInfo],
       partitionIdOpt: Option[Int] = None): CommitResult = {
     val shuffleCommittedInfo = committedPartitionInfo.get(shuffleId)
     val primaryPartMap = JavaUtils.newConcurrentHashMap[String, PartitionLocation]
@@ -396,7 +439,8 @@ abstract class CommitHandler(
     val workerPartitionLocations = allocatedWorkers.asScala.filter(!_._2.isEmpty)
 
     val params = new ArrayBuffer[CommitFilesParam](workerPartitionLocations.size)
-    workerPartitionLocations.foreach { case (worker, partitionLocationInfo) =>
+    workerPartitionLocations.foreach { case (_, partitionLocationInfo) =>
+      val worker = partitionLocationInfo.workerInfo
       val primaryParts =
         partitionLocationInfo.getPrimaryPartitions(partitionIdOpt)
       val replicaParts = partitionLocationInfo.getReplicaPartitions(partitionIdOpt)
@@ -463,7 +507,8 @@ abstract class CommitHandler(
       primaryPartitionUniqueIds: util.Iterator[String],
       replicaPartitionUniqueIds: util.Iterator[String],
       primaryPartMap: ConcurrentHashMap[String, PartitionLocation],
-      replicaPartMap: ConcurrentHashMap[String, PartitionLocation]): Unit = {
+      replicaPartMap: ConcurrentHashMap[String, PartitionLocation],
+      isSegmentGranularityVisible: Boolean = false): Unit = {
     val committedPartitions = new util.HashMap[String, PartitionLocation]
     primaryPartitionUniqueIds.asScala.foreach { id =>
       val partitionLocation = primaryPartMap.get(id)
@@ -488,6 +533,8 @@ abstract class CommitHandler(
       }
     }
 
+    // TODO: if support upstream task write and downstream task read simultaneously,
+    //  should record the partition locations information in upstream task start time, rather than end time.
     committedPartitions.values().asScala.foreach { partition =>
       val partitionLocations = reducerFileGroupsMap.get(shuffleId).computeIfAbsent(
         partition.getId,
