@@ -17,9 +17,10 @@
 
 package org.apache.celeborn.client;
 
-import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.*;
 import static org.mockito.ArgumentMatchers.any;
 import static org.mockito.ArgumentMatchers.anyLong;
+import static org.mockito.BDDMockito.willAnswer;
 import static org.mockito.Mockito.mock;
 import static org.mockito.Mockito.when;
 
@@ -47,6 +48,7 @@ import org.apache.celeborn.common.exception.CelebornIOException;
 import org.apache.celeborn.common.identity.UserIdentifier;
 import org.apache.celeborn.common.network.client.TransportClient;
 import org.apache.celeborn.common.network.client.TransportClientFactory;
+import org.apache.celeborn.common.network.protocol.SerdeVersion;
 import org.apache.celeborn.common.protocol.CompressionCodec;
 import org.apache.celeborn.common.protocol.PartitionLocation;
 import org.apache.celeborn.common.protocol.message.ControlMessages.GetReducerFileGroupResponse$;
@@ -127,6 +129,31 @@ public class ShuffleClientSuiteJ {
     }
   }
 
+  @Test
+  public void testPushDataAndInterrupted() throws IOException, InterruptedException {
+    CelebornConf conf = setupEnv(CompressionCodec.NONE, StatusCode.SUCCESS, true);
+    try {
+      shuffleClient.pushData(
+          TEST_SHUFFLE_ID,
+          TEST_ATTEMPT_ID,
+          TEST_ATTEMPT_ID,
+          TEST_REDUCRE_ID,
+          TEST_BUF1,
+          0,
+          TEST_BUF1.length,
+          1,
+          1);
+      Thread.sleep(10 * 1000); // waiting for interrupt
+      fail();
+    } catch (Exception e) {
+      if (e instanceof InterruptedException) {
+        assertTrue(true);
+      } else {
+        fail();
+      }
+    }
+  }
+
   @Test
   public void testMergeData() throws IOException, InterruptedException {
     for (CompressionCodec codec : CompressionCodec.values()) {
@@ -179,59 +206,31 @@ public class ShuffleClientSuiteJ {
 
   @Test
   public void testRegisterShuffleFailed() throws IOException, InterruptedException {
-    setupEnv(CompressionCodec.NONE, StatusCode.SLOT_NOT_AVAILABLE);
-    try {
-      shuffleClient.pushData(
-          TEST_SHUFFLE_ID,
-          TEST_ATTEMPT_ID,
-          TEST_ATTEMPT_ID,
-          TEST_REDUCRE_ID,
-          TEST_BUF1,
-          0,
-          TEST_BUF1.length,
-          1,
-          1);
-      assert false;
-    } catch (CelebornIOException e) {
-      assert e.getMessage()
-          .contains("Register shuffle failed for shuffle 1, reason: SLOT_NOT_AVAILABLE");
-    }
+    verifyRegisterShuffleFailure(StatusCode.SLOT_NOT_AVAILABLE);
+    verifyRegisterShuffleFailure(StatusCode.RESERVE_SLOTS_FAILED);
+    verifyRegisterShuffleFailure(StatusCode.REQUEST_FAILED);
+  }
 
-    setupEnv(CompressionCodec.NONE, StatusCode.RESERVE_SLOTS_FAILED);
-    try {
-      shuffleClient.pushData(
-          TEST_SHUFFLE_ID,
-          TEST_ATTEMPT_ID,
-          TEST_ATTEMPT_ID,
-          TEST_REDUCRE_ID,
-          TEST_BUF1,
-          0,
-          TEST_BUF1.length,
-          1,
-          1);
-      assert false;
-    } catch (CelebornIOException e) {
-      assert e.getMessage()
-          .contains("Register shuffle failed for shuffle 1, reason: RESERVE_SLOTS_FAILED");
-    }
-
-    setupEnv(CompressionCodec.NONE, StatusCode.REQUEST_FAILED);
-    try {
-      shuffleClient.pushData(
-          TEST_SHUFFLE_ID,
-          TEST_ATTEMPT_ID,
-          TEST_ATTEMPT_ID,
-          TEST_REDUCRE_ID,
-          TEST_BUF1,
-          0,
-          TEST_BUF1.length,
-          1,
-          1);
-      assert false;
-    } catch (CelebornIOException e) {
-      assert e.getMessage()
-          .contains("Register shuffle failed for shuffle 1, reason: REQUEST_FAILED");
-    }
+  private void verifyRegisterShuffleFailure(StatusCode statusCode)
+      throws IOException, InterruptedException {
+    setupEnv(CompressionCodec.NONE, statusCode);
+    CelebornIOException e =
+        assertThrows(
+            CelebornIOException.class,
+            () ->
+                shuffleClient.pushData(
+                    TEST_SHUFFLE_ID,
+                    TEST_ATTEMPT_ID,
+                    TEST_ATTEMPT_ID,
+                    TEST_REDUCRE_ID,
+                    TEST_BUF1,
+                    0,
+                    TEST_BUF1.length,
+                    1,
+                    1));
+    assertTrue(
+        e.getMessage()
+            .contains("Register shuffle failed for shuffle 1, reason: " + statusCode.name()));
   }
 
   private CelebornConf setupEnv(CompressionCodec codec) throws IOException, InterruptedException {
@@ -240,6 +239,12 @@ public class ShuffleClientSuiteJ {
 
   private CelebornConf setupEnv(CompressionCodec codec, StatusCode statusCode)
       throws IOException, InterruptedException {
+    return setupEnv(codec, statusCode, false);
+  }
+
+  private CelebornConf setupEnv(
+      CompressionCodec codec, StatusCode statusCode, boolean interruptWhenPushData)
+      throws IOException, InterruptedException {
     CelebornConf conf = new CelebornConf();
     conf.set(CelebornConf.SHUFFLE_COMPRESSION_CODEC().key(), codec.name());
     conf.set(CelebornConf.CLIENT_PUSH_RETRY_THREADS().key(), "1");
@@ -387,7 +392,16 @@ public class ShuffleClientSuiteJ {
           }
         };
 
-    when(client.pushData(any(), anyLong(), any())).thenAnswer(t -> mockedFuture);
+    if (interruptWhenPushData) {
+      willAnswer(
+              invocation -> {
+                throw new InterruptedException("test");
+              })
+          .given(client)
+          .pushData(any(), anyLong(), any());
+    } else {
+      when(client.pushData(any(), anyLong(), any())).thenAnswer(t -> mockedFuture);
+    }
     when(clientFactory.createClient(
             primaryLocation.getHost(), primaryLocation.getPushPort(), TEST_REDUCRE_ID))
         .thenAnswer(t -> client);
@@ -410,7 +424,13 @@ public class ShuffleClientSuiteJ {
             t -> {
               Thread.sleep(60 * 1000);
               return GetReducerFileGroupResponse$.MODULE$.apply(
-                  StatusCode.SUCCESS, locations, new int[0], Collections.emptySet());
+                  StatusCode.SUCCESS,
+                  locations,
+                  new int[0],
+                  Collections.emptySet(),
+                  Collections.emptyMap(),
+                  new byte[0],
+                  SerdeVersion.V1);
             });
 
     when(endpointRef.askSync(any(), any(), any(Integer.class), any(Long.class), any()))
@@ -418,7 +438,13 @@ public class ShuffleClientSuiteJ {
             t -> {
               Thread.sleep(60 * 1000);
               return GetReducerFileGroupResponse$.MODULE$.apply(
-                  StatusCode.SUCCESS, locations, new int[0], Collections.emptySet());
+                  StatusCode.SUCCESS,
+                  locations,
+                  new int[0],
+                  Collections.emptySet(),
+                  Collections.emptyMap(),
+                  new byte[0],
+                  SerdeVersion.V1);
             });
 
     shuffleClient =
@@ -457,14 +483,26 @@ public class ShuffleClientSuiteJ {
         .thenAnswer(
             t -> {
               return GetReducerFileGroupResponse$.MODULE$.apply(
-                  StatusCode.SHUFFLE_NOT_REGISTERED, locations, new int[0], Collections.emptySet());
+                  StatusCode.SHUFFLE_NOT_REGISTERED,
+                  locations,
+                  new int[0],
+                  Collections.emptySet(),
+                  Collections.emptyMap(),
+                  new byte[0],
+                  SerdeVersion.V1);
             });
 
     when(endpointRef.askSync(any(), any(), any(Integer.class), any(Long.class), any()))
         .thenAnswer(
             t -> {
               return GetReducerFileGroupResponse$.MODULE$.apply(
-                  StatusCode.SHUFFLE_NOT_REGISTERED, locations, new int[0], Collections.emptySet());
+                  StatusCode.SHUFFLE_NOT_REGISTERED,
+                  locations,
+                  new int[0],
+                  Collections.emptySet(),
+                  Collections.emptyMap(),
+                  new byte[0],
+                  SerdeVersion.V1);
             });
 
     shuffleClient =
@@ -481,14 +519,26 @@ public class ShuffleClientSuiteJ {
         .thenAnswer(
             t -> {
               return GetReducerFileGroupResponse$.MODULE$.apply(
-                  StatusCode.STAGE_END_TIME_OUT, locations, new int[0], Collections.emptySet());
+                  StatusCode.STAGE_END_TIME_OUT,
+                  locations,
+                  new int[0],
+                  Collections.emptySet(),
+                  Collections.emptyMap(),
+                  new byte[0],
+                  SerdeVersion.V1);
             });
 
     when(endpointRef.askSync(any(), any(), any(Integer.class), any(Long.class), any()))
         .thenAnswer(
             t -> {
               return GetReducerFileGroupResponse$.MODULE$.apply(
-                  StatusCode.STAGE_END_TIME_OUT, locations, new int[0], Collections.emptySet());
+                  StatusCode.STAGE_END_TIME_OUT,
+                  locations,
+                  new int[0],
+                  Collections.emptySet(),
+                  Collections.emptyMap(),
+                  new byte[0],
+                  SerdeVersion.V1);
             });
 
     shuffleClient =
@@ -505,14 +555,26 @@ public class ShuffleClientSuiteJ {
         .thenAnswer(
             t -> {
               return GetReducerFileGroupResponse$.MODULE$.apply(
-                  StatusCode.SHUFFLE_DATA_LOST, locations, new int[0], Collections.emptySet());
+                  StatusCode.SHUFFLE_DATA_LOST,
+                  locations,
+                  new int[0],
+                  Collections.emptySet(),
+                  Collections.emptyMap(),
+                  new byte[0],
+                  SerdeVersion.V1);
             });
 
     when(endpointRef.askSync(any(), any(), any(Integer.class), any(Long.class), any()))
         .thenAnswer(
             t -> {
               return GetReducerFileGroupResponse$.MODULE$.apply(
-                  StatusCode.SHUFFLE_DATA_LOST, locations, new int[0], Collections.emptySet());
+                  StatusCode.SHUFFLE_DATA_LOST,
+                  locations,
+                  new int[0],
+                  Collections.emptySet(),
+                  Collections.emptyMap(),
+                  new byte[0],
+                  SerdeVersion.V1);
             });
 
     shuffleClient =
