@@ -19,8 +19,9 @@ package org.apache.celeborn.common.identity
 
 import org.apache.celeborn.common.CelebornConf
 import org.apache.celeborn.common.internal.Logging
+import org.apache.celeborn.common.util.Utils
 
-abstract class IdentityProvider {
+abstract class IdentityProvider(conf: CelebornConf) {
   def provide(): UserIdentifier
 }
 
@@ -29,20 +30,8 @@ object IdentityProvider extends Logging {
   val DEFAULT_USERNAME = "default"
 
   def instantiate(conf: CelebornConf): IdentityProvider = {
-    val className = conf.quotaIdentityProviderClass
-    logDebug(s"Creating identity provider $className")
-    val clazz = Class.forName(
-      className,
-      true,
-      Thread.currentThread().getContextClassLoader).asInstanceOf[Class[IdentityProvider]]
-    try {
-      val ctor = clazz.getDeclaredConstructor()
-      logDebug("Using (String, String, Boolean) constructor")
-      ctor.newInstance()
-    } catch {
-      case e: NoSuchMethodException =>
-        logError(s"Falling to instantiate identity provider $className", e)
-        throw e
-    }
+    val className = conf.identityProviderClass
+    logDebug(s"Creating instance of $className")
+    Utils.instantiateClassWithCelebornConf[IdentityProvider](className, conf)
   }
 }
