@@ -40,6 +40,9 @@ case class NamedCounter(name: String, counter: Counter, labels: Map[String, Stri
 case class NamedGauge[T](name: String, gauge: Gauge[T], labels: Map[String, String])
   extends MetricLabels
 
+case class NamedMeter(name: String, meter: Meter, labels: Map[String, String])
+  extends MetricLabels
+
 case class NamedHistogram(name: String, histogram: Histogram, labels: Map[String, String])
   extends MetricLabels
 
@@ -59,11 +62,20 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
 
   val timerSupplier = new TimerSupplier(metricsSlidingWindowSize)
 
+  val histogramSupplier = new HistogramSupplier(metricsSlidingWindowSize)
+
   val metricsCleaner: ScheduledExecutorService =
     ThreadUtils.newDaemonSingleThreadScheduledExecutor("worker-metrics-cleaner")
 
   val roleLabel: (String, String) = "role" -> role
-  val staticLabels: Map[String, String] = conf.metricsExtraLabels + roleLabel
+  val instanceLabel: Map[String, String] = role match {
+    case Role.MASTER =>
+      Map("instance" -> s"${Utils.localHostName(conf)}:${conf.masterHttpPort}")
+    case Role.WORKER =>
+      Map("instance" -> s"${Utils.localHostName(conf)}:${conf.workerHttpPort}")
+    case _ => Map.empty
+  }
+  val staticLabels: Map[String, String] = conf.metricsExtraLabels + roleLabel ++ instanceLabel
   val staticLabelsString: String = MetricLabels.labelString(staticLabels)
 
   val applicationLabel = "applicationId"
@@ -80,6 +92,12 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
   protected val namedCounters: ConcurrentHashMap[String, NamedCounter] =
     JavaUtils.newConcurrentHashMap[String, NamedCounter]()
 
+  protected val namedMeters: ConcurrentHashMap[String, NamedMeter] =
+    JavaUtils.newConcurrentHashMap[String, NamedMeter]()
+
+  protected val namedHistogram: ConcurrentHashMap[String, NamedHistogram] =
+    JavaUtils.newConcurrentHashMap[String, NamedHistogram]()
+
   def addTimerMetrics(namedTimer: NamedTimer): Unit = {
     val timerMetricsString = getTimerMetrics(namedTimer)
     timerMetrics.add(timerMetricsString)
@@ -118,6 +136,33 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
     addGauge(name, Map.empty[String, String], gauge)
   }
 
+  def addMeter(
+      name: String,
+      labels: Map[String, String],
+      meter: Meter): Unit = {
+    namedMeters.putIfAbsent(
+      metricNameWithCustomizedLabels(name, labels),
+      NamedMeter(name, meter, labels ++ staticLabels))
+  }
+
+  def addMeter(
+      name: String,
+      labels: JMap[String, String],
+      meter: Meter): Unit = {
+    addMeter(name, labels.asScala.toMap, meter)
+  }
+
+  def addMeter(name: String, labels: Map[String, String] = Map.empty)(f: () => Long): Unit = {
+    addMeter(
+      name,
+      labels,
+      metricRegistry.meter(metricNameWithCustomizedLabels(name, labels), new MeterSupplier(f)))
+  }
+
+  def addMeter(name: String, meter: Meter): Unit = {
+    addMeter(name, Map.empty[String, String], meter)
+  }
+
   def addTimer(name: String): Unit = addTimer(name, Map.empty[String, String])
 
   def addTimer(name: String, labels: Map[String, String]): Unit = {
@@ -143,6 +188,20 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
       NamedCounter(name, metricRegistry.counter(metricNameWithLabel), labels ++ staticLabels))
   }
 
+  def addHistogram(name: String): Unit = {
+    addHistogram(name, Map.empty)
+  }
+
+  def addHistogram(name: String, labels: Map[String, String]): Unit = {
+    val metricNameWithLabel = metricNameWithCustomizedLabels(name, labels)
+    namedHistogram.putIfAbsent(
+      metricNameWithLabel,
+      NamedHistogram(
+        name,
+        metricRegistry.histogram(name, histogramSupplier),
+        labels ++ staticLabels))
+  }
+
   def counters(): List[NamedCounter] = {
     namedCounters.values().asScala.toList
   }
@@ -151,8 +210,12 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
     namedGauges.values().asScala.toList
   }
 
+  def meters(): List[NamedMeter] = {
+    namedMeters.values().asScala.toList
+  }
+
   def histograms(): List[NamedHistogram] = {
-    List.empty[NamedHistogram]
+    namedHistogram.values().asScala.toList
   }
 
   def timers(): List[NamedTimer] = {
@@ -250,15 +313,20 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
   protected def doStopTimer(metricsName: String, key: String, labels: Map[String, String]): Unit = {
     val metricNameWithLabel = metricNameWithCustomizedLabels(metricsName, labels)
     try {
-      val (namedTimer, map) = namedTimers.get(metricNameWithLabel)
-      val startTime = Option(map.remove(key))
-      startTime match {
-        case Some(t) =>
-          namedTimer.timer.update(System.nanoTime() - t, TimeUnit.NANOSECONDS)
-          if (namedTimer.timer.getCount % metricsSlidingWindowSize == 0) {
-            addTimerMetrics(namedTimer)
-          }
-        case None =>
+      val pair = namedTimers.get(metricNameWithLabel)
+      if (pair != null) {
+        val (namedTimer, map) = pair
+        val startTime = Option(map.remove(key))
+        startTime match {
+          case Some(t) =>
+            namedTimer.timer.update(System.nanoTime() - t, TimeUnit.NANOSECONDS)
+            if (namedTimer.timer.getCount % metricsSlidingWindowSize == 0) {
+              addTimerMetrics(namedTimer)
+            }
+          case None =>
+        }
+      } else {
+        logWarning(s"Metric $metricNameWithLabel not found!")
       }
     } catch {
       case e: Exception =>
@@ -266,6 +334,19 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
     }
   }
 
+  def updateTimer(name: String, value: Long): Unit = {
+    updateTimer(name, value, Map.empty[String, String])
+  }
+
+  def updateTimer(metricsName: String, value: Long, labels: Map[String, String]): Unit = {
+    val metricNameWithLabel = metricNameWithCustomizedLabels(metricsName, labels)
+    if (!namedTimers.containsKey(metricNameWithLabel)) {
+      addTimer(metricsName, labels)
+    }
+    val (namedTimer, _) = namedTimers.get(metricNameWithLabel)
+    namedTimer.timer.update(value, TimeUnit.NANOSECONDS)
+  }
+
   def incCounter(metricsName: String): Unit = {
     incCounter(metricsName, 1)
   }
@@ -284,6 +365,20 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
     }
   }
 
+  def updateHistogram(name: String, value: Long): Unit = {
+    updateHistogram(name, Map.empty, value)
+  }
+
+  def updateHistogram(name: String, labels: Map[String, String], value: Long): Unit = {
+    val metricNameWithLabel = metricNameWithCustomizedLabels(name, labels)
+    val histogram = namedHistogram.get(metricNameWithLabel)
+    if (histogram != null) {
+      histogram.histogram.update(value)
+    } else {
+      logWarning(s"Metric $metricNameWithLabel not found!")
+    }
+  }
+
   private def clearOldValues(map: ConcurrentHashMap[String, Long]): Unit = {
     if (map.size > 5000) {
       // remove values has existed more than 15 min
@@ -308,18 +403,80 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
     metricsCleaner.scheduleWithFixedDelay(cleanTask, 10, 10, TimeUnit.MINUTES)
   }
 
+  private def addMetricsWithPrometheusHelpType(
+      metricName: String,
+      label: String,
+      value: Any,
+      timestamp: Long,
+      promType: String): String = {
+    val sb = new StringBuilder
+    sb.append(s"# HELP ${metricName}\n")
+    sb.append(s"# TYPE ${metricName} ${promType}\n")
+    sb.append(s"${metricName}$label ${value} $timestamp\n")
+    sb.toString()
+  }
+
   def getCounterMetrics(nc: NamedCounter): String = {
     val timestamp = System.currentTimeMillis
+    val sb = new StringBuilder
     val label = nc.labelString
-    val str = s"${normalizeKey(nc.name)}Count$label ${nc.counter.getCount} $timestamp\n"
-    str
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${normalizeKey(nc.name)}Count",
+      label,
+      nc.counter.getCount,
+      timestamp,
+      "counter"))
+    sb.toString()
   }
 
   def getGaugeMetrics(ng: NamedGauge[_]): String = {
     val timestamp = System.currentTimeMillis
     val sb = new StringBuilder
     val label = ng.labelString
-    sb.append(s"${normalizeKey(ng.name)}Value$label ${ng.gauge.getValue} $timestamp\n")
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${normalizeKey(ng.name)}Value",
+      label,
+      ng.gauge.getValue,
+      timestamp,
+      "gauge"))
+    sb.toString()
+  }
+
+  def getMeterMetrics(nm: NamedMeter): String = {
+    val timestamp = System.currentTimeMillis
+    val sb = new StringBuilder
+    val prefix = normalizeKey(nm.name)
+    val label = nm.labelString
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Count",
+      label,
+      nm.meter.getCount,
+      timestamp,
+      "counter"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}MeanRate",
+      label,
+      nm.meter.getMeanRate,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}OneMinuteRate",
+      label,
+      nm.meter.getOneMinuteRate,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}FiveMinuteRate",
+      label,
+      nm.meter.getFiveMinuteRate,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}FifteenMinuteRate",
+      label,
+      nm.meter.getFifteenMinuteRate,
+      timestamp,
+      "gauge"))
     sb.toString()
   }
 
@@ -329,22 +486,66 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
     val snapshot = nh.histogram.getSnapshot
     val prefix = normalizeKey(nh.name)
     val label = nh.labelString
-    sb.append(s"${prefix}Count$label ${nh.histogram.getCount} $timestamp\n")
-    sb.append(s"${prefix}Max$label ${reportNanosAsMills(snapshot.getMax)} $timestamp\n")
-    sb.append(s"${prefix}Mean$label ${reportNanosAsMills(snapshot.getMean)} $timestamp\n")
-    sb.append(s"${prefix}Min$label ${reportNanosAsMills(snapshot.getMin)} $timestamp\n")
-    sb.append(s"${prefix}50thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.getMedian)} $timestamp\n")
-    sb.append(s"${prefix}75thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get75thPercentile)} $timestamp\n")
-    sb.append(s"${prefix}95thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get95thPercentile)} $timestamp\n")
-    sb.append(s"${prefix}98thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get98thPercentile)} $timestamp\n")
-    sb.append(s"${prefix}99thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get99thPercentile)} $timestamp\n")
-    sb.append(s"${prefix}999thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get999thPercentile)} $timestamp\n")
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Count",
+      label,
+      nh.histogram.getCount,
+      timestamp,
+      "counter"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Max",
+      label,
+      snapshot.getMax,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Mean",
+      label,
+      snapshot.getMean,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Min",
+      label,
+      snapshot.getMin,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}50thPercentile",
+      label,
+      snapshot.getMedian,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}75thPercentile",
+      label,
+      snapshot.get75thPercentile,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}95thPercentile",
+      label,
+      snapshot.get95thPercentile,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}98thPercentile",
+      label,
+      snapshot.get98thPercentile,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}99thPercentile",
+      label,
+      snapshot.get99thPercentile,
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}999thPercentile",
+      label,
+      snapshot.get999thPercentile,
+      timestamp,
+      "gauge"))
     sb.toString()
   }
 
@@ -354,39 +555,87 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
     val snapshot = nt.timer.getSnapshot
     val prefix = normalizeKey(nt.name)
     val label = nt.labelString
-    sb.append(s"${prefix}Count$label ${nt.timer.getCount} $timestamp\n")
-    sb.append(s"${prefix}Max$label ${reportNanosAsMills(snapshot.getMax)} $timestamp\n")
-    sb.append(s"${prefix}Mean$label ${reportNanosAsMills(snapshot.getMean)} $timestamp\n")
-    sb.append(s"${prefix}Min$label ${reportNanosAsMills(snapshot.getMin)} $timestamp\n")
-    sb.append(s"${prefix}50thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.getMedian)} $timestamp\n")
-    sb.append(s"${prefix}75thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get75thPercentile)} $timestamp\n")
-    sb.append(s"${prefix}95thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get95thPercentile)} $timestamp\n")
-    sb.append(s"${prefix}98thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get98thPercentile)} $timestamp\n")
-    sb.append(s"${prefix}99thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get99thPercentile)} $timestamp\n")
-    sb.append(s"${prefix}999thPercentile$label" +
-      s" ${reportNanosAsMills(snapshot.get999thPercentile)} $timestamp\n")
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Count",
+      label,
+      nt.timer.getCount,
+      timestamp,
+      "counter"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Max",
+      label,
+      reportNanosAsMills(snapshot.getMax),
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Mean",
+      label,
+      reportNanosAsMills(snapshot.getMean),
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}Min",
+      label,
+      reportNanosAsMills(snapshot.getMin),
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}50thPercentile",
+      label,
+      reportNanosAsMills(snapshot.getMedian),
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}75thPercentile",
+      label,
+      reportNanosAsMills(snapshot.get75thPercentile),
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}95thPercentile",
+      label,
+      reportNanosAsMills(snapshot.get95thPercentile),
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}98thPercentile",
+      label,
+      reportNanosAsMills(snapshot.get98thPercentile),
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}99thPercentile",
+      label,
+      reportNanosAsMills(snapshot.get99thPercentile),
+      timestamp,
+      "gauge"))
+    sb.append(addMetricsWithPrometheusHelpType(
+      s"${prefix}999thPercentile",
+      label,
+      reportNanosAsMills(snapshot.get999thPercentile),
+      timestamp,
+      "gauge"))
     sb.toString()
   }
 
   def getAllMetricsNum: Int = {
     val sum = timerMetrics.size() +
       namedTimers.size() +
+      namedMeters.size() +
       namedGauges.size() +
       namedCounters.size()
     sum
   }
 
-  override def getMetrics(): String = {
+  // Do use this method to get metrics, because it will clear the timeMetrics queue
+  // Do not use the LogReporter to report metrics
+  override def getMetrics: String = {
     var leftMetricsNum = metricsCapacity
     val sb = new mutable.StringBuilder
     leftMetricsNum = fillInnerMetricsSnapshot(getAndClearTimerMetrics(), leftMetricsNum, sb)
     leftMetricsNum = fillInnerMetricsSnapshot(timers(), leftMetricsNum, sb)
     leftMetricsNum = fillInnerMetricsSnapshot(histograms(), leftMetricsNum, sb)
+    leftMetricsNum = fillInnerMetricsSnapshot(meters(), leftMetricsNum, sb)
     leftMetricsNum = fillInnerMetricsSnapshot(gauges(), leftMetricsNum, sb)
     leftMetricsNum = fillInnerMetricsSnapshot(counters(), leftMetricsNum, sb)
     if (leftMetricsNum <= 0) {
@@ -409,10 +658,10 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
         sb.append(getCounterMetrics(c))
       case g: NamedGauge[_] =>
         sb.append(getGaugeMetrics(g))
+      case m: NamedMeter =>
+        sb.append(getMeterMetrics(m))
       case h: NamedHistogram =>
         sb.append(getHistogramMetrics(h))
-        h.asInstanceOf[CelebornHistogram].reservoir
-          .asInstanceOf[ResettableSlidingWindowReservoir].reset()
       case t: NamedTimer =>
         sb.append(getTimerMetrics(t))
         t.timer.asInstanceOf[CelebornTimer].reservoir
@@ -427,8 +676,10 @@ abstract class AbstractSource(conf: CelebornConf, role: String)
     metricsCleaner.shutdown()
     namedCounters.clear()
     namedGauges.clear()
+    namedMeters.clear()
     namedTimers.clear()
     timerMetrics.clear()
+    namedHistogram.clear()
     metricRegistry.removeMatching(new MetricFilter {
       override def matches(s: String, metric: Metric): Boolean = true
     })
@@ -463,3 +714,14 @@ class TimerSupplier(val slidingWindowSize: Int)
 class GaugeSupplier[T](f: () => T) extends MetricRegistry.MetricSupplier[Gauge[_]] {
   override def newMetric(): Gauge[T] = new Gauge[T] { override def getValue: T = f() }
 }
+
+class MeterSupplier(f: () => Long) extends MetricRegistry.MetricSupplier[Meter] {
+  override def newMetric(): Meter = new Meter { override def getCount: Long = f() }
+}
+
+class HistogramSupplier(val slidingWindowSize: Int)
+  extends MetricRegistry.MetricSupplier[Histogram] {
+  override def newMetric(): Histogram = {
+    new CelebornHistogram(new ResettableSlidingWindowReservoir(slidingWindowSize))
+  }
+}
