@@ -22,16 +22,18 @@ import java.util.{Collections, UUID}
 
 import scala.collection.JavaConverters._
 
+import com.google.protobuf.ByteString
 import org.roaringbitmap.RoaringBitmap
 
 import org.apache.celeborn.common.identity.UserIdentifier
 import org.apache.celeborn.common.internal.Logging
 import org.apache.celeborn.common.meta.{DiskInfo, WorkerInfo, WorkerStatus}
-import org.apache.celeborn.common.network.protocol.TransportMessage
+import org.apache.celeborn.common.network.protocol.{SerdeVersion, TransportMessage}
 import org.apache.celeborn.common.protocol._
 import org.apache.celeborn.common.protocol.MessageType._
 import org.apache.celeborn.common.quota.ResourceConsumption
 import org.apache.celeborn.common.util.{PbSerDeUtils, Utils}
+import org.apache.celeborn.common.write.LocationPushFailedBatches
 
 sealed trait Message extends Serializable
 
@@ -68,7 +70,7 @@ object ControlMessages extends Logging {
 
   case object CheckForWorkerUnavailableInfoTimeout extends Message
 
-  case object CheckForHDFSExpiredDirsTimeout extends Message
+  case object CheckForDFSExpiredDirsTimeout extends Message
 
   case object RemoveExpiredShuffle extends Message
 
@@ -118,7 +120,6 @@ object ControlMessages extends Logging {
       disks: Seq[DiskInfo],
       userResourceConsumption: util.Map[UserIdentifier, ResourceConsumption],
       activeShuffleKeys: util.Set[String],
-      estimatedAppDiskUsage: util.HashMap[String, java.lang.Long],
       highWorkload: Boolean,
       workerStatus: WorkerStatus,
       override var requestId: String = ZERO_UUID) extends MasterRequestMessage
@@ -147,13 +148,15 @@ object ControlMessages extends Logging {
         numMappers: Int,
         mapId: Int,
         attemptId: Int,
-        partitionId: Int): PbRegisterMapPartitionTask =
+        partitionId: Int,
+        isSegmentGranularityVisible: Boolean): PbRegisterMapPartitionTask =
       PbRegisterMapPartitionTask.newBuilder()
         .setShuffleId(shuffleId)
         .setNumMappers(numMappers)
         .setMapId(mapId)
         .setAttemptId(attemptId)
         .setPartitionId(partitionId)
+        .setIsSegmentGranularityVisible(isSegmentGranularityVisible)
         .build()
   }
 
@@ -181,6 +184,7 @@ object ControlMessages extends Logging {
       availableStorageTypes: Int,
       excludedWorkerSet: Set[WorkerInfo] = Set.empty,
       packed: Boolean = false,
+      tagsExpr: String = "",
       override var requestId: String = ZERO_UUID)
     extends MasterRequestMessage
 
@@ -269,20 +273,29 @@ object ControlMessages extends Logging {
       mapId: Int,
       attemptId: Int,
       numMappers: Int,
-      partitionId: Int)
+      partitionId: Int,
+      failedBatchSet: util.Map[String, LocationPushFailedBatches])
     extends MasterMessage
 
   case class MapperEndResponse(status: StatusCode) extends MasterMessage
 
-  case class GetReducerFileGroup(shuffleId: Int) extends MasterMessage
+  case class GetReducerFileGroup(
+      shuffleId: Int,
+      isSegmentGranularityVisible: Boolean,
+      serdeVersion: SerdeVersion)
+    extends MasterMessage
 
   // util.Set[String] -> util.Set[Path.toString]
   // Path can't be serialized
   case class GetReducerFileGroupResponse(
       status: StatusCode,
-      fileGroup: util.Map[Integer, util.Set[PartitionLocation]],
-      attempts: Array[Int],
-      partitionIds: util.Set[Integer] = Collections.emptySet[Integer]())
+      fileGroup: util.Map[Integer, util.Set[PartitionLocation]] = Collections.emptyMap(),
+      attempts: Array[Int] = Array.emptyIntArray,
+      partitionIds: util.Set[Integer] = Collections.emptySet[Integer](),
+      pushFailedBatches: util.Map[String, LocationPushFailedBatches] =
+        Collections.emptyMap(),
+      broadcast: Array[Byte] = Array.emptyByteArray,
+      serdeVersion: SerdeVersion = SerdeVersion.V1)
     extends MasterMessage
 
   object WorkerExclude {
@@ -331,6 +344,26 @@ object ControlMessages extends Logging {
         .build()
   }
 
+  object ReviseLostShuffles {
+    def apply(
+        appId: String,
+        lostShuffles: java.util.List[Integer],
+        requestId: String): PbReviseLostShuffles =
+      PbReviseLostShuffles.newBuilder()
+        .setAppId(appId)
+        .addAllLostShuffles(lostShuffles)
+        .setRequestId(requestId)
+        .build()
+  }
+
+  object ReviseLostShufflesResponse {
+    def apply(success: Boolean, message: String): PbReviseLostShufflesResponse =
+      PbReviseLostShufflesResponse.newBuilder()
+        .setSuccess(success)
+        .setMessage(message)
+        .build()
+  }
+
   case class StageEnd(shuffleId: Int) extends MasterMessage
 
   case class StageEndResponse(status: StatusCode)
@@ -348,6 +381,18 @@ object ControlMessages extends Logging {
         .build()
   }
 
+  object BatchUnregisterShuffles {
+    def apply(
+        appId: String,
+        shuffleIds: util.List[Integer],
+        requestId: String): PbBatchUnregisterShuffles =
+      PbBatchUnregisterShuffles.newBuilder()
+        .setAppId(appId)
+        .addAllShuffleIds(shuffleIds)
+        .setRequestId(requestId)
+        .build()
+  }
+
   object UnregisterShuffleResponse {
     def apply(status: StatusCode): PbUnregisterShuffleResponse =
       PbUnregisterShuffleResponse.newBuilder()
@@ -355,6 +400,17 @@ object ControlMessages extends Logging {
         .build()
   }
 
+  object BatchUnregisterShuffleResponse {
+    def apply(
+        status: StatusCode,
+        shuffleIds: util.List[Integer] = Collections.emptyList())
+        : PbBatchUnregisterShuffleResponse =
+      PbBatchUnregisterShuffleResponse.newBuilder()
+        .setStatus(status.getValue)
+        .addAllShuffleIds(shuffleIds)
+        .build()
+  }
+
   case class ApplicationLost(
       appId: String,
       override var requestId: String = ZERO_UUID) extends MasterRequestMessage
@@ -365,6 +421,10 @@ object ControlMessages extends Logging {
       appId: String,
       totalWritten: Long,
       fileCount: Long,
+      shuffleCount: Long,
+      applicationCount: Long,
+      shuffleFallbackCounts: util.Map[String, java.lang.Long],
+      applicationFallbackCounts: util.Map[String, java.lang.Long],
       needCheckedWorkerList: util.List[WorkerInfo],
       override var requestId: String = ZERO_UUID,
       shouldResponse: Boolean = false) extends MasterRequestMessage
@@ -373,7 +433,9 @@ object ControlMessages extends Logging {
       statusCode: StatusCode,
       excludedWorkers: util.List[WorkerInfo],
       unknownWorkers: util.List[WorkerInfo],
-      shuttingWorkers: util.List[WorkerInfo]) extends Message
+      shuttingWorkers: util.List[WorkerInfo],
+      registeredShuffles: util.List[Integer],
+      checkQuotaResponse: CheckQuotaResponse) extends Message
 
   case class CheckQuota(userIdentifier: UserIdentifier) extends Message
 
@@ -383,6 +445,10 @@ object ControlMessages extends Logging {
       unavailable: util.List[WorkerInfo],
       override var requestId: String = ZERO_UUID) extends MasterRequestMessage
 
+  case class ReportWorkerDecommission(
+      unavailable: util.List[WorkerInfo],
+      override var requestId: String = ZERO_UUID) extends MasterRequestMessage
+
   object CheckWorkersAvailable {
     def apply(): PbCheckWorkersAvailable = {
       PbCheckWorkersAvailable.newBuilder().build()
@@ -408,14 +474,21 @@ object ControlMessages extends Logging {
         .build()
   }
 
+  object RemoveWorkersUnavailableInfoResponse {
+    def apply(success: Boolean): PbRemoveWorkersUnavailableInfoResponse =
+      PbRemoveWorkersUnavailableInfoResponse.newBuilder()
+        .setSuccess(success)
+        .build()
+  }
+
   object WorkerEventRequest {
     def apply(
         workers: util.List[WorkerInfo],
-        eventType: String,
+        eventType: WorkerEventType,
         requestId: String): PbWorkerEventRequest =
       PbWorkerEventRequest.newBuilder()
         .setRequestId(requestId)
-        .setWorkerEventType(WorkerEventType.valueOf(eventType))
+        .setWorkerEventType(eventType)
         .addAllWorkers(workers.asScala.map { workerInfo =>
           PbSerDeUtils.toPbWorkerInfo(workerInfo, true, false)
         }.toList.asJava)
@@ -446,7 +519,8 @@ object ControlMessages extends Logging {
       rangeReadFilter: Boolean,
       userIdentifier: UserIdentifier,
       pushDataTimeout: Long,
-      partitionSplitEnabled: Boolean = false)
+      partitionSplitEnabled: Boolean = false,
+      isSegmentGranularityVisible: Boolean = false)
     extends WorkerMessage
 
   case class ReserveSlotsResponse(
@@ -505,8 +579,8 @@ object ControlMessages extends Logging {
     case CheckForApplicationTimeOut =>
       new TransportMessage(MessageType.CHECK_APPLICATION_TIMEOUT, null)
 
-    case CheckForHDFSExpiredDirsTimeout =>
-      new TransportMessage(MessageType.CHECK_FOR_HDFS_EXPIRED_DIRS_TIMEOUT, null)
+    case CheckForDFSExpiredDirsTimeout =>
+      new TransportMessage(MessageType.CHECK_FOR_DFS_EXPIRED_DIRS_TIMEOUT, null)
 
     case RemoveExpiredShuffle =>
       new TransportMessage(MessageType.REMOVE_EXPIRED_SHUFFLE, null)
@@ -526,6 +600,12 @@ object ControlMessages extends Logging {
     case pb: PbReportShuffleFetchFailureResponse =>
       new TransportMessage(MessageType.REPORT_SHUFFLE_FETCH_FAILURE_RESPONSE, pb.toByteArray)
 
+    case pb: PbReviseLostShuffles =>
+      new TransportMessage(MessageType.REVISE_LOST_SHUFFLES, pb.toByteArray)
+
+    case pb: PbReviseLostShufflesResponse =>
+      new TransportMessage(MessageType.REVISE_LOST_SHUFFLES_RESPONSE, pb.toByteArray)
+
     case pb: PbReportBarrierStageAttemptFailure =>
       new TransportMessage(MessageType.REPORT_BARRIER_STAGE_ATTEMPT_FAILURE, pb.toByteArray)
 
@@ -534,6 +614,15 @@ object ControlMessages extends Logging {
         MessageType.REPORT_BARRIER_STAGE_ATTEMPT_FAILURE_RESPONSE,
         pb.toByteArray)
 
+    case pb: PbPushMergedDataSplitPartitionInfo =>
+      new TransportMessage(MessageType.PUSH_MERGED_DATA_SPLIT_PARTITION_INFO, pb.toByteArray)
+
+    case pb: PbGetStageEnd =>
+      new TransportMessage(MessageType.GET_STAGE_END, pb.toByteArray)
+
+    case pb: PbGetStageEndResponse =>
+      new TransportMessage(MessageType.GET_STAGE_END_RESPONSE, pb.toByteArray)
+
     case HeartbeatFromWorker(
           host,
           rpcPort,
@@ -543,7 +632,6 @@ object ControlMessages extends Logging {
           disks,
           userResourceConsumption,
           activeShuffleKeys,
-          estimatedAppDiskUsage,
           highWorkload,
           workerStatus,
           requestId) =>
@@ -559,7 +647,6 @@ object ControlMessages extends Logging {
         .putAllUserResourceConsumption(pbUserResourceConsumption)
         .setReplicatePort(replicatePort)
         .addAllActiveShuffleKeys(activeShuffleKeys)
-        .putAllEstimatedAppDiskUsage(estimatedAppDiskUsage)
         .setHighWorkload(highWorkload)
         .setWorkerStatus(PbSerDeUtils.toPbWorkerStatus(workerStatus))
         .setRequestId(requestId)
@@ -595,6 +682,7 @@ object ControlMessages extends Logging {
           availableStorageTypes,
           excludedWorkerSet,
           packed,
+          tagsExpr,
           requestId) =>
       val payload = PbRequestSlots.newBuilder()
         .setApplicationId(applicationId)
@@ -610,6 +698,7 @@ object ControlMessages extends Logging {
         .addAllExcludedWorkerSet(excludedWorkerSet.map(
           PbSerDeUtils.toPbWorkerInfo(_, true, true)).asJava)
         .setPacked(packed)
+        .setTagsExpr(tagsExpr)
         .build().toByteArray
       new TransportMessage(MessageType.REQUEST_SLOTS, payload)
 
@@ -650,13 +739,18 @@ object ControlMessages extends Logging {
     case pb: PbChangeLocationResponse =>
       new TransportMessage(MessageType.CHANGE_LOCATION_RESPONSE, pb.toByteArray)
 
-    case MapperEnd(shuffleId, mapId, attemptId, numMappers, partitionId) =>
+    case MapperEnd(shuffleId, mapId, attemptId, numMappers, partitionId, pushFailedBatch) =>
+      val pushFailedMap = pushFailedBatch.asScala.map { case (k, v) =>
+        val resultValue = PbSerDeUtils.toPbLocationPushFailedBatches(v)
+        (k, resultValue)
+      }.toMap.asJava
       val payload = PbMapperEnd.newBuilder()
         .setShuffleId(shuffleId)
         .setMapId(mapId)
         .setAttemptId(attemptId)
         .setNumMappers(numMappers)
         .setPartitionId(partitionId)
+        .putAllPushFailureBatches(pushFailedMap)
         .build().toByteArray
       new TransportMessage(MessageType.MAPPER_END, payload)
 
@@ -666,13 +760,21 @@ object ControlMessages extends Logging {
         .build().toByteArray
       new TransportMessage(MessageType.MAPPER_END_RESPONSE, payload)
 
-    case GetReducerFileGroup(shuffleId) =>
+    case GetReducerFileGroup(shuffleId, isSegmentGranularityVisible, serdeVersion) =>
       val payload = PbGetReducerFileGroup.newBuilder()
         .setShuffleId(shuffleId)
+        .setIsSegmentGranularityVisible(isSegmentGranularityVisible)
         .build().toByteArray
-      new TransportMessage(MessageType.GET_REDUCER_FILE_GROUP, payload)
+      new TransportMessage(MessageType.GET_REDUCER_FILE_GROUP, payload, serdeVersion)
 
-    case GetReducerFileGroupResponse(status, fileGroup, attempts, partitionIds) =>
+    case GetReducerFileGroupResponse(
+          status,
+          fileGroup,
+          attempts,
+          partitionIds,
+          failedBatches,
+          broadcast,
+          serdeVersion) =>
       val builder = PbGetReducerFileGroupResponse
         .newBuilder()
         .setStatus(status.getValue)
@@ -685,8 +787,14 @@ object ControlMessages extends Logging {
         }.asJava)
       builder.addAllAttempts(attempts.map(Integer.valueOf).toIterable.asJava)
       builder.addAllPartitionIds(partitionIds)
+      builder.putAllPushFailedBatches(
+        failedBatches.asScala.map {
+          case (uniqueId, pushFailedBatchSet) =>
+            (uniqueId, PbSerDeUtils.toPbLocationPushFailedBatches(pushFailedBatchSet))
+        }.asJava)
+      builder.setBroadcast(ByteString.copyFrom(broadcast))
       val payload = builder.build().toByteArray
-      new TransportMessage(MessageType.GET_REDUCER_FILE_GROUP_RESPONSE, payload)
+      new TransportMessage(MessageType.GET_REDUCER_FILE_GROUP_RESPONSE, payload, serdeVersion)
 
     case pb: PbWorkerExclude =>
       new TransportMessage(MessageType.WORKER_EXCLUDE, pb.toByteArray)
@@ -715,9 +823,15 @@ object ControlMessages extends Logging {
     case pb: PbUnregisterShuffle =>
       new TransportMessage(MessageType.UNREGISTER_SHUFFLE, pb.toByteArray)
 
+    case pb: PbBatchUnregisterShuffles =>
+      new TransportMessage(MessageType.BATCH_UNREGISTER_SHUFFLES, pb.toByteArray)
+
     case pb: PbUnregisterShuffleResponse =>
       new TransportMessage(MessageType.UNREGISTER_SHUFFLE_RESPONSE, pb.toByteArray)
 
+    case pb: PbBatchUnregisterShuffleResponse =>
+      new TransportMessage(MessageType.BATCH_UNREGISTER_SHUFFLE_RESPONSE, pb.toByteArray)
+
     case ApplicationLost(appId, requestId) =>
       val payload = PbApplicationLost.newBuilder()
         .setAppId(appId).setRequestId(requestId)
@@ -733,6 +847,10 @@ object ControlMessages extends Logging {
           appId,
           totalWritten,
           fileCount,
+          shuffleCount,
+          applicationCount,
+          shuffleFallbackCounts,
+          applicationFallbackCounts,
           needCheckedWorkerList,
           requestId,
           shouldResponse) =>
@@ -741,6 +859,10 @@ object ControlMessages extends Logging {
         .setRequestId(requestId)
         .setTotalWritten(totalWritten)
         .setFileCount(fileCount)
+        .setShuffleCount(shuffleCount)
+        .setApplicationCount(applicationCount)
+        .putAllShuffleFallbackCounts(shuffleFallbackCounts)
+        .putAllApplicationFallbackCounts(applicationFallbackCounts)
         .addAllNeedCheckedWorkerList(needCheckedWorkerList.asScala.map(
           PbSerDeUtils.toPbWorkerInfo(_, true, true)).toList.asJava)
         .setShouldResponse(shouldResponse)
@@ -751,7 +873,11 @@ object ControlMessages extends Logging {
           statusCode,
           excludedWorkers,
           unknownWorkers,
-          shuttingWorkers) =>
+          shuttingWorkers,
+          registeredShuffles,
+          checkQuotaResponse) =>
+      val pbCheckQuotaResponse = PbCheckQuotaResponse.newBuilder().setAvailable(
+        checkQuotaResponse.isAvailable).setReason(checkQuotaResponse.reason)
       val payload = PbHeartbeatFromApplicationResponse.newBuilder()
         .setStatus(statusCode.getValue)
         .addAllExcludedWorkers(
@@ -760,6 +886,8 @@ object ControlMessages extends Logging {
           unknownWorkers.asScala.map(PbSerDeUtils.toPbWorkerInfo(_, true, true)).toList.asJava)
         .addAllShuttingWorkers(
           shuttingWorkers.asScala.map(PbSerDeUtils.toPbWorkerInfo(_, true, true)).toList.asJava)
+        .addAllRegisteredShuffles(registeredShuffles)
+        .setCheckQuotaResponse(pbCheckQuotaResponse)
         .build().toByteArray
       new TransportMessage(MessageType.HEARTBEAT_FROM_APPLICATION_RESPONSE, payload)
 
@@ -786,6 +914,14 @@ object ControlMessages extends Logging {
         .setRequestId(requestId).build().toByteArray
       new TransportMessage(MessageType.REPORT_WORKER_FAILURE, payload)
 
+    case ReportWorkerDecommission(workers, requestId) =>
+      val payload = PbReportWorkerDecommission.newBuilder()
+        .addAllWorkers(workers.asScala.map { workerInfo =>
+          PbSerDeUtils.toPbWorkerInfo(workerInfo, true, false)
+        }.toList.asJava)
+        .setRequestId(requestId).build().toByteArray
+      new TransportMessage(MessageType.REPORT_WORKER_DECOMMISSION, payload)
+
     case pb: PbRemoveWorkersUnavailableInfo =>
       new TransportMessage(MessageType.REMOVE_WORKERS_UNAVAILABLE_INFO, pb.toByteArray)
 
@@ -809,7 +945,8 @@ object ControlMessages extends Logging {
           rangeReadFilter,
           userIdentifier,
           pushDataTimeout,
-          partitionSplitEnabled) =>
+          partitionSplitEnabled,
+          isSegmentGranularityVisible) =>
       val payload = PbReserveSlots.newBuilder()
         .setApplicationId(applicationId)
         .setShuffleId(shuffleId)
@@ -822,6 +959,7 @@ object ControlMessages extends Logging {
         .setUserIdentifier(PbSerDeUtils.toPbUserIdentifier(userIdentifier))
         .setPushDataTimeout(pushDataTimeout)
         .setPartitionSplitEnabled(partitionSplitEnabled)
+        .setIsSegmentGranularityVisible(isSegmentGranularityVisible)
         .build().toByteArray
       new TransportMessage(MessageType.RESERVE_SLOTS, payload)
 
@@ -943,21 +1081,17 @@ object ControlMessages extends Logging {
 
       case RELEASE_SLOTS_RESPONSE_VALUE =>
         val pbReleaseSlotsResponse = PbReleaseSlotsResponse.parseFrom(message.getPayload)
-        ReleaseSlotsResponse(Utils.toStatusCode(pbReleaseSlotsResponse.getStatus))
+        ReleaseSlotsResponse(StatusCode.fromValue(pbReleaseSlotsResponse.getStatus))
 
       case REGISTER_WORKER_VALUE =>
         PbRegisterWorker.parseFrom(message.getPayload)
 
       case HEARTBEAT_FROM_WORKER_VALUE =>
         val pbHeartbeatFromWorker = PbHeartbeatFromWorker.parseFrom(message.getPayload)
-        val estimatedAppDiskUsage = new util.HashMap[String, java.lang.Long]()
         val userResourceConsumption = PbSerDeUtils.fromPbUserResourceConsumption(
           pbHeartbeatFromWorker.getUserResourceConsumptionMap)
         val pbDisks =
           pbHeartbeatFromWorker.getDisksList.asScala.toSeq.map(PbSerDeUtils.fromPbDiskInfo)
-        if (!pbHeartbeatFromWorker.getEstimatedAppDiskUsageMap.isEmpty) {
-          estimatedAppDiskUsage.putAll(pbHeartbeatFromWorker.getEstimatedAppDiskUsageMap)
-        }
         val activeShuffleKeys = new util.HashSet[String]()
         if (!pbHeartbeatFromWorker.getActiveShuffleKeysList.isEmpty) {
           activeShuffleKeys.addAll(pbHeartbeatFromWorker.getActiveShuffleKeysList)
@@ -974,7 +1108,6 @@ object ControlMessages extends Logging {
           pbDisks,
           userResourceConsumption,
           activeShuffleKeys,
-          estimatedAppDiskUsage,
           pbHeartbeatFromWorker.getHighWorkload,
           workerStatus,
           pbHeartbeatFromWorker.getRequestId)
@@ -1018,6 +1151,7 @@ object ControlMessages extends Logging {
           pbRequestSlots.getAvailableStorageTypes,
           excludedWorkerInfoSet,
           pbRequestSlots.getPacked,
+          pbRequestSlots.getTagsExpr,
           pbRequestSlots.getRequestId)
 
       case REQUEST_SLOTS_RESPONSE_VALUE =>
@@ -1031,7 +1165,7 @@ object ControlMessages extends Logging {
               pbRequestSlotsResponse.getWorkerResourceMap)
           }
         RequestSlotsResponse(
-          Utils.toStatusCode(pbRequestSlotsResponse.getStatus),
+          StatusCode.fromValue(pbRequestSlotsResponse.getStatus),
           workerResource)
 
       case CHANGE_LOCATION_VALUE =>
@@ -1047,16 +1181,22 @@ object ControlMessages extends Logging {
           pbMapperEnd.getMapId,
           pbMapperEnd.getAttemptId,
           pbMapperEnd.getNumMappers,
-          pbMapperEnd.getPartitionId)
+          pbMapperEnd.getPartitionId,
+          pbMapperEnd.getPushFailureBatchesMap.asScala.map {
+            case (partitionId, pushFailedBatchSet) =>
+              (partitionId, PbSerDeUtils.fromPbLocationPushFailedBatches(pushFailedBatchSet))
+          }.toMap.asJava)
 
       case MAPPER_END_RESPONSE_VALUE =>
         val pbMapperEndResponse = PbMapperEndResponse.parseFrom(message.getPayload)
-        MapperEndResponse(Utils.toStatusCode(pbMapperEndResponse.getStatus))
+        MapperEndResponse(StatusCode.fromValue(pbMapperEndResponse.getStatus))
 
       case GET_REDUCER_FILE_GROUP_VALUE =>
         val pbGetReducerFileGroup = PbGetReducerFileGroup.parseFrom(message.getPayload)
         GetReducerFileGroup(
-          pbGetReducerFileGroup.getShuffleId)
+          pbGetReducerFileGroup.getShuffleId,
+          pbGetReducerFileGroup.getIsSegmentGranularityVisible,
+          message.getSerdeVersion)
 
       case GET_REDUCER_FILE_GROUP_RESPONSE_VALUE =>
         val pbGetReducerFileGroupResponse = PbGetReducerFileGroupResponse
@@ -1082,11 +1222,18 @@ object ControlMessages extends Logging {
 
         val attempts = pbGetReducerFileGroupResponse.getAttemptsList.asScala.map(_.toInt).toArray
         val partitionIds = new util.HashSet(pbGetReducerFileGroupResponse.getPartitionIdsList)
+        val pushFailedBatches = pbGetReducerFileGroupResponse.getPushFailedBatchesMap.asScala.map {
+          case (uniqueId, pushFailedBatchSet) =>
+            (uniqueId, PbSerDeUtils.fromPbLocationPushFailedBatches(pushFailedBatchSet))
+        }.toMap.asJava
+        val broadcast = pbGetReducerFileGroupResponse.getBroadcast.toByteArray
         GetReducerFileGroupResponse(
-          Utils.toStatusCode(pbGetReducerFileGroupResponse.getStatus),
+          StatusCode.fromValue(pbGetReducerFileGroupResponse.getStatus),
           fileGroup,
           attempts,
-          partitionIds)
+          partitionIds,
+          pushFailedBatches,
+          broadcast)
 
       case GET_SHUFFLE_ID_VALUE =>
         message.getParsedPayload()
@@ -1103,16 +1250,22 @@ object ControlMessages extends Logging {
       case UNREGISTER_SHUFFLE_VALUE =>
         PbUnregisterShuffle.parseFrom(message.getPayload)
 
+      case BATCH_UNREGISTER_SHUFFLES_VALUE =>
+        PbBatchUnregisterShuffles.parseFrom(message.getPayload)
+
       case UNREGISTER_SHUFFLE_RESPONSE_VALUE =>
         PbUnregisterShuffleResponse.parseFrom(message.getPayload)
 
+      case BATCH_UNREGISTER_SHUFFLE_RESPONSE_VALUE =>
+        PbBatchUnregisterShuffleResponse.parseFrom(message.getPayload)
+
       case APPLICATION_LOST_VALUE =>
         val pbApplicationLost = PbApplicationLost.parseFrom(message.getPayload)
         ApplicationLost(pbApplicationLost.getAppId, pbApplicationLost.getRequestId)
 
       case APPLICATION_LOST_RESPONSE_VALUE =>
         val pbApplicationLostResponse = PbApplicationLostResponse.parseFrom(message.getPayload)
-        ApplicationLostResponse(Utils.toStatusCode(pbApplicationLostResponse.getStatus))
+        ApplicationLostResponse(StatusCode.fromValue(pbApplicationLostResponse.getStatus))
 
       case HEARTBEAT_FROM_APPLICATION_VALUE =>
         val pbHeartbeatFromApplication = PbHeartbeatFromApplication.parseFrom(message.getPayload)
@@ -1120,6 +1273,10 @@ object ControlMessages extends Logging {
           pbHeartbeatFromApplication.getAppId,
           pbHeartbeatFromApplication.getTotalWritten,
           pbHeartbeatFromApplication.getFileCount,
+          pbHeartbeatFromApplication.getShuffleCount,
+          pbHeartbeatFromApplication.getApplicationCount,
+          pbHeartbeatFromApplication.getShuffleFallbackCountsMap,
+          pbHeartbeatFromApplication.getApplicationFallbackCountsMap,
           new util.ArrayList[WorkerInfo](
             pbHeartbeatFromApplication.getNeedCheckedWorkerListList.asScala
               .map(PbSerDeUtils.fromPbWorkerInfo).toList.asJava),
@@ -1129,14 +1286,17 @@ object ControlMessages extends Logging {
       case HEARTBEAT_FROM_APPLICATION_RESPONSE_VALUE =>
         val pbHeartbeatFromApplicationResponse =
           PbHeartbeatFromApplicationResponse.parseFrom(message.getPayload)
+        val pbCheckQuotaResponse = pbHeartbeatFromApplicationResponse.getCheckQuotaResponse
         HeartbeatFromApplicationResponse(
-          Utils.toStatusCode(pbHeartbeatFromApplicationResponse.getStatus),
+          StatusCode.fromValue(pbHeartbeatFromApplicationResponse.getStatus),
           pbHeartbeatFromApplicationResponse.getExcludedWorkersList.asScala
             .map(PbSerDeUtils.fromPbWorkerInfo).toList.asJava,
           pbHeartbeatFromApplicationResponse.getUnknownWorkersList.asScala
             .map(PbSerDeUtils.fromPbWorkerInfo).toList.asJava,
           pbHeartbeatFromApplicationResponse.getShuttingWorkersList.asScala
-            .map(PbSerDeUtils.fromPbWorkerInfo).toList.asJava)
+            .map(PbSerDeUtils.fromPbWorkerInfo).toList.asJava,
+          pbHeartbeatFromApplicationResponse.getRegisteredShufflesList,
+          CheckQuotaResponse(pbCheckQuotaResponse.getAvailable, pbCheckQuotaResponse.getReason))
 
       case CHECK_QUOTA_VALUE =>
         val pbCheckAvailable = PbCheckQuota.parseFrom(message.getPayload)
@@ -1156,6 +1316,13 @@ object ControlMessages extends Logging {
             .asScala.map(PbSerDeUtils.fromPbWorkerInfo).toList.asJava),
           pbReportWorkerUnavailable.getRequestId)
 
+      case REPORT_WORKER_DECOMMISSION_VALUE =>
+        val pbReportWorkerDecommission = PbReportWorkerDecommission.parseFrom(message.getPayload)
+        ReportWorkerDecommission(
+          new util.ArrayList[WorkerInfo](pbReportWorkerDecommission.getWorkersList
+            .asScala.map(PbSerDeUtils.fromPbWorkerInfo).toList.asJava),
+          pbReportWorkerDecommission.getRequestId)
+
       case REMOVE_WORKERS_UNAVAILABLE_INFO_VALUE =>
         PbRemoveWorkersUnavailableInfo.parseFrom(message.getPayload)
 
@@ -1193,12 +1360,13 @@ object ControlMessages extends Logging {
           pbReserveSlots.getRangeReadFilter,
           userIdentifier,
           pbReserveSlots.getPushDataTimeout,
-          pbReserveSlots.getPartitionSplitEnabled)
+          pbReserveSlots.getPartitionSplitEnabled,
+          pbReserveSlots.getIsSegmentGranularityVisible)
 
       case RESERVE_SLOTS_RESPONSE_VALUE =>
         val pbReserveSlotsResponse = PbReserveSlotsResponse.parseFrom(message.getPayload)
         ReserveSlotsResponse(
-          Utils.toStatusCode(pbReserveSlotsResponse.getStatus),
+          StatusCode.fromValue(pbReserveSlotsResponse.getStatus),
           pbReserveSlotsResponse.getReason)
 
       case COMMIT_FILES_VALUE =>
@@ -1225,7 +1393,7 @@ object ControlMessages extends Logging {
           committedBitMap.put(entry._1, Utils.byteStringToRoaringBitmap(entry._2))
         }
         CommitFilesResponse(
-          Utils.toStatusCode(pbCommitFilesResponse.getStatus),
+          StatusCode.fromValue(pbCommitFilesResponse.getStatus),
           pbCommitFilesResponse.getCommittedPrimaryIdsList,
           pbCommitFilesResponse.getCommittedReplicaIdsList,
           pbCommitFilesResponse.getFailedPrimaryIdsList,
@@ -1247,7 +1415,7 @@ object ControlMessages extends Logging {
       case DESTROY_RESPONSE_VALUE =>
         val pbDestroyResponse = PbDestroyWorkerSlotsResponse.parseFrom(message.getPayload)
         DestroyWorkerSlotsResponse(
-          Utils.toStatusCode(pbDestroyResponse.getStatus),
+          StatusCode.fromValue(pbDestroyResponse.getStatus),
           pbDestroyResponse.getFailedPrimariesList,
           pbDestroyResponse.getFailedReplicasList)
 
@@ -1263,8 +1431,8 @@ object ControlMessages extends Logging {
       case CHECK_APPLICATION_TIMEOUT_VALUE =>
         CheckForApplicationTimeOut
 
-      case CHECK_FOR_HDFS_EXPIRED_DIRS_TIMEOUT_VALUE =>
-        CheckForHDFSExpiredDirsTimeout
+      case CHECK_FOR_DFS_EXPIRED_DIRS_TIMEOUT_VALUE =>
+        CheckForDFSExpiredDirsTimeout
 
       case WORKER_LOST_VALUE =>
         PbWorkerLost.parseFrom(message.getPayload)
@@ -1281,7 +1449,7 @@ object ControlMessages extends Logging {
 
       case STAGE_END_RESPONSE_VALUE =>
         val pbStageEndResponse = PbStageEndResponse.parseFrom(message.getPayload)
-        StageEndResponse(Utils.toStatusCode(pbStageEndResponse.getStatus))
+        StageEndResponse(StatusCode.fromValue(pbStageEndResponse.getStatus))
 
       case CHECK_WORKERS_AVAILABLE_VALUE =>
         PbCheckWorkersAvailable.parseFrom(message.getPayload)
@@ -1300,6 +1468,15 @@ object ControlMessages extends Logging {
 
       case REPORT_BARRIER_STAGE_ATTEMPT_FAILURE_RESPONSE_VALUE =>
         PbReportBarrierStageAttemptFailureResponse.parseFrom(message.getPayload)
+
+      case PUSH_MERGED_DATA_SPLIT_PARTITION_INFO_VALUE =>
+        PbPushMergedDataSplitPartitionInfo.parseFrom(message.getPayload)
+
+      case GET_STAGE_END_VALUE =>
+        PbGetStageEnd.parseFrom(message.getPayload)
+
+      case GET_STAGE_END_RESPONSE_VALUE =>
+        PbGetStageEndResponse.parseFrom(message.getPayload)
     }
   }
 }
