@@ -34,11 +34,11 @@ import org.apache.celeborn.common.CelebornConf
 import org.apache.celeborn.common.internal.Logging
 import org.apache.celeborn.common.network.TransportContext
 import org.apache.celeborn.common.network.client._
-import org.apache.celeborn.common.network.protocol.{RequestMessage => NRequestMessage, RpcRequest}
+import org.apache.celeborn.common.network.protocol.{RequestMessage => NRequestMessage, RpcRequest, SerdeVersion, TransportMessage}
 import org.apache.celeborn.common.network.sasl.{SaslClientBootstrap, SaslServerBootstrap}
 import org.apache.celeborn.common.network.sasl.registration.{RegistrationClientBootstrap, RegistrationServerBootstrap}
 import org.apache.celeborn.common.network.server._
-import org.apache.celeborn.common.protocol.{RpcNameConstants, TransportModuleConstants}
+import org.apache.celeborn.common.protocol.RpcNameConstants
 import org.apache.celeborn.common.rpc._
 import org.apache.celeborn.common.serializer.{JavaSerializer, JavaSerializerInstance, SerializationStream}
 import org.apache.celeborn.common.util.{ByteBufferInputStream, ByteBufferOutputStream, JavaUtils, ThreadUtils, Utils}
@@ -55,7 +55,9 @@ class NettyRpcEnv(
     config.transportModule,
     celebornConf.rpcIoThreads.getOrElse(config.numUsableCores))
 
-  private val dispatcher: Dispatcher = new Dispatcher(this)
+  private val _rpcSource: RpcSource = new RpcSource(celebornConf, config.role)
+
+  private val dispatcher: Dispatcher = new Dispatcher(this, _rpcSource)
 
   private var worker: RpcEndpoint = null
 
@@ -291,7 +293,7 @@ class NettyRpcEnv(
       case NonFatal(e) =>
         onFailure(e)
     }
-    promise.future.mapTo[T].recover(timeout.addMessageIfTimeout)(ThreadUtils.sameThread)
+    promise.future.mapTo[T].recover(timeout.addMessageIfTimeout(address))(ThreadUtils.sameThread)
   }
 
   private[celeborn] def serialize(content: Any): ByteBuffer = {
@@ -361,6 +363,8 @@ class NettyRpcEnv(
       deserializationAction()
     }
   }
+
+  override def rpcSource(): RpcSource = _rpcSource
 }
 
 private[celeborn] object NettyRpcEnv extends Logging {
@@ -395,8 +399,13 @@ private[celeborn] class NettyRpcEnvFactory extends RpcEnvFactory with Logging {
       new JavaSerializer(celebornConf).newInstance().asInstanceOf[JavaSerializerInstance]
     val nettyEnv = new NettyRpcEnv(config, javaSerializerInstance)
     val startNettyRpcEnv: Int => (NettyRpcEnv, Int) = { actualPort =>
-      logInfo(s"Starting RPC Server [${config.name}] on ${config.bindAddress}:$actualPort " +
-        s"with advisor endpoint ${config.advertiseAddress}:$actualPort")
+      if (celebornConf.bindWildcardAddress) {
+        logInfo(s"Starting RPC Server [${config.name}] on wildcard address with port" +
+          s" $actualPort, and advertised endpoint ${config.advertiseAddress}:$actualPort")
+      } else {
+        logInfo(s"Starting RPC Server [${config.name}] on ${config.bindAddress}:$actualPort " +
+          s"with advertised endpoint ${config.advertiseAddress}:$actualPort")
+      }
       nettyEnv.startServer(config.bindAddress, actualPort)
       (nettyEnv, nettyEnv.address.port)
     }
@@ -495,6 +504,20 @@ private[celeborn] class RequestMessage(
       writeRpcAddress(out, senderAddress)
       writeRpcAddress(out, receiver.address)
       out.writeUTF(receiver.name)
+      val msg = Utils.toTransportMessage(content)
+      msg match {
+        case transMsg: TransportMessage =>
+          // Check if the msg is a TransportMessage with language-agnostic V2 serdeVersion.
+          // If so, write the marker and the body explicitly.
+          if (transMsg.getSerdeVersion == SerdeVersion.V2) {
+            val out = new DataOutputStream(bos)
+            out.writeByte(SerdeVersion.V2.getMarker)
+            out.write(transMsg.toByteBuffer.array)
+            out.close()
+            return bos.toByteBuffer
+          }
+        case _ =>
+      }
       val s = nettyEnv.serializeStream(out)
       try {
         s.writeObject(Utils.toTransportMessage(content))
@@ -553,6 +576,7 @@ private[celeborn] object RequestMessage {
 /**
  * A response that indicates some failure happens in the receiver side.
  */
+@SerialVersionUID(2793139166962436434L)
 private[celeborn] case class RpcFailure(e: Throwable)
 
 /**
