@@ -19,22 +19,29 @@ package org.apache.celeborn.common.util
 
 import java.util
 import java.util.concurrent.ConcurrentHashMap
-import java.util.function.IntFunction
 
 import scala.collection.JavaConverters._
 
 import com.google.protobuf.InvalidProtocolBufferException
 
 import org.apache.celeborn.common.identity.UserIdentifier
-import org.apache.celeborn.common.meta.{AppDiskUsage, AppDiskUsageSnapShot, ApplicationMeta, DeviceInfo, DiskFileInfo, DiskInfo, MapFileMeta, ReduceFileMeta, WorkerEventInfo, WorkerInfo, WorkerStatus}
+import org.apache.celeborn.common.meta.{ApplicationMeta, DeviceInfo, DiskFileInfo, DiskInfo, MapFileMeta, ReduceFileMeta, WorkerEventInfo, WorkerInfo, WorkerStatus}
+import org.apache.celeborn.common.meta.MapFileMeta.SegmentIndex
 import org.apache.celeborn.common.protocol._
 import org.apache.celeborn.common.protocol.PartitionLocation.Mode
 import org.apache.celeborn.common.protocol.message.ControlMessages.WorkerResource
 import org.apache.celeborn.common.quota.ResourceConsumption
 import org.apache.celeborn.common.util.{CollectionUtils => localCollectionUtils}
+import org.apache.celeborn.common.write.LocationPushFailedBatches
 
 object PbSerDeUtils {
 
+  private var masterPersistWorkerNetworkLocation: Boolean = false
+
+  def setMasterPersistWorkerNetworkLocation(masterPersistWorkerNetworkLocation: Boolean) = {
+    this.masterPersistWorkerNetworkLocation = masterPersistWorkerNetworkLocation
+  }
+
   @throws[InvalidProtocolBufferException]
   def fromPbSortedShuffleFileSet(data: Array[Byte]): util.Set[String] = {
     val pbSortedShuffleFileSet = PbSortedShuffleFileSet.parseFrom(data)
@@ -97,7 +104,15 @@ object PbSerDeUtils {
       case PartitionType.REDUCE =>
         new ReduceFileMeta(pbFileInfo.getChunkOffsetsList)
       case PartitionType.MAP =>
-        new MapFileMeta(pbFileInfo.getBufferSize, pbFileInfo.getNumSubpartitions)
+        val fileMeta = new MapFileMeta(
+          pbFileInfo.getBufferSize,
+          pbFileInfo.getNumSubpartitions,
+          pbFileInfo.getIsSegmentGranularityVisible,
+          pbFileInfo.getPartitionWritingSegmentMap,
+          fromPbSegmentIndexList(pbFileInfo.getSegmentIndexList))
+        // writer always closed as this is committed file info.
+        fileMeta.setIsWriterClosed(true)
+        fileMeta
       case PartitionType.MAPGROUP =>
         throw new NotImplementedError("Map group is not implemented")
     }
@@ -109,6 +124,19 @@ object PbSerDeUtils {
       pbFileInfo.getBytesFlushed)
   }
 
+  private def fromPbSegmentIndexList(
+      segmentIndexList: util.List[PbSegmentIndex]): util.List[SegmentIndex] = {
+    segmentIndexList.asScala.map(firstBufferIndexToSegment =>
+      new SegmentIndex(
+        firstBufferIndexToSegment.getFirstBufferIndexToSegmentMap)).asJava
+  }
+
+  private def toPbSegmentIndex(
+      firstBufferIndexToSegment: SegmentIndex): PbSegmentIndex = {
+    PbSegmentIndex.newBuilder().putAllFirstBufferIndexToSegment(
+      firstBufferIndexToSegment.getFirstBufferIndexToSegment).build()
+  }
+
   def toPbFileInfo(fileInfo: DiskFileInfo): PbFileInfo = {
     val builder = PbFileInfo.newBuilder
       .setFilePath(fileInfo.getFilePath)
@@ -120,6 +148,10 @@ object PbSerDeUtils {
       builder.setPartitionType(PartitionType.MAP.getValue)
       builder.setBufferSize(mapFileMeta.getBufferSize)
       builder.setNumSubpartitions(mapFileMeta.getNumSubpartitions)
+      builder.setIsSegmentGranularityVisible(mapFileMeta.isSegmentGranularityVisible)
+      builder.putAllPartitionWritingSegment(mapFileMeta.getSubPartitionWritingSegmentId)
+      builder.addAllSegmentIndex(
+        mapFileMeta.getSubPartitionSegmentIndexes.asScala.map(toPbSegmentIndex).asJava)
     } else {
       val reduceFileMeta = fileInfo.getFileMeta.asInstanceOf[ReduceFileMeta]
       builder.setPartitionType(PartitionType.REDUCE.getValue)
@@ -234,7 +266,7 @@ object PbSerDeUtils {
     }
     val userResourceConsumption =
       PbSerDeUtils.fromPbUserResourceConsumption(pbWorkerInfo.getUserResourceConsumptionMap)
-    new WorkerInfo(
+    val workerInfo = new WorkerInfo(
       pbWorkerInfo.getHost,
       pbWorkerInfo.getRpcPort,
       pbWorkerInfo.getPushPort,
@@ -243,6 +275,10 @@ object PbSerDeUtils {
       pbWorkerInfo.getInternalPort,
       disks,
       userResourceConsumption)
+    if (masterPersistWorkerNetworkLocation) {
+      workerInfo.networkLocation = pbWorkerInfo.getNetworkLocation
+    }
+    workerInfo
   }
 
   def toPbWorkerInfo(
@@ -256,6 +292,10 @@ object PbSerDeUtils {
       .setPushPort(workerInfo.pushPort)
       .setReplicatePort(workerInfo.replicatePort)
       .setInternalPort(workerInfo.internalPort)
+    if (masterPersistWorkerNetworkLocation) {
+      builder.setNetworkLocation(workerInfo.networkLocation)
+    }
+
     if (!eliminateUserResourceConsumption) {
       builder.putAllUserResourceConsumption(
         PbSerDeUtils.toPbUserResourceConsumption(workerInfo.userResourceConsumption))
@@ -372,52 +412,13 @@ object PbSerDeUtils {
         .addAllReplicaPartitions(replicaPartitions)
         .setNetworkLocation(workerInfo.networkLocation)
         .build()
-      workerInfo.toUniqueId() -> pbWorkerResource
+      workerInfo.toUniqueId -> pbWorkerResource
     }.asJava
   }
 
-  def fromPbAppDiskUsage(pbAppDiskUsage: PbAppDiskUsage): AppDiskUsage = {
-    AppDiskUsage(pbAppDiskUsage.getAppId, pbAppDiskUsage.getEstimatedUsage)
-  }
-
-  def toPbAppDiskUsage(appDiskUsage: AppDiskUsage): PbAppDiskUsage = {
-    PbAppDiskUsage.newBuilder()
-      .setAppId(appDiskUsage.appId)
-      .setEstimatedUsage(appDiskUsage.estimatedUsage)
-      .build()
-  }
-
-  def fromPbAppDiskUsageSnapshot(
-      pbAppDiskUsageSnapShot: PbAppDiskUsageSnapshot): AppDiskUsageSnapShot = {
-    val snapShot = new AppDiskUsageSnapShot(pbAppDiskUsageSnapShot.getTopItemCount)
-    snapShot.startSnapShotTime = pbAppDiskUsageSnapShot.getStartSnapShotTime
-    snapShot.endSnapShotTime = pbAppDiskUsageSnapShot.getEndSnapshotTime
-    snapShot.restoreFromSnapshot(
-      pbAppDiskUsageSnapShot
-        .getTopNItemsList
-        .asScala
-        .map(fromPbAppDiskUsage)
-        .asJava
-        .stream()
-        .toArray(new IntFunction[Array[AppDiskUsage]]() {
-          override def apply(value: Int): Array[AppDiskUsage] = new Array[AppDiskUsage](value)
-        }))
-    snapShot
-  }
-
-  def toPbAppDiskUsageSnapshot(snapshots: AppDiskUsageSnapShot): PbAppDiskUsageSnapshot = {
-    PbAppDiskUsageSnapshot.newBuilder()
-      .setTopItemCount(snapshots.topItemCount)
-      .setStartSnapShotTime(snapshots.startSnapShotTime)
-      .setEndSnapshotTime(snapshots.endSnapShotTime)
-      // topNItems some value could be null
-      .addAllTopNItems(snapshots.topNItems.filter(_ != null).map(toPbAppDiskUsage).toList.asJava)
-      .build()
-  }
-
   def toPbSnapshotMetaInfo(
       estimatedPartitionSize: java.lang.Long,
-      registeredShuffle: java.util.Set[String],
+      registeredShuffle: java.util.Map[String, java.util.Set[Integer]],
       hostnameSet: java.util.Set[String],
       excludedWorkers: java.util.Set[WorkerInfo],
       manuallyExcludedWorkers: java.util.Set[WorkerInfo],
@@ -426,15 +427,20 @@ object PbSerDeUtils {
       workers: java.util.Set[WorkerInfo],
       partitionTotalWritten: java.lang.Long,
       partitionTotalFileCount: java.lang.Long,
-      appDiskUsageMetricSnapshots: Array[AppDiskUsageSnapShot],
-      currentAppDiskUsageMetricsSnapshot: AppDiskUsageSnapShot,
+      shuffleTotalCount: java.lang.Long,
+      applicationTotalCount: java.lang.Long,
+      shuffleFallbackCounts: java.util.Map[String, java.lang.Long],
+      applicationFallbackCounts: java.util.Map[String, java.lang.Long],
       lostWorkers: ConcurrentHashMap[WorkerInfo, java.lang.Long],
       shutdownWorkers: java.util.Set[WorkerInfo],
       workerEventInfos: ConcurrentHashMap[WorkerInfo, WorkerEventInfo],
-      applicationMetas: ConcurrentHashMap[String, ApplicationMeta]): PbSnapshotMetaInfo = {
+      applicationMetas: ConcurrentHashMap[String, ApplicationMeta],
+      decommissionWorkers: java.util.Set[WorkerInfo]): PbSnapshotMetaInfo = {
     val builder = PbSnapshotMetaInfo.newBuilder()
       .setEstimatedPartitionSize(estimatedPartitionSize)
-      .addAllRegisteredShuffle(registeredShuffle)
+      .addAllRegisteredShuffle(registeredShuffle.asScala.flatMap { appIdAndShuffleId =>
+        appIdAndShuffleId._2.asScala.map(i => Utils.makeShuffleKey(appIdAndShuffleId._1, i))
+      }.asJava)
       .addAllHostnameSet(hostnameSet)
       .addAllExcludedWorkers(excludedWorkers.asScala.map(toPbWorkerInfo(_, true, false)).asJava)
       .addAllManuallyExcludedWorkers(manuallyExcludedWorkers.asScala
@@ -444,22 +450,21 @@ object PbSerDeUtils {
       .addAllWorkers(workers.asScala.map(toPbWorkerInfo(_, true, false)).asJava)
       .setPartitionTotalWritten(partitionTotalWritten)
       .setPartitionTotalFileCount(partitionTotalFileCount)
-      // appDiskUsageMetricSnapshots can have null values,
-      // protobuf repeated value can't support null value in list.
-      .addAllAppDiskUsageMetricSnapshots(appDiskUsageMetricSnapshots.filter(_ != null)
-        .map(toPbAppDiskUsageSnapshot).toList.asJava)
+      .setShuffleTotalCount(shuffleTotalCount)
+      .setApplicationTotalCount(applicationTotalCount)
+      .putAllShuffleFallbackCounts(shuffleFallbackCounts)
+      .putAllApplicationFallbackCounts(applicationFallbackCounts)
       .putAllLostWorkers(lostWorkers.asScala.map {
-        case (worker: WorkerInfo, time: java.lang.Long) => (worker.toUniqueId(), time)
+        case (worker: WorkerInfo, time: java.lang.Long) => (worker.toUniqueId, time)
       }.asJava)
       .addAllShutdownWorkers(shutdownWorkers.asScala.map(toPbWorkerInfo(_, true, false)).asJava)
       .putAllWorkerEventInfos(workerEventInfos.asScala.map {
         case (worker, workerEventInfo) =>
-          (worker.toUniqueId(), PbSerDeUtils.toPbWorkerEventInfo(workerEventInfo))
+          (worker.toUniqueId, PbSerDeUtils.toPbWorkerEventInfo(workerEventInfo))
       }.asJava)
-    if (currentAppDiskUsageMetricsSnapshot != null) {
-      builder.setCurrentAppDiskUsageMetricsSnapshot(
-        toPbAppDiskUsageSnapshot(currentAppDiskUsageMetricsSnapshot))
-    }
+      .addAllDecommissionWorkers(decommissionWorkers.asScala.map(
+        toPbWorkerInfo(_, true, false)).asJava)
+
     val pbApplicationMetas = applicationMetas.asScala.map {
       case (appId, applicationMeta) => (appId, toPbApplicationMeta(applicationMeta))
     }.asJava
@@ -476,7 +481,7 @@ object PbSerDeUtils {
   }
 
   def fromPbApplicationMeta(pbApplicationMeta: PbApplicationMeta): ApplicationMeta = {
-    new ApplicationMeta(pbApplicationMeta.getAppId, pbApplicationMeta.getSecret)
+    ApplicationMeta(pbApplicationMeta.getAppId, pbApplicationMeta.getSecret)
   }
 
   def toPbWorkerStatus(workerStatus: WorkerStatus): PbWorkerStatus = {
@@ -500,7 +505,7 @@ object PbSerDeUtils {
   def fromPbWorkerEventInfo(pbWorkerEventInfo: PbWorkerEventInfo): WorkerEventInfo = {
     new WorkerEventInfo(
       pbWorkerEventInfo.getWorkerEventType.getNumber,
-      pbWorkerEventInfo.getEventStartTime())
+      pbWorkerEventInfo.getEventStartTime)
   }
 
   private def toPackedPartitionLocation(
@@ -510,7 +515,7 @@ object PbSerDeUtils {
       location: PartitionLocation): PbPackedPartitionLocations.Builder = {
     pbPackedLocationsBuilder.addIds(location.getId)
     pbPackedLocationsBuilder.addEpoches(location.getEpoch)
-    pbPackedLocationsBuilder.addWorkerIds(workerIdIndex(location.getWorker.toUniqueId()))
+    pbPackedLocationsBuilder.addWorkerIds(workerIdIndex(location.getWorker.toUniqueId))
     pbPackedLocationsBuilder.addMapIdBitMap(
       Utils.roaringBitmapToByteString(location.getMapIdBitMap))
     pbPackedLocationsBuilder.addTypes(location.getStorageInfo.getType.getValue)
@@ -524,6 +529,14 @@ object PbSerDeUtils {
       pbPackedLocationsBuilder.addFilePaths("")
     }
     pbPackedLocationsBuilder.addAvailableStorageTypes(location.getStorageInfo.availableStorageTypes)
+    pbPackedLocationsBuilder.addFileSizes(location.getStorageInfo.getFileSize)
+    val chunkOffsets = PbChunkOffsets.newBuilder()
+    if (null != location.getStorageInfo.chunkOffsets && !location.getStorageInfo.chunkOffsets.isEmpty) {
+      chunkOffsets.addAllChunkOffset(location.getStorageInfo.chunkOffsets).build()
+      pbPackedLocationsBuilder.addChunksOffsets(chunkOffsets)
+    } else {
+      pbPackedLocationsBuilder.addChunksOffsets(chunkOffsets.build())
+    }
     pbPackedLocationsBuilder.addModes(location.getMode.mode())
   }
 
@@ -536,7 +549,7 @@ object PbSerDeUtils {
 
     val allLocations = (inputLocations ++ implicateLocations)
     val workerIdList = new util.ArrayList[String](
-      allLocations.map(_.getWorker.toUniqueId()).toSet.asJava)
+      allLocations.map(_.getWorker.toUniqueId).toSet.asJava)
     val workerIdIndex = workerIdList.asScala.zipWithIndex.toMap
     val mountPointsList = new util.ArrayList[String](
       allLocations.map(
@@ -624,6 +637,28 @@ object PbSerDeUtils {
         Mode.REPLICA
       }
 
+    val storageInfo =
+      if (pbPackedPartitionLocations.getFileSizesList.isEmpty ||
+        pbPackedPartitionLocations.getChunksOffsetsList.isEmpty) {
+        new StorageInfo(
+          StorageInfo.typesMap.get(pbPackedPartitionLocations.getTypes(index)),
+          pbPackedPartitionLocations.getMountPointsSet(
+            pbPackedPartitionLocations.getMountPoints(index)),
+          pbPackedPartitionLocations.getFinalResult(index),
+          filePath,
+          pbPackedPartitionLocations.getAvailableStorageTypes(index))
+      } else {
+        new StorageInfo(
+          StorageInfo.typesMap.get(pbPackedPartitionLocations.getTypes(index)),
+          pbPackedPartitionLocations.getMountPointsSet(
+            pbPackedPartitionLocations.getMountPoints(index)),
+          pbPackedPartitionLocations.getFinalResult(index),
+          filePath,
+          pbPackedPartitionLocations.getAvailableStorageTypes(index),
+          pbPackedPartitionLocations.getFileSizes(index),
+          pbPackedPartitionLocations.getChunksOffsets(index).getChunkOffsetList)
+      }
+
     new PartitionLocation(
       pbPackedPartitionLocations.getIds(index),
       pbPackedPartitionLocations.getEpoches(index),
@@ -634,13 +669,7 @@ object PbSerDeUtils {
       workerIdParts(4).toInt,
       mode,
       null,
-      new StorageInfo(
-        StorageInfo.typesMap.get(pbPackedPartitionLocations.getTypes(index)),
-        pbPackedPartitionLocations.getMountPointsSet(
-          pbPackedPartitionLocations.getMountPoints(index)),
-        pbPackedPartitionLocations.getFinalResult(index),
-        filePath,
-        pbPackedPartitionLocations.getAvailableStorageTypes(index)),
+      storageInfo,
       Utils.byteStringToRoaringBitmap(pbPackedPartitionLocations.getMapIdBitMap(index)))
   }
 
@@ -666,8 +695,40 @@ object PbSerDeUtils {
           primaryLocations.asScala.toList ++ replicaLocations.asScala.toList))
         .setNetworkLocation(workerInfo.networkLocation)
         .build()
-      workerInfo.toUniqueId() -> pbWorkerResource
+      workerInfo.toUniqueId -> pbWorkerResource
     }.asJava
   }
 
+  def toPbFailedBatches(failedBatches: util.Set[Int]): PbFailedBatches = {
+    PbFailedBatches.newBuilder()
+      .addAllFailedBatches(failedBatches.asScala.map(new Integer(_)).asJava)
+      .build()
+  }
+
+  def fromPbFailedBatches(failedBatches: PbFailedBatches): util.Set[Int] = {
+    failedBatches.getFailedBatchesList.asScala.map(_.intValue()).toSet.asJava
+  }
+
+  def toPbLocationPushFailedBatches(locationPushFailedBatches: LocationPushFailedBatches)
+      : PbLocationPushFailedBatches = {
+    val builder = PbLocationPushFailedBatches.newBuilder()
+    builder.putAllFailedBatches(
+      locationPushFailedBatches
+        .getFailedBatches
+        .asScala
+        .map(item =>
+          (item._1, toPbFailedBatches(item._2.asScala.map(_.intValue()).toSet.asJava))).asJava)
+    builder.build()
+  }
+
+  def fromPbLocationPushFailedBatches(pbLocationPushFailedBatches: PbLocationPushFailedBatches)
+      : LocationPushFailedBatches = {
+    val batches = new LocationPushFailedBatches()
+    pbLocationPushFailedBatches.getFailedBatchesMap.asScala.foreach { case (key, value) =>
+      batches.getFailedBatches.put(
+        key,
+        fromPbFailedBatches(value).asScala.map(new Integer(_)).asJava)
+    }
+    batches
+  }
 }
