@@ -49,9 +49,10 @@ import org.apache.celeborn.common.internal.Logging
 import org.apache.celeborn.common.meta.{DiskStatus, WorkerInfo}
 import org.apache.celeborn.common.network.protocol.TransportMessage
 import org.apache.celeborn.common.network.util.TransportConf
-import org.apache.celeborn.common.protocol.{PartitionLocation, PartitionSplitMode, PartitionType, TransportModuleConstants}
-import org.apache.celeborn.common.protocol.message.{ControlMessages, Message, StatusCode}
+import org.apache.celeborn.common.protocol.{PartitionLocation, PartitionSplitMode, PartitionType, RpcNameConstants, TransportModuleConstants}
+import org.apache.celeborn.common.protocol.message.{ControlMessages, Message}
 import org.apache.celeborn.common.protocol.message.ControlMessages.WorkerResource
+import org.apache.celeborn.common.rpc.RpcTimeoutException
 import org.apache.celeborn.reflect.DynConstructors
 
 object Utils extends Logging {
@@ -245,8 +246,15 @@ object Utils extends Logging {
     }
   }
 
-  def selectRandomPort(from: Int, to: Int): Int = {
-    ScalaRandom.nextInt(to - from) + from
+  /**
+   * Select a random integer within the specified range.
+   *
+   * @param from the lower bound of the range (inclusive)
+   * @param until the upper bound of the range (exclusive)
+   * @return a randomly selected integer within the range [from, until)
+   */
+  def selectRandomInt(from: Int, until: Int): Int = {
+    ScalaRandom.nextInt(until - 1 - from) + from
   }
 
   def startServiceOnPort[T](
@@ -321,8 +329,7 @@ object Utils extends Logging {
         }
         isBindCollision(e.getCause)
       case e: NativeIoException =>
-        (e.getMessage != null && e.getMessage.startsWith("bind() failed: ")) ||
-          isBindCollision(e.getCause)
+        e.getMessage != null && e.getMessage.startsWith("bind")
       case e: Exception => isBindCollision(e.getCause)
       case _ => false
     }
@@ -415,8 +422,29 @@ object Utils extends Logging {
     customHostname = Some(hostname)
   }
 
-  def localHostName(conf: CelebornConf): String = customHostname.getOrElse {
-    if (conf.bindPreferIP) {
+  def localHostName(conf: CelebornConf): String = {
+    getHostName(conf.bindPreferIP)
+  }
+
+  def localHostNameForAdvertiseAddress(conf: CelebornConf, env: String): String = {
+    if (env.equals(RpcNameConstants.MASTER_SYS) || env.equals(
+        RpcNameConstants.MASTER_INTERNAL_SYS)) {
+      getAdvertiseAddressForMaster(conf)
+    } else {
+      getAdvertiseAddressForWorker(conf)
+    }
+  }
+
+  private def getAdvertiseAddressForMaster(conf: CelebornConf) = {
+    conf.advertiseAddressMasterHost
+  }
+
+  private def getAdvertiseAddressForWorker(conf: CelebornConf) = {
+    getHostName(conf.advertisePreferIP)
+  }
+
+  def getHostName(preferIP: Boolean): String = customHostname.getOrElse {
+    if (preferIP) {
       localIpAddress match {
         case ipv6Address: Inet6Address =>
           val ip = ipv6Address.getHostAddress
@@ -586,7 +614,7 @@ object Utils extends Logging {
     }
   }
 
-  def instantiateDynamicConfigStoreBackend[T](className: String, conf: CelebornConf): T = {
+  def instantiateClassWithCelebornConf[T](className: String, conf: CelebornConf): T = {
     try {
       DynConstructors.builder().impl(className, classOf[CelebornConf])
         .build[T]()
@@ -594,7 +622,7 @@ object Utils extends Logging {
     } catch {
       case e: Throwable =>
         throw new CelebornException(
-          s"Failed to instantiate dynamic config store backend $className.",
+          s"Failed to instantiate class $className with celeborn conf.",
           e)
     }
   }
@@ -690,6 +718,17 @@ object Utils extends Logging {
     s"$shuffleId-$mapId-$attemptId"
   }
 
+  def makeAttemptKey(mapId: Int, attemptId: Int): String = {
+    s"$mapId-$attemptId"
+  }
+
+  def splitAttemptKey(attemptKey: String): (Int, Int) = {
+    val splits = attemptKey.split("-")
+    val mapId = splits(0).toInt
+    val attemptId = splits(1).toInt
+    (mapId, attemptId)
+  }
+
   def shuffleKeyPrefix(shuffleKey: String): String = {
     shuffleKey + "-"
   }
@@ -745,29 +784,39 @@ object Utils extends Logging {
    * Note: code was initially copied from Apache Spark(v3.5.1).
    */
   private def threadInfoToThreadStackTrace(threadInfo: ThreadInfo): ThreadStackTrace = {
-    val monitors = threadInfo.getLockedMonitors.map(m => m.getLockedStackFrame -> m).toMap
-    val stackTrace = StackTrace(threadInfo.getStackTrace.map { frame =>
-      monitors.get(frame) match {
-        case Some(monitor) =>
-          monitor.getLockedStackFrame.toString + s" => holding ${monitor.lockString}"
-        case None =>
-          frame.toString
-      }
+    val threadState = threadInfo.getThreadState
+    val monitors = threadInfo.getLockedMonitors.map(m => m.getLockedStackDepth -> m.toString).toMap
+    val stackTrace = StackTrace(threadInfo.getStackTrace.zipWithIndex.map { case (frame, idx) =>
+      val locked =
+        if (idx == 0 && threadInfo.getLockInfo != null) {
+          threadState match {
+            case Thread.State.BLOCKED =>
+              s"\t-  blocked on ${threadInfo.getLockInfo}\n"
+            case Thread.State.WAITING | Thread.State.TIMED_WAITING =>
+              s"\t-  waiting on ${threadInfo.getLockInfo}\n"
+            case _ => ""
+          }
+        } else ""
+      val locking = monitors.get(idx).map(mi => s"\t-  locked $mi\n").getOrElse("")
+      s"${frame.toString}\n$locked$locking"
     })
 
-    // use a set to dedup re-entrant locks that are held at multiple places
-    val heldLocks =
-      (threadInfo.getLockedSynchronizers ++ threadInfo.getLockedMonitors).map(_.lockString).toSet
-
+    val synchronizers = threadInfo.getLockedSynchronizers.map(_.toString)
+    val monitorStrs = monitors.values.toSeq
     ThreadStackTrace(
-      threadId = threadInfo.getThreadId,
-      threadName = threadInfo.getThreadName,
-      threadState = threadInfo.getThreadState,
-      stackTrace = stackTrace,
-      blockedByThreadId =
-        if (threadInfo.getLockOwnerId < 0) None else Some(threadInfo.getLockOwnerId),
-      blockedByLock = Option(threadInfo.getLockInfo).map(_.lockString).getOrElse(""),
-      holdingLocks = heldLocks.toSeq)
+      threadInfo.getThreadId,
+      threadInfo.getThreadName,
+      threadState,
+      stackTrace,
+      if (threadInfo.getLockOwnerId < 0) None else Some(threadInfo.getLockOwnerId),
+      Option(threadInfo.getLockInfo).map(_.lockString).getOrElse(""),
+      synchronizers ++ monitorStrs,
+      synchronizers,
+      monitorStrs,
+      Option(threadInfo.getLockName),
+      Option(threadInfo.getLockOwnerName),
+      threadInfo.isSuspended,
+      threadInfo.isInNative)
   }
 
   private def readProcessStdout(process: Process): String = {
@@ -991,6 +1040,17 @@ object Utils extends Logging {
     }
   }
 
+  def tryWithResources[R <: Closeable, U](f: => R)(func: R => U): U = {
+    val res = f
+    try {
+      func(f)
+    } finally {
+      if (null != res) {
+        res.close()
+      }
+    }
+  }
+
   def toTransportMessage(message: Any): Any = {
     message match {
       case legacy: Message =>
@@ -1010,117 +1070,6 @@ object Utils extends Logging {
     }
   }
 
-  def toStatusCode(status: Int): StatusCode = {
-    status match {
-      case 0 =>
-        StatusCode.SUCCESS
-      case 1 =>
-        StatusCode.PARTIAL_SUCCESS
-      case 2 =>
-        StatusCode.REQUEST_FAILED
-      case 3 =>
-        StatusCode.SHUFFLE_ALREADY_REGISTERED
-      case 4 =>
-        StatusCode.SHUFFLE_NOT_REGISTERED
-      case 5 =>
-        StatusCode.RESERVE_SLOTS_FAILED
-      case 6 =>
-        StatusCode.SLOT_NOT_AVAILABLE
-      case 7 =>
-        StatusCode.WORKER_NOT_FOUND
-      case 8 =>
-        StatusCode.PARTITION_NOT_FOUND
-      case 9 =>
-        StatusCode.REPLICA_PARTITION_NOT_FOUND
-      case 10 =>
-        StatusCode.DELETE_FILES_FAILED
-      case 11 =>
-        StatusCode.PARTITION_EXISTS
-      case 12 =>
-        StatusCode.REVIVE_FAILED
-      case 13 =>
-        StatusCode.REPLICATE_DATA_FAILED
-      case 14 =>
-        StatusCode.NUM_MAPPER_ZERO
-      case 15 =>
-        StatusCode.MAP_ENDED
-      case 16 =>
-        StatusCode.STAGE_ENDED
-      case 17 =>
-        StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_PRIMARY
-      case 18 =>
-        StatusCode.PUSH_DATA_WRITE_FAIL_REPLICA
-      case 19 =>
-        StatusCode.PUSH_DATA_WRITE_FAIL_PRIMARY
-      case 20 =>
-        StatusCode.PUSH_DATA_FAIL_PARTITION_NOT_FOUND
-      case 21 =>
-        StatusCode.HARD_SPLIT
-      case 22 =>
-        StatusCode.SOFT_SPLIT
-      case 23 =>
-        StatusCode.STAGE_END_TIME_OUT
-      case 24 =>
-        StatusCode.SHUFFLE_DATA_LOST
-      case 25 =>
-        StatusCode.WORKER_SHUTDOWN
-      case 26 =>
-        StatusCode.NO_AVAILABLE_WORKING_DIR
-      case 27 =>
-        StatusCode.WORKER_EXCLUDED
-      case 28 =>
-        StatusCode.WORKER_UNKNOWN
-      case 29 =>
-        StatusCode.COMMIT_FILE_EXCEPTION
-      case 30 =>
-        StatusCode.PUSH_DATA_SUCCESS_PRIMARY_CONGESTED
-      case 31 =>
-        StatusCode.PUSH_DATA_SUCCESS_REPLICA_CONGESTED
-      case 32 =>
-        StatusCode.PUSH_DATA_HANDSHAKE_FAIL_REPLICA
-      case 33 =>
-        StatusCode.PUSH_DATA_HANDSHAKE_FAIL_PRIMARY
-      case 34 =>
-        StatusCode.REGION_START_FAIL_REPLICA
-      case 35 =>
-        StatusCode.REGION_START_FAIL_PRIMARY
-      case 36 =>
-        StatusCode.REGION_FINISH_FAIL_REPLICA
-      case 37 =>
-        StatusCode.REGION_FINISH_FAIL_PRIMARY
-      case 38 =>
-        StatusCode.PUSH_DATA_CREATE_CONNECTION_FAIL_PRIMARY
-      case 39 =>
-        StatusCode.PUSH_DATA_CREATE_CONNECTION_FAIL_REPLICA
-      case 40 =>
-        StatusCode.PUSH_DATA_CONNECTION_EXCEPTION_PRIMARY
-      case 41 =>
-        StatusCode.PUSH_DATA_CONNECTION_EXCEPTION_REPLICA
-      case 42 =>
-        StatusCode.PUSH_DATA_TIMEOUT_PRIMARY
-      case 43 =>
-        StatusCode.PUSH_DATA_TIMEOUT_REPLICA
-      case 44 =>
-        StatusCode.PUSH_DATA_PRIMARY_WORKER_EXCLUDED
-      case 45 =>
-        StatusCode.PUSH_DATA_REPLICA_WORKER_EXCLUDED
-      case 46 =>
-        StatusCode.FETCH_DATA_TIMEOUT
-      case 47 =>
-        StatusCode.REVIVE_INITIALIZED
-      case 48 =>
-        StatusCode.DESTROY_SLOTS_MOCK_FAILURE
-      case 49 =>
-        StatusCode.COMMIT_FILES_MOCK_FAILURE
-      case 50 =>
-        StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_REPLICA
-      case 51 =>
-        StatusCode.OPEN_STREAM_FAILED
-      case _ =>
-        null
-    }
-  }
-
   def toShuffleSplitMode(mode: Int): PartitionSplitMode = {
     mode match {
       case 0 => PartitionSplitMode.SOFT
@@ -1164,7 +1113,9 @@ object Utils extends Logging {
   val SORTED_SUFFIX = ".sorted"
   val INDEX_SUFFIX = ".index"
   val SUFFIX_HDFS_WRITE_SUCCESS = ".success"
-  val COMPATIBLE_HDFS_REGEX = "^[a-zA-Z0-9]+://.*"
+  val COMPATIBLE_HDFS_REGEX = "^(?!s3://)(?!s3a://)(?!oss://)[a-zA-Z0-9]+://.*"
+  val S3_REGEX = "^s3[a]?://([a-z0-9][a-z0-9-]{1,61}[a-z0-9])(/.*)?$"
+  val OSS_REGEX = "^oss?://([a-z0-9][a-z0-9-]{1,61}[a-z0-9])(/.*)?$"
 
   val UNKNOWN_APP_SHUFFLE_ID = -1
 
@@ -1172,6 +1123,14 @@ object Utils extends Logging {
     path.matches(COMPATIBLE_HDFS_REGEX)
   }
 
+  def isS3Path(path: String): Boolean = {
+    path.matches(S3_REGEX)
+  }
+
+  def isOssPath(path: String): Boolean = {
+    path.matches(OSS_REGEX)
+  }
+
   def getSortedFilePath(path: String): String = {
     path + SORTED_SUFFIX
   }
@@ -1286,6 +1245,23 @@ object Utils extends Logging {
     }.asInstanceOf[Seq[(K, V)]]
   }
 
+  def instantiate[T](className: String)(implicit tag: scala.reflect.ClassTag[T]): T = {
+    logDebug(s"Creating instance of $className")
+    val clazz = Class.forName(
+      className,
+      true,
+      Thread.currentThread().getContextClassLoader).asInstanceOf[Class[T]]
+    try {
+      val ctor = clazz.getDeclaredConstructor()
+      logDebug("Using no-arg constructor")
+      ctor.newInstance()
+    } catch {
+      case e: NoSuchMethodException =>
+        logError(s"Failed to instantiate class $className", e)
+        throw e
+    }
+  }
+
   def isCriticalCauseForFetch(e: Exception) = {
     val rpcTimeout =
       e.isInstanceOf[IOException] && e.getCause != null && e.getCause.isInstanceOf[TimeoutException]
@@ -1297,4 +1273,44 @@ object Utils extends Logging {
     connectException || rpcTimeout || fetchChunkTimeout
   }
 
+  def makeOpenStreamRequestId(
+      shuffleKey: String,
+      clientChannelId: String,
+      rpcRequestId: Long): String = {
+    s"$shuffleKey-$clientChannelId-$rpcRequestId"
+  }
+
+  def withRetryOnTimeoutOrIOException[T](numRetries: Int, retryWait: Long)(block: => T): T = {
+    var retriesLeft = numRetries
+
+    def waitOrThrow(e: Throwable): Unit = {
+      if (retriesLeft > 0) {
+        val retryWaitMs = new Random().nextInt(retryWait.toInt)
+        try {
+          TimeUnit.MILLISECONDS.sleep(retryWaitMs)
+        } catch {
+          case _: InterruptedException =>
+            throw e
+        }
+      } else {
+        throw e
+      }
+    }
+
+    while (retriesLeft >= 0) {
+      retriesLeft -= 1
+      try {
+        return block
+      } catch {
+        case e: CelebornException if e.getCause != null && e.getCause.isInstanceOf[IOException] =>
+          waitOrThrow(e)
+        case e @ (_: RpcTimeoutException | _: IOException) =>
+          waitOrThrow(e)
+        case e: Exception =>
+          throw e
+      }
+    }
+    // should never be here
+    null.asInstanceOf[T]
+  }
 }
