@@ -22,18 +22,16 @@ import java.io.File;
 import java.io.FileInputStream;
 import java.io.IOException;
 import java.nio.file.Files;
-import java.util.List;
-import java.util.Map;
-import java.util.Optional;
-import java.util.Set;
+import java.util.*;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.atomic.AtomicLong;
-import java.util.concurrent.atomic.AtomicReference;
 import java.util.concurrent.atomic.LongAdder;
 import java.util.stream.Collectors;
 
 import scala.Option;
+import scala.Tuple2;
 
+import com.google.common.annotations.VisibleForTesting;
 import org.apache.hadoop.net.NetworkTopology;
 import org.apache.hadoop.net.Node;
 import org.slf4j.Logger;
@@ -41,8 +39,6 @@ import org.slf4j.LoggerFactory;
 
 import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.common.identity.UserIdentifier;
-import org.apache.celeborn.common.meta.AppDiskUsageMetric;
-import org.apache.celeborn.common.meta.AppDiskUsageSnapShot;
 import org.apache.celeborn.common.meta.ApplicationMeta;
 import org.apache.celeborn.common.meta.DiskInfo;
 import org.apache.celeborn.common.meta.DiskStatus;
@@ -59,13 +55,22 @@ import org.apache.celeborn.common.util.PbSerDeUtils;
 import org.apache.celeborn.common.util.Utils;
 import org.apache.celeborn.common.util.WorkerStatusUtils;
 
+/**
+ * Note: Do not update the worker collections directly from outside the metadata manager, especially
+ * {@link #workersMap}, {@link #workerEventInfos}, {@link #shutdownWorkers}, {@link
+ * #excludedWorkers}, {@link #manuallyExcludedWorkers}, {@link #availableWorkers}.
+ *
+ * <p>All updates should be done through the provided methods to ensure consistency.
+ */
 public abstract class AbstractMetaManager implements IMetadataHandler {
   private static final Logger LOG = LoggerFactory.getLogger(AbstractMetaManager.class);
 
   // Metadata for master service
-  public final Set<String> registeredShuffle = ConcurrentHashMap.newKeySet();
+  public final Map<String, Set<Integer>> registeredAppAndShuffles =
+      JavaUtils.newConcurrentHashMap();
   public final Set<String> hostnameSet = ConcurrentHashMap.newKeySet();
-  public final Set<WorkerInfo> workers = ConcurrentHashMap.newKeySet();
+  public final Map<String, WorkerInfo> workersMap = JavaUtils.newConcurrentHashMap();
+  public final Set<WorkerInfo> availableWorkers = ConcurrentHashMap.newKeySet();
 
   public final ConcurrentHashMap<WorkerInfo, Long> lostWorkers = JavaUtils.newConcurrentHashMap();
   public final ConcurrentHashMap<WorkerInfo, WorkerEventInfo> workerEventInfos =
@@ -74,6 +79,7 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
   public final Set<WorkerInfo> excludedWorkers = ConcurrentHashMap.newKeySet();
   public final Set<WorkerInfo> manuallyExcludedWorkers = ConcurrentHashMap.newKeySet();
   public final Set<WorkerInfo> shutdownWorkers = ConcurrentHashMap.newKeySet();
+  public final Set<WorkerInfo> decommissionWorkers = ConcurrentHashMap.newKeySet();
   public final Set<WorkerInfo> workerLostEvents = ConcurrentHashMap.newKeySet();
 
   protected RpcEnv rpcEnv;
@@ -82,18 +88,25 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
 
   public long initialEstimatedPartitionSize;
   public long estimatedPartitionSize;
+  public double unhealthyDiskRatioThreshold;
   public final LongAdder partitionTotalWritten = new LongAdder();
   public final LongAdder partitionTotalFileCount = new LongAdder();
-  public AppDiskUsageMetric appDiskUsageMetric = null;
+  public final LongAdder shuffleTotalCount = new LongAdder();
+  public final LongAdder applicationTotalCount = new LongAdder();
+  public final Map<String, Long> shuffleFallbackCounts = JavaUtils.newConcurrentHashMap();
+  public final Map<String, Long> applicationFallbackCounts = JavaUtils.newConcurrentHashMap();
 
   public final ConcurrentHashMap<String, ApplicationMeta> applicationMetas =
       JavaUtils.newConcurrentHashMap();
 
   public void updateRequestSlotsMeta(
       String shuffleKey, String hostName, Map<String, Map<String, Integer>> workerWithAllocations) {
-    registeredShuffle.add(shuffleKey);
+    Tuple2<String, Object> appIdShuffleId = Utils.splitShuffleKey(shuffleKey);
+    registeredAppAndShuffles
+        .computeIfAbsent(appIdShuffleId._1(), v -> new HashSet<>())
+        .add((Integer) appIdShuffleId._2);
 
-    String appId = Utils.splitShuffleKey(shuffleKey)._1;
+    String appId = appIdShuffleId._1;
     appHeartbeatTime.compute(
         appId,
         (applicationId, oldTimestamp) -> {
@@ -110,25 +123,90 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
   }
 
   public void updateUnregisterShuffleMeta(String shuffleKey) {
-    registeredShuffle.remove(shuffleKey);
+    Tuple2<String, Object> appIdShuffleId = Utils.splitShuffleKey(shuffleKey);
+    Set<Integer> shuffleIds = registeredAppAndShuffles.get(appIdShuffleId._1());
+    if (shuffleIds != null) {
+      shuffleIds.remove(appIdShuffleId._2);
+      registeredAppAndShuffles.compute(
+          appIdShuffleId._1(),
+          (s, shuffles) -> {
+            if (shuffles.size() == 0) {
+              return null;
+            }
+            return shuffles;
+          });
+    }
   }
 
-  public void updateAppHeartbeatMeta(String appId, long time, long totalWritten, long fileCount) {
+  public void updateBatchUnregisterShuffleMeta(List<String> shuffleKeys) {
+    for (String shuffleKey : shuffleKeys) {
+      Tuple2<String, Object> appIdShuffleId = Utils.splitShuffleKey(shuffleKey);
+      String appId = appIdShuffleId._1;
+      if (registeredAppAndShuffles.containsKey(appId)) {
+        registeredAppAndShuffles.get(appId).remove(appIdShuffleId._2);
+      }
+    }
+  }
+
+  public void updateAppHeartbeatMeta(
+      String appId,
+      long time,
+      long totalWritten,
+      long fileCount,
+      long shuffleCount,
+      long applicationCount,
+      Map<String, Long> shuffleFallbackCounts,
+      Map<String, Long> applicationFallbackCounts) {
     appHeartbeatTime.put(appId, time);
     partitionTotalWritten.add(totalWritten);
     partitionTotalFileCount.add(fileCount);
+    shuffleTotalCount.add(shuffleCount);
+    applicationTotalCount.add(applicationCount);
+    addFallbackCounts(this.shuffleFallbackCounts, shuffleFallbackCounts);
+    addFallbackCounts(this.applicationFallbackCounts, applicationFallbackCounts);
   }
 
   public void updateAppLostMeta(String appId) {
-    registeredShuffle.removeIf(shuffleKey -> shuffleKey.startsWith(appId));
+    registeredAppAndShuffles.remove(appId);
     appHeartbeatTime.remove(appId);
     applicationMetas.remove(appId);
   }
 
-  public void updateWorkerExcludeMeta(
+  @VisibleForTesting
+  public void updateExcludedWorkersMeta(
       List<WorkerInfo> workersToAdd, List<WorkerInfo> workersToRemove) {
-    manuallyExcludedWorkers.addAll(workersToAdd);
-    workersToRemove.forEach(manuallyExcludedWorkers::remove);
+    workersToAdd.forEach(
+        worker -> {
+          excludedWorkers.add(worker);
+          availableWorkers.remove(worker);
+        });
+    workersToRemove.forEach(
+        worker -> {
+          excludedWorkers.remove(worker);
+          updateAvailableWorkers(worker);
+        });
+  }
+
+  public void updateManuallyExcludedWorkersMeta(
+      List<WorkerInfo> workersToAdd, List<WorkerInfo> workersToRemove) {
+    workersToAdd.forEach(
+        worker -> {
+          manuallyExcludedWorkers.add(worker);
+          availableWorkers.remove(worker);
+        });
+    workersToRemove.forEach(
+        worker -> {
+          manuallyExcludedWorkers.remove(worker);
+          updateAvailableWorkers(worker);
+        });
+  }
+
+  public void reviseLostShuffles(String appId, List<Integer> lostShuffles) {
+    registeredAppAndShuffles.computeIfAbsent(appId, v -> new HashSet<>()).addAll(lostShuffles);
+  }
+
+  public void deleteApp(String appId) {
+    registeredAppAndShuffles.remove(appId);
   }
 
   public void updateWorkerLostMeta(
@@ -136,9 +214,10 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
     WorkerInfo worker = new WorkerInfo(host, rpcPort, pushPort, fetchPort, replicatePort);
     workerLostEvents.add(worker);
     // remove worker from workers
-    synchronized (workers) {
-      workers.remove(worker);
+    synchronized (workersMap) {
+      workersMap.remove(worker.toUniqueId());
       lostWorkers.put(worker, System.currentTimeMillis());
+      availableWorkers.remove(worker);
     }
     excludedWorkers.remove(worker);
     workerLostEvents.remove(worker);
@@ -148,20 +227,23 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
       String host, int rpcPort, int pushPort, int fetchPort, int replicatePort) {
     WorkerInfo worker = new WorkerInfo(host, rpcPort, pushPort, fetchPort, replicatePort);
     // remove worker from workers
-    synchronized (workers) {
-      workers.remove(worker);
+    synchronized (workersMap) {
+      workersMap.remove(worker.toUniqueId());
       lostWorkers.put(worker, System.currentTimeMillis());
+      availableWorkers.remove(worker);
     }
     excludedWorkers.remove(worker);
   }
 
   public void removeWorkersUnavailableInfoMeta(List<WorkerInfo> unavailableWorkers) {
-    synchronized (workers) {
+    synchronized (workersMap) {
       for (WorkerInfo workerInfo : unavailableWorkers) {
         if (lostWorkers.containsKey(workerInfo)) {
           lostWorkers.remove(workerInfo);
           shutdownWorkers.remove(workerInfo);
           workerEventInfos.remove(workerInfo);
+          decommissionWorkers.remove(workerInfo);
+          updateAvailableWorkers(workerInfo);
         }
       }
     }
@@ -174,22 +256,18 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
       int fetchPort,
       int replicatePort,
       Map<String, DiskInfo> disks,
-      Map<UserIdentifier, ResourceConsumption> userResourceConsumption,
-      Map<String, Long> estimatedAppDiskUsage,
       long time,
       WorkerStatus workerStatus,
       boolean highWorkload) {
     WorkerInfo worker =
-        new WorkerInfo(
-            host, rpcPort, pushPort, fetchPort, replicatePort, -1, disks, userResourceConsumption);
+        new WorkerInfo(host, rpcPort, pushPort, fetchPort, replicatePort, -1, disks, null);
     AtomicLong availableSlots = new AtomicLong();
     LOG.debug("update worker {}:{} heartbeat {}", host, rpcPort, disks);
-    synchronized (workers) {
-      Optional<WorkerInfo> workerInfo = workers.stream().filter(w -> w.equals(worker)).findFirst();
+    synchronized (workersMap) {
+      Optional<WorkerInfo> workerInfo = Optional.ofNullable(workersMap.get(worker.toUniqueId()));
       workerInfo.ifPresent(
           info -> {
             info.updateThenGetDiskInfos(disks, Option.apply(estimatedPartitionSize));
-            info.updateThenGetUserResourceConsumption(userResourceConsumption);
             availableSlots.set(info.totalAvailableSlots());
             info.lastHeartbeat_$eq(time);
             info.setWorkerStatus(workerStatus);
@@ -205,18 +283,32 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
       }
     }
 
-    appDiskUsageMetric.update(estimatedAppDiskUsage);
     // If using HDFSONLY mode, workers with empty disks should not be put into excluded worker list.
-    long healthyDiskNum =
-        disks.values().stream().filter(s -> s.status().equals(DiskStatus.HEALTHY)).count();
+    long unhealthyDiskNum =
+        disks.values().stream().filter(s -> !s.status().equals(DiskStatus.HEALTHY)).count();
+    boolean exceed = unhealthyDiskNum * 1.0 / disks.size() >= unhealthyDiskRatioThreshold;
     if (!excludedWorkers.contains(worker)
-        && (((disks.isEmpty() || healthyDiskNum <= 0) && !conf.hasHDFSStorage()) || highWorkload)) {
-      LOG.debug("Worker: {} num total slots is 0, add to excluded list", worker);
+        && (((disks.isEmpty() || exceed)
+                && !conf.hasHDFSStorage()
+                && !conf.hasS3Storage()
+                && !conf.hasOssStorage())
+            || highWorkload)) {
+      LOG.warn(
+          "Worker {} (unhealthy disks num: {}) adds to excluded workers", worker, unhealthyDiskNum);
       excludedWorkers.add(worker);
-    } else if ((availableSlots.get() > 0 || conf.hasHDFSStorage()) && !highWorkload) {
+    } else if ((availableSlots.get() > 0
+            || conf.hasHDFSStorage()
+            || conf.hasS3Storage()
+            || conf.hasOssStorage())
+        && !highWorkload) {
       // only unblack if numSlots larger than 0
       excludedWorkers.remove(worker);
     }
+
+    // try to update the available workers if the worker status is Normal
+    if (workerStatus.getState() == PbWorkerStatus.State.Normal) {
+      updateAvailableWorkers(worker);
+    }
   }
 
   public void updateRegisterWorkerMeta(
@@ -227,8 +319,7 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
       int replicatePort,
       int internalPort,
       String networkLocation,
-      Map<String, DiskInfo> disks,
-      Map<UserIdentifier, ResourceConsumption> userResourceConsumption) {
+      Map<String, DiskInfo> disks) {
     WorkerInfo workerInfo =
         new WorkerInfo(
             host,
@@ -238,7 +329,7 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
             replicatePort,
             internalPort,
             disks,
-            userResourceConsumption);
+            new HashMap<>());
     workerInfo.lastHeartbeat_$eq(System.currentTimeMillis());
     if (networkLocation != null
         && !networkLocation.isEmpty()
@@ -247,15 +338,15 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
     } else {
       workerInfo.networkLocation_$eq(rackResolver.resolve(host).getNetworkLocation());
     }
-    workerInfo.updateDiskMaxSlots(estimatedPartitionSize);
-    synchronized (workers) {
-      if (!workers.contains(workerInfo)) {
-        workers.add(workerInfo);
-      }
+    workerInfo.updateDiskSlots(estimatedPartitionSize);
+    synchronized (workersMap) {
+      workersMap.putIfAbsent(workerInfo.toUniqueId(), workerInfo);
       shutdownWorkers.remove(workerInfo);
       lostWorkers.remove(workerInfo);
       excludedWorkers.remove(workerInfo);
       workerEventInfos.remove(workerInfo);
+      decommissionWorkers.remove(workerInfo);
+      updateAvailableWorkers(workerInfo);
     }
   }
 
@@ -269,21 +360,24 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
     byte[] snapshotBytes =
         PbSerDeUtils.toPbSnapshotMetaInfo(
                 estimatedPartitionSize,
-                registeredShuffle,
+                registeredAppAndShuffles,
                 hostnameSet,
                 excludedWorkers,
                 manuallyExcludedWorkers,
                 workerLostEvents,
                 appHeartbeatTime,
-                workers,
+                new HashSet(workersMap.values()),
                 partitionTotalWritten.sum(),
                 partitionTotalFileCount.sum(),
-                appDiskUsageMetric.snapShots(),
-                appDiskUsageMetric.currentSnapShot().get(),
+                shuffleTotalCount.sum(),
+                applicationTotalCount.sum(),
+                shuffleFallbackCounts,
+                applicationFallbackCounts,
                 lostWorkers,
                 shutdownWorkers,
                 workerEventInfos,
-                applicationMetas)
+                applicationMetas,
+                decommissionWorkers)
             .toByteArray();
     Files.write(file.toPath(), snapshotBytes);
   }
@@ -301,7 +395,12 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
 
       estimatedPartitionSize = snapshotMetaInfo.getEstimatedPartitionSize();
 
-      registeredShuffle.addAll(snapshotMetaInfo.getRegisteredShuffleList());
+      for (String shuffleKey : snapshotMetaInfo.getRegisteredShuffleList()) {
+        Tuple2<String, Object> appIdShuffleId = Utils.splitShuffleKey(shuffleKey);
+        registeredAppAndShuffles
+            .computeIfAbsent(appIdShuffleId._1, v -> new HashSet<>())
+            .add((Integer) appIdShuffleId._2);
+      }
       hostnameSet.addAll(snapshotMetaInfo.getHostnameSetList());
       excludedWorkers.addAll(
           snapshotMetaInfo.getExcludedWorkersList().stream()
@@ -317,9 +416,8 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
               .collect(Collectors.toSet()));
       appHeartbeatTime.putAll(snapshotMetaInfo.getAppHeartbeatTimeMap());
 
-      registeredShuffle.forEach(
-          shuffleKey -> {
-            String appId = Utils.splitShuffleKey(shuffleKey)._1;
+      registeredAppAndShuffles.forEach(
+          (appId, shuffleId) -> {
             if (!appHeartbeatTime.containsKey(appId)) {
               appHeartbeatTime.put(appId, System.currentTimeMillis());
             }
@@ -336,7 +434,7 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
               .collect(Collectors.toList());
       scala.collection.immutable.Map<String, Node> resolveMap =
           rackResolver.resolveToMap(workerHostList);
-      workers.addAll(
+      workersMap.putAll(
           workerInfoSet.stream()
               .peek(
                   workerInfo -> {
@@ -346,7 +444,7 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
                           resolveMap.get(workerInfo.host()).get().getNetworkLocation());
                     }
                   })
-              .collect(Collectors.toSet()));
+              .collect(Collectors.toMap(WorkerInfo::toUniqueId, w -> w)));
 
       snapshotMetaInfo
           .getLostWorkersMap()
@@ -366,54 +464,69 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
               .map(PbSerDeUtils::fromPbWorkerInfo)
               .collect(Collectors.toSet()));
 
+      decommissionWorkers.addAll(
+          snapshotMetaInfo.getDecommissionWorkersList().stream()
+              .map(PbSerDeUtils::fromPbWorkerInfo)
+              .collect(Collectors.toSet()));
+
       partitionTotalWritten.add(snapshotMetaInfo.getPartitionTotalWritten());
       partitionTotalFileCount.add(snapshotMetaInfo.getPartitionTotalFileCount());
-      appDiskUsageMetric.restoreFromSnapshot(
-          snapshotMetaInfo.getAppDiskUsageMetricSnapshotsList().stream()
-              .map(PbSerDeUtils::fromPbAppDiskUsageSnapshot)
-              .toArray(AppDiskUsageSnapShot[]::new));
-      appDiskUsageMetric.currentSnapShot_$eq(
-          new AtomicReference<AppDiskUsageSnapShot>(
-              PbSerDeUtils.fromPbAppDiskUsageSnapshot(
-                  snapshotMetaInfo.getCurrentAppDiskUsageMetricsSnapshot())));
+      shuffleTotalCount.add(snapshotMetaInfo.getShuffleTotalCount());
+      applicationTotalCount.add(snapshotMetaInfo.getApplicationTotalCount());
+      addFallbackCounts(shuffleFallbackCounts, snapshotMetaInfo.getShuffleFallbackCountsMap());
+      addFallbackCounts(
+          applicationFallbackCounts, snapshotMetaInfo.getApplicationFallbackCountsMap());
 
       snapshotMetaInfo
           .getApplicationMetasMap()
           .forEach(
               (key, value) -> applicationMetas.put(key, PbSerDeUtils.fromPbApplicationMeta(value)));
+
+      availableWorkers.addAll(
+          workersMap.values().stream()
+              .filter(worker -> isWorkerAvailable(worker))
+              .collect(Collectors.toSet()));
     } catch (Exception e) {
       throw new IOException(e);
     }
     LOG.info("Successfully restore meta info from snapshot {}", file.getAbsolutePath());
     LOG.info(
         "Worker size: {}, Registered shuffle size: {}. Worker excluded list size: {}. Manually Excluded list size: {}",
-        workers.size(),
-        registeredShuffle.size(),
+        workersMap.size(),
+        registeredAppAndShuffles.size(),
         excludedWorkers.size(),
         manuallyExcludedWorkers.size());
-    workers.forEach(workerInfo -> LOG.info(workerInfo.toString()));
-    registeredShuffle.forEach(shuffle -> LOG.info("RegisteredShuffle {}", shuffle));
+    workersMap.values().forEach(workerInfo -> LOG.info(workerInfo.toString()));
+    registeredAppAndShuffles.forEach(
+        (appId, shuffleId) -> LOG.info("RegisteredShuffle {}-{}", appId, shuffleId));
   }
 
   private void cleanUpState() {
-    registeredShuffle.clear();
+    registeredAppAndShuffles.clear();
     hostnameSet.clear();
-    workers.clear();
+    workersMap.clear();
+    availableWorkers.clear();
     lostWorkers.clear();
     appHeartbeatTime.clear();
     excludedWorkers.clear();
     shutdownWorkers.clear();
+    decommissionWorkers.clear();
     manuallyExcludedWorkers.clear();
     workerLostEvents.clear();
     partitionTotalWritten.reset();
     partitionTotalFileCount.reset();
+    shuffleTotalCount.reset();
+    applicationTotalCount.reset();
+    shuffleFallbackCounts.clear();
+    applicationFallbackCounts.clear();
     workerEventInfos.clear();
     applicationMetas.clear();
   }
 
   public void updateMetaByReportWorkerUnavailable(List<WorkerInfo> failedWorkers) {
-    synchronized (this.workers) {
+    synchronized (this.workersMap) {
       shutdownWorkers.addAll(failedWorkers);
+      availableWorkers.removeAll(failedWorkers);
     }
   }
 
@@ -421,21 +534,30 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
     long eventTime = System.currentTimeMillis();
     ResourceProtos.WorkerEventType eventType =
         ResourceProtos.WorkerEventType.forNumber(workerEventTypeValue);
-    synchronized (this.workers) {
+    synchronized (this.workersMap) {
       for (WorkerInfo workerInfo : workerInfoList) {
         WorkerEventInfo workerEventInfo = workerEventInfos.get(workerInfo);
         LOG.info("Received worker event: {} for worker: {}", eventType, workerInfo.toUniqueId());
         if (workerEventInfo == null || !workerEventInfo.isSameEvent(eventType.getNumber())) {
           if (eventType == ResourceProtos.WorkerEventType.None) {
             workerEventInfos.remove(workerInfo);
+            updateAvailableWorkers(workerInfo);
           } else {
             workerEventInfos.put(workerInfo, new WorkerEventInfo(eventType.getNumber(), eventTime));
+            availableWorkers.remove(workerInfo);
           }
         }
       }
     }
   }
 
+  public void updateMetaByReportWorkerDecommission(List<WorkerInfo> workers) {
+    synchronized (this.workersMap) {
+      decommissionWorkers.addAll(workers);
+      availableWorkers.removeAll(workers);
+    }
+  }
+
   public void updatePartitionSize() {
     long oldEstimatedPartitionSize = estimatedPartitionSize;
     long tmpTotalWritten = partitionTotalWritten.sumThenReset();
@@ -452,23 +574,40 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
     } else {
       estimatedPartitionSize = initialEstimatedPartitionSize;
     }
+
+    // Do not trigger update is estimated partition size value is unchanged
+    if (estimatedPartitionSize == oldEstimatedPartitionSize) {
+      return;
+    }
+
     LOG.warn(
         "Celeborn cluster estimated partition size changed from {} to {}",
         Utils.bytesToString(oldEstimatedPartitionSize),
         Utils.bytesToString(estimatedPartitionSize));
-    workers.stream()
-        .filter(
-            worker ->
-                !excludedWorkers.contains(worker) && !manuallyExcludedWorkers.contains(worker))
-        .forEach(workerInfo -> workerInfo.updateDiskMaxSlots(estimatedPartitionSize));
+
+    HashSet<WorkerInfo> workers = new HashSet(workersMap.values());
+    excludedWorkers.forEach(workers::remove);
+    manuallyExcludedWorkers.forEach(workers::remove);
+    workers.forEach(workerInfo -> workerInfo.updateDiskSlots(estimatedPartitionSize));
   }
 
-  public boolean isWorkerAvailable(WorkerInfo workerInfo) {
-    return !excludedWorkers.contains(workerInfo)
+  private boolean isWorkerAvailable(WorkerInfo workerInfo) {
+    return (workerInfo.getWorkerStatus().getState() == PbWorkerStatus.State.Normal
+            && !workerEventInfos.containsKey(workerInfo))
+        && !excludedWorkers.contains(workerInfo)
         && !shutdownWorkers.contains(workerInfo)
-        && !manuallyExcludedWorkers.contains(workerInfo)
-        && (!workerEventInfos.containsKey(workerInfo)
-            && workerInfo.getWorkerStatus().getState() == PbWorkerStatus.State.Normal);
+        && !manuallyExcludedWorkers.contains(workerInfo);
+  }
+
+  private void updateAvailableWorkers(WorkerInfo worker) {
+    synchronized (workersMap) {
+      Optional<WorkerInfo> workerInfo = Optional.ofNullable(workersMap.get(worker.toUniqueId()));
+      if (workerInfo.map(this::isWorkerAvailable).orElse(false)) {
+        availableWorkers.add(workerInfo.get());
+      } else {
+        availableWorkers.remove(worker);
+      }
+    }
   }
 
   public void updateApplicationMeta(ApplicationMeta applicationMeta) {
@@ -478,4 +617,30 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
   public void removeApplicationMeta(String appId) {
     applicationMetas.remove(appId);
   }
+
+  public int registeredShuffleCount() {
+    return registeredAppAndShuffles.values().stream().mapToInt(Set::size).sum();
+  }
+
+  private void addFallbackCounts(Map<String, Long> fallbackCounts, Map<String, Long> counts) {
+    for (String fallbackPolicy : counts.keySet()) {
+      fallbackCounts.compute(
+          fallbackPolicy, (k, v) -> v == null ? counts.get(k) : v + counts.get(k));
+    }
+  }
+
+  public void updateWorkerResourceConsumptions(
+      String host,
+      int rpcPort,
+      int pushPort,
+      int fetchPort,
+      int replicatePort,
+      Map<UserIdentifier, ResourceConsumption> resourceConsumptions) {
+    WorkerInfo worker =
+        new WorkerInfo(host, rpcPort, pushPort, fetchPort, replicatePort, -1, null, null);
+    synchronized (workersMap) {
+      Optional<WorkerInfo> workerInfo = Optional.ofNullable(workersMap.get(worker.toUniqueId()));
+      workerInfo.ifPresent(info -> info.updateThenGetUserResourceConsumption(resourceConsumptions));
+    }
+  }
 }
