@@ -18,10 +18,13 @@
 package org.apache.celeborn.server.common.service.store;
 
 import java.util.List;
+import java.util.Map;
 
-import org.apache.celeborn.server.common.service.config.SystemConfig;
+import org.apache.celeborn.server.common.service.config.ConfigLevel;
 import org.apache.celeborn.server.common.service.config.TenantConfig;
 import org.apache.celeborn.server.common.service.model.ClusterInfo;
+import org.apache.celeborn.server.common.service.model.ClusterSystemConfig;
+import org.apache.celeborn.server.common.service.model.ClusterTag;
 
 public interface IServiceManager {
 
@@ -33,5 +36,17 @@ public interface IServiceManager {
 
   List<TenantConfig> getAllTenantUserConfigs();
 
-  SystemConfig getSystemConfig();
+  List<ClusterSystemConfig> getSystemConfig();
+
+  List<ClusterTag> getClusterTags();
+
+  void upsertSystemConfig(Map<String, String> systemConfigs);
+
+  void upsertTenantConfig(
+      ConfigLevel configLevel, String tenantId, String name, Map<String, String> tenantConfigs);
+
+  void deleteSystemConfigByKeys(List<String> configKeys);
+
+  void deleteTenantConfigByKeys(
+      ConfigLevel configLevel, String tenantId, String name, List<String> configKeys);
 }
