@@ -21,8 +21,8 @@ import java.io.File
 
 import scala.collection.JavaConverters._
 
-import org.apache.flink.api.common.{ExecutionMode, RuntimeExecutionMode}
-import org.apache.flink.configuration.{Configuration, ExecutionOptions, RestOptions}
+import org.apache.flink.api.common.RuntimeExecutionMode
+import org.apache.flink.configuration.{Configuration, ExecutionOptions}
 import org.apache.flink.runtime.jobgraph.JobType
 import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment
 import org.apache.flink.streaming.api.graph.GlobalStreamExchangeMode
@@ -32,6 +32,7 @@ import org.scalatest.funsuite.AnyFunSuite
 import org.apache.celeborn.common.CelebornConf
 import org.apache.celeborn.common.CelebornConf.{AUTH_ENABLED, INTERNAL_PORT_ENABLED}
 import org.apache.celeborn.common.internal.Logging
+import org.apache.celeborn.common.protocol.FallbackPolicy
 import org.apache.celeborn.service.deploy.MiniClusterFeature
 import org.apache.celeborn.service.deploy.worker.Worker
 
@@ -65,6 +66,14 @@ abstract class WordCountTestBase extends AnyFunSuite with Logging with MiniClust
   }
 
   test(getClass.getName + ": celeborn flink integration test - word count") {
+    testWordCount(FallbackPolicy.AUTO)
+  }
+
+  test(getClass.getName + ": celeborn flink integration test with fallback - word count") {
+    testWordCount(FallbackPolicy.ALWAYS)
+  }
+
+  private def testWordCount(fallbackPolicy: FallbackPolicy): Unit = {
     // set up execution environment
     val configuration = new Configuration
     val parallelism = 8
@@ -72,10 +81,11 @@ abstract class WordCountTestBase extends AnyFunSuite with Logging with MiniClust
       "shuffle-service-factory.class",
       "org.apache.celeborn.plugin.flink.RemoteShuffleServiceFactory")
     configuration.setString("celeborn.master.endpoints", s"localhost:$port")
+    configuration.setString("celeborn.client.flink.shuffle.fallback.policy", fallbackPolicy.name())
     configuration.setString("execution.batch-shuffle-mode", "ALL_EXCHANGES_BLOCKING")
     configuration.set(ExecutionOptions.RUNTIME_MODE, RuntimeExecutionMode.BATCH)
     configuration.setString("taskmanager.memory.network.min", "1024m")
-    configuration.setString(RestOptions.BIND_PORT, "8081-8089")
+    configuration.setString("rest.bind-port", "8081-8099")
     configuration.setString(
       "execution.batch.adaptive.auto-parallelism.min-parallelism",
       "" + parallelism)
@@ -84,7 +94,6 @@ abstract class WordCountTestBase extends AnyFunSuite with Logging with MiniClust
     configuration.setString("restart-strategy.fixed-delay.delay", "5s")
     addClientConf(configuration)
     val env = StreamExecutionEnvironment.createLocalEnvironmentWithWebUI(configuration)
-    env.getConfig.setExecutionMode(ExecutionMode.BATCH)
     env.getConfig.setParallelism(parallelism)
     env.disableOperatorChaining()
     // make parameters available in the web interface
