@@ -17,18 +17,18 @@
 
 package org.apache.celeborn.tests.client
 
+import java.nio.charset.StandardCharsets
 import java.util
 
 import scala.collection.JavaConverters._
 import scala.collection.mutable.ArrayBuffer
 
-import org.apache.celeborn.client.{LifecycleManager, WithShuffleClientSuite}
+import org.apache.celeborn.client.{LifecycleManager, ShuffleClientImpl, WithShuffleClientSuite}
 import org.apache.celeborn.client.LifecycleManager.ShuffleFailedWorkers
 import org.apache.celeborn.client.commit.CommitFilesParam
 import org.apache.celeborn.common.CelebornConf
-import org.apache.celeborn.common.protocol.message.ControlMessages.WorkerResource
 import org.apache.celeborn.common.protocol.message.StatusCode
-import org.apache.celeborn.common.util.{CelebornExitKind, Utils}
+import org.apache.celeborn.common.util.Utils
 import org.apache.celeborn.service.deploy.MiniClusterFeature
 import org.apache.celeborn.service.deploy.worker.CommitInfo
 
@@ -59,7 +59,10 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
     assert(res.status == StatusCode.SUCCESS)
     assert(res.workerResource.keySet().size() == 3)
 
-    lifecycleManager.setupEndpoints(res.workerResource, shuffleId, new ShuffleFailedWorkers())
+    lifecycleManager.setupEndpoints(
+      res.workerResource.keySet(),
+      shuffleId,
+      new ShuffleFailedWorkers())
 
     lifecycleManager.reserveSlotsWithRetry(
       shuffleId,
@@ -67,7 +70,7 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
       res.workerResource,
       updateEpoch = false)
 
-    lifecycleManager.commitManager.registerShuffle(shuffleId, 1)
+    lifecycleManager.commitManager.registerShuffle(shuffleId, 1, false)
     0 until 10 foreach { partitionId =>
       lifecycleManager.commitManager.finishMapperAttempt(shuffleId, 0, 0, 1, partitionId)
     }
@@ -87,12 +90,15 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
       new ShuffleFailedWorkers)
 
     workerInfos.keySet.foreach { worker =>
-      worker.controller.shuffleCommitInfos.get(
-        Utils.makeShuffleKey(APP, shuffleId)).values().asScala.foreach { commitInfo =>
-        commitInfo.status == CommitInfo.COMMIT_INPROCESS || commitInfo.status == CommitInfo.COMMIT_FINISHED
+      val commitInfoList =
+        worker.controller.shuffleCommitInfos.get(Utils.makeShuffleKey(APP, shuffleId))
+      if (commitInfoList != null) {
+        commitInfoList.values().asScala.foreach { commitInfo =>
+          assert(
+            commitInfo.status == CommitInfo.COMMIT_INPROCESS || commitInfo.status == CommitInfo.COMMIT_FINISHED)
+        }
       }
     }
-
     lifecycleManager.stop()
   }
 
@@ -109,7 +115,10 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
     assert(res.status == StatusCode.SUCCESS)
     assert(res.workerResource.keySet().size() == 3)
 
-    lifecycleManager.setupEndpoints(res.workerResource, shuffleId, new ShuffleFailedWorkers())
+    lifecycleManager.setupEndpoints(
+      res.workerResource.keySet(),
+      shuffleId,
+      new ShuffleFailedWorkers())
 
     lifecycleManager.reserveSlotsWithRetry(
       shuffleId,
@@ -117,7 +126,7 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
       res.workerResource,
       updateEpoch = false)
 
-    lifecycleManager.commitManager.registerShuffle(shuffleId, 1)
+    lifecycleManager.commitManager.registerShuffle(shuffleId, 1, false)
     0 until 10 foreach { partitionId =>
       lifecycleManager.commitManager.finishMapperAttempt(shuffleId, 0, 0, 1, partitionId)
     }
@@ -137,9 +146,13 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
       new ShuffleFailedWorkers)
 
     workerInfos.keySet.foreach { worker =>
-      worker.controller.shuffleCommitInfos.get(
-        Utils.makeShuffleKey(APP, shuffleId)).values().asScala.foreach { commitInfo =>
-        commitInfo.status == CommitInfo.COMMIT_INPROCESS || commitInfo.status == CommitInfo.COMMIT_FINISHED
+      val commitInfoList =
+        worker.controller.shuffleCommitInfos.get(Utils.makeShuffleKey(APP, shuffleId))
+      if (commitInfoList != null) {
+        commitInfoList.values().asScala.foreach { commitInfo =>
+          assert(
+            commitInfo.status == CommitInfo.COMMIT_INPROCESS || commitInfo.status == CommitInfo.COMMIT_FINISHED)
+        }
       }
     }
 
@@ -172,7 +185,10 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
     val res = lifecycleManager.requestMasterRequestSlotsWithRetry(shuffleId, ids)
     assert(res.status == StatusCode.SUCCESS)
 
-    lifecycleManager.setupEndpoints(res.workerResource, shuffleId, new ShuffleFailedWorkers())
+    lifecycleManager.setupEndpoints(
+      res.workerResource.keySet(),
+      shuffleId,
+      new ShuffleFailedWorkers())
 
     lifecycleManager.reserveSlotsWithRetry(
       shuffleId,
@@ -180,7 +196,7 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
       res.workerResource,
       updateEpoch = false)
 
-    lifecycleManager.commitManager.registerShuffle(shuffleId, 1)
+    lifecycleManager.commitManager.registerShuffle(shuffleId, 1, false)
     0 until 1000 foreach { partitionId =>
       lifecycleManager.commitManager.finishMapperAttempt(shuffleId, 0, 0, 1, partitionId)
     }
@@ -213,6 +229,70 @@ class LifecycleManagerCommitFilesSuite extends WithShuffleClientSuite with MiniC
     lifecycleManager.stop()
   }
 
+  test("CELEBORN-1319: test commit files and check commit info") {
+    val shuffleId = nextShuffleId
+    val conf = celebornConf.clone
+    conf.set(CelebornConf.TEST_MOCK_COMMIT_FILES_FAILURE.key, "false")
+    val lifecycleManager: LifecycleManager = new LifecycleManager(APP, conf)
+    val shuffleClient = new ShuffleClientImpl(APP, conf, userIdentifier)
+    shuffleClient.setupLifecycleManagerRef(lifecycleManager.self)
+
+    val ids = new util.ArrayList[Integer](3)
+    0 until 3 foreach {
+      ids.add(_)
+    }
+    val res = lifecycleManager.requestMasterRequestSlotsWithRetry(shuffleId, ids)
+    assert(res.status == StatusCode.SUCCESS)
+    assert(res.workerResource.keySet().size() == 3)
+
+    lifecycleManager.setupEndpoints(
+      res.workerResource.keySet,
+      shuffleId,
+      new ShuffleFailedWorkers())
+
+    lifecycleManager.reserveSlotsWithRetry(
+      shuffleId,
+      new util.HashSet(res.workerResource.keySet()),
+      res.workerResource,
+      updateEpoch = false)
+
+    lifecycleManager.commitManager.registerShuffle(shuffleId, 1, false)
+
+    val buffer = "hello world".getBytes(StandardCharsets.UTF_8)
+
+    var bufferLength = -1
+
+    0 until 3 foreach { partitionId =>
+      bufferLength =
+        shuffleClient.pushData(shuffleId, 0, 0, partitionId, buffer, 0, buffer.length, 1, 3)
+      lifecycleManager.commitManager.finishMapperAttempt(shuffleId, 0, 0, 1, partitionId)
+    }
+
+    val commitHandler = lifecycleManager.commitManager.getCommitHandler(shuffleId)
+    val params = new ArrayBuffer[CommitFilesParam](res.workerResource.size())
+    res.workerResource.asScala.foreach { case (workerInfo, (primaryIds, replicaIds)) =>
+      params += CommitFilesParam(
+        workerInfo,
+        primaryIds.asScala.map(_.getUniqueId).toList.asJava,
+        replicaIds.asScala.map(_.getUniqueId).toList.asJava)
+    }
+
+    val shuffleCommittedInfo = lifecycleManager.commitManager.committedPartitionInfo.get(shuffleId)
+    commitHandler.doParallelCommitFiles(
+      shuffleId,
+      shuffleCommittedInfo,
+      params,
+      new ShuffleFailedWorkers)
+
+    shuffleCommittedInfo.committedReplicaStorageInfos.values().asScala.foreach { storageInfo =>
+      assert(storageInfo.fileSize == bufferLength)
+      // chunkOffsets contains 0 by default, and bufferFlushOffset
+      assert(storageInfo.chunkOffsets.size() == 2)
+    }
+
+    lifecycleManager.stop()
+  }
+
   override def afterAll(): Unit = {
     logInfo("all test complete , stop celeborn mini cluster")
     shutdownMiniCluster()
