@@ -24,12 +24,17 @@ import java.util.concurrent.ScheduledExecutorService;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
 
+import com.google.common.annotations.VisibleForTesting;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
+import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.common.identity.UserIdentifier;
+import org.apache.celeborn.common.quota.UserTrafficQuota;
+import org.apache.celeborn.common.quota.WorkerTrafficQuota;
 import org.apache.celeborn.common.util.JavaUtils;
 import org.apache.celeborn.common.util.ThreadUtils;
+import org.apache.celeborn.server.common.service.config.ConfigService;
 import org.apache.celeborn.service.deploy.worker.WorkerSource;
 import org.apache.celeborn.service.deploy.worker.memory.MemoryManager;
 
@@ -40,36 +45,54 @@ public class CongestionController {
   private final WorkerSource workerSource;
 
   private final int sampleTimeWindowSeconds;
-  private final long highWatermark;
-  private final long lowWatermark;
   private final long userInactiveTimeMills;
 
   private final AtomicBoolean overHighWatermark = new AtomicBoolean(false);
 
   private final BufferStatusHub consumedBufferStatusHub;
 
+  private final BufferStatusHub producedBufferStatusHub;
+
   private final ConcurrentHashMap<UserIdentifier, UserBufferInfo> userBufferStatuses;
 
   private final ScheduledExecutorService removeUserExecutorService;
 
   private final ScheduledExecutorService checkService;
 
+  private final ConcurrentHashMap<UserIdentifier, UserCongestionControlContext>
+      userCongestionContextMap;
+
+  private final ConfigService configService;
+
+  private final UserTrafficQuota defaultUserQuota;
+
+  private volatile WorkerTrafficQuota workerTrafficQuota;
+
   protected CongestionController(
       WorkerSource workerSource,
       int sampleTimeWindowSeconds,
-      long highWatermark,
-      long lowWatermark,
-      long userInactiveTimeMills,
-      long checkIntervalTimeMills) {
-    assert (highWatermark > lowWatermark);
+      CelebornConf conf,
+      ConfigService configService) {
 
     this.workerSource = workerSource;
     this.sampleTimeWindowSeconds = sampleTimeWindowSeconds;
-    this.highWatermark = highWatermark;
-    this.lowWatermark = lowWatermark;
-    this.userInactiveTimeMills = userInactiveTimeMills;
+    this.userInactiveTimeMills = conf.workerCongestionControlUserInactiveIntervalMs();
     this.consumedBufferStatusHub = new BufferStatusHub(sampleTimeWindowSeconds);
+    this.producedBufferStatusHub = new BufferStatusHub(sampleTimeWindowSeconds);
     this.userBufferStatuses = JavaUtils.newConcurrentHashMap();
+    this.userCongestionContextMap = JavaUtils.newConcurrentHashMap();
+
+    defaultUserQuota =
+        new UserTrafficQuota(
+            conf.workerCongestionControlUserProduceSpeedHighWatermark(),
+            conf.workerCongestionControlUserProduceSpeedLowWatermark());
+
+    workerTrafficQuota =
+        new WorkerTrafficQuota(
+            conf.workerCongestionControlDiskBufferHighWatermark(),
+            conf.workerCongestionControlDiskBufferLowWatermark(),
+            conf.workerCongestionControlWorkerProduceSpeedHighWatermark(),
+            conf.workerCongestionControlWorkerProduceSpeedLowWatermark());
 
     this.removeUserExecutorService =
         ThreadUtils.newDaemonSingleThreadScheduledExecutor(
@@ -82,30 +105,31 @@ public class CongestionController {
         ThreadUtils.newDaemonSingleThreadScheduledExecutor("worker-congestion-controller-checker");
 
     this.checkService.scheduleWithFixedDelay(
-        this::checkCongestion, 0, checkIntervalTimeMills, TimeUnit.MILLISECONDS);
+        this::checkCongestion,
+        0,
+        conf.workerCongestionControlCheckIntervalMs(),
+        TimeUnit.MILLISECONDS);
 
     this.workerSource.addGauge(
         WorkerSource.POTENTIAL_CONSUME_SPEED(), this::getPotentialConsumeSpeed);
 
     this.workerSource.addGauge(
         WorkerSource.WORKER_CONSUME_SPEED(), consumedBufferStatusHub::avgBytesPerSec);
+
+    this.configService = configService;
+
+    if (configService != null) {
+      updateQuota();
+      configService.registerListenerOnConfigUpdate(this::updateQuota);
+    }
   }
 
   public static synchronized CongestionController initialize(
       WorkerSource workSource,
       int sampleTimeWindowSeconds,
-      long highWatermark,
-      long lowWatermark,
-      long userInactiveTimeMills,
-      long checkIntervalTimeMills) {
-    _INSTANCE =
-        new CongestionController(
-            workSource,
-            sampleTimeWindowSeconds,
-            highWatermark,
-            lowWatermark,
-            userInactiveTimeMills,
-            checkIntervalTimeMills);
+      CelebornConf conf,
+      ConfigService configService) {
+    _INSTANCE = new CongestionController(workSource, sampleTimeWindowSeconds, conf, configService);
     return _INSTANCE;
   }
 
@@ -122,25 +146,43 @@ public class CongestionController {
    * <p>3. If the pending bytes doesn't exceed the high watermark, will allow all users to try to
    * get max throughout capacity.
    */
-  public boolean isUserCongested(UserIdentifier userIdentifier) {
-    if (userBufferStatuses.size() == 0) {
+  public boolean isUserCongested(UserCongestionControlContext userCongestionControlContext) {
+    if (userBufferStatuses.isEmpty()) {
       return false;
     }
+
+    UserIdentifier userIdentifier = userCongestionControlContext.getUserIdentifier();
+    long userProduceSpeed = getUserProduceSpeed(userCongestionControlContext.getUserBufferInfo());
+    UserTrafficQuota userTrafficQuota = userCongestionControlContext.getUserTrafficQuota();
+    // If the user produce speed is higher that the avg consume speed, will congest it
     if (overHighWatermark.get()) {
-      // If the user produce speed is higher that the avg consume speed, will congest it
-      long userProduceSpeed = getUserProduceSpeed(userBufferStatuses.get(userIdentifier));
-      long avgConsumeSpeed = getPotentialConsumeSpeed();
+      long avgConsumeSpeed = getPotentialProduceSpeed();
+      if (userProduceSpeed > avgConsumeSpeed) {
+        if (logger.isDebugEnabled()) {
+          logger.debug(
+              "The user {}, produceSpeed is {}, while consumeSpeed is {}, need to congest it.",
+              userIdentifier,
+              userProduceSpeed,
+              avgConsumeSpeed);
+        }
+        return true;
+      }
+    }
+
+    if (userProduceSpeed > userTrafficQuota.userProduceSpeedHighWatermark()) {
+      userCongestionControlContext.onCongestionControl();
       if (logger.isDebugEnabled()) {
         logger.debug(
-            "The user {}, produceSpeed is {}, while consumeSpeed is {}, need to congest it: {}",
+            "The user {}, produceSpeed is {}, while userProduceSpeedHighWatermark is {}, need to congest it.",
             userIdentifier,
             userProduceSpeed,
-            avgConsumeSpeed,
-            userProduceSpeed > avgConsumeSpeed);
+            userTrafficQuota.userProduceSpeedHighWatermark());
       }
-      return userProduceSpeed > avgConsumeSpeed;
+    } else if (userCongestionControlContext.inCongestionControl()
+        && userProduceSpeed < userTrafficQuota.userProduceSpeedLowWatermark()) {
+      userCongestionControlContext.offCongestionControl();
     }
-    return false;
+    return userCongestionControlContext.inCongestionControl();
   }
 
   public UserBufferInfo getUserBuffer(UserIdentifier userIdentifier) {
@@ -150,10 +192,6 @@ public class CongestionController {
           logger.info("New user {} comes, initializing its rate status", user);
           BufferStatusHub bufferStatusHub = new BufferStatusHub(sampleTimeWindowSeconds);
           UserBufferInfo userInfo = new UserBufferInfo(System.currentTimeMillis(), bufferStatusHub);
-          workerSource.addGauge(
-              WorkerSource.USER_PRODUCE_SPEED(),
-              userIdentifier.toJMap(),
-              () -> getUserProduceSpeed(userInfo));
           return userInfo;
         });
   }
@@ -184,6 +222,14 @@ public class CongestionController {
     return consumedBufferStatusHub.avgBytesPerSec() / userBufferStatuses.size();
   }
 
+  public long getPotentialProduceSpeed() {
+    if (userBufferStatuses.size() == 0) {
+      return 0;
+    }
+
+    return producedBufferStatusHub.avgBytesPerSec() / userBufferStatuses.size();
+  }
+
   /** Get the avg user produce speed, the unit is bytes/sec. */
   private long getUserProduceSpeed(UserBufferInfo userBufferInfo) {
     if (userBufferInfo != null) {
@@ -205,6 +251,7 @@ public class CongestionController {
         UserBufferInfo userBufferInfo = next.getValue();
         if (currentTimeMillis - userBufferInfo.getTimestamp() >= userInactiveTimeMills) {
           userBufferStatuses.remove(userIdentifier);
+          userCongestionContextMap.remove(userIdentifier);
           workerSource.removeGauge(WorkerSource.USER_PRODUCE_SPEED(), userIdentifier.toMap());
           logger.info("User {} has been expired, remove from rate limit list", userIdentifier);
         }
@@ -217,13 +264,19 @@ public class CongestionController {
   protected void checkCongestion() {
     try {
       long pendingConsume = getTotalPendingBytes();
-      if (pendingConsume < lowWatermark) {
+      long workerProduceSpeed = producedBufferStatusHub.avgBytesPerSec();
+      if (pendingConsume < workerTrafficQuota.diskBufferLowWatermark()
+          && workerProduceSpeed < workerTrafficQuota.workerProduceSpeedLowWatermark()) {
         if (overHighWatermark.compareAndSet(true, false)) {
-          logger.info("Pending consume is lower than low watermark, exit congestion control");
+          logger.info(
+              "Pending consume and produce speed is lower than low watermark, exit congestion control");
         }
         return;
-      } else if (pendingConsume > highWatermark && overHighWatermark.compareAndSet(false, true)) {
-        logger.info("Pending consume is higher than high watermark, need congestion control");
+      } else if ((pendingConsume > workerTrafficQuota.diskBufferHighWatermark()
+              || workerProduceSpeed > workerTrafficQuota.workerProduceSpeedHighWatermark())
+          && overHighWatermark.compareAndSet(false, true)) {
+        logger.info(
+            "Pending consume or produce speed is higher than high watermark, need congestion control");
       }
       if (overHighWatermark.get()) {
         trimMemoryUsage();
@@ -239,10 +292,16 @@ public class CongestionController {
 
   public void close() {
     logger.info("Closing {}", this.getClass().getSimpleName());
-    this.removeUserExecutorService.shutdownNow();
-    this.checkService.shutdownNow();
+    ThreadUtils.shutdown(this.removeUserExecutorService);
+    ThreadUtils.shutdown(this.checkService);
     this.userBufferStatuses.clear();
     this.consumedBufferStatusHub.clear();
+    this.producedBufferStatusHub.clear();
+  }
+
+  @VisibleForTesting
+  public void shutDownCheckService() {
+    ThreadUtils.shutdown(this.checkService);
   }
 
   public static synchronized void destroy() {
@@ -251,4 +310,53 @@ public class CongestionController {
       _INSTANCE = null;
     }
   }
+
+  public BufferStatusHub getProducedBufferStatusHub() {
+    return producedBufferStatusHub;
+  }
+
+  public UserCongestionControlContext getUserCongestionContext(UserIdentifier userIdentifier) {
+    return userCongestionContextMap.computeIfAbsent(
+        userIdentifier,
+        user -> {
+          UserBufferInfo userBufferInfo = getUserBuffer(userIdentifier);
+          UserTrafficQuota userTrafficQuota;
+          if (configService == null) {
+            userTrafficQuota = defaultUserQuota;
+          } else {
+            userTrafficQuota =
+                configService
+                    .getTenantUserConfigFromCache(userIdentifier.tenantId(), userIdentifier.name())
+                    .getUserTrafficQuota();
+          }
+          return new UserCongestionControlContext(
+              userTrafficQuota,
+              producedBufferStatusHub,
+              userBufferInfo,
+              workerSource,
+              userIdentifier);
+        });
+  }
+
+  public ConcurrentHashMap<UserIdentifier, UserCongestionControlContext>
+      getUserCongestionContextMap() {
+    return userCongestionContextMap;
+  }
+
+  public BufferStatusHub getConsumedBufferStatusHub() {
+    return consumedBufferStatusHub;
+  }
+
+  private void updateQuota() {
+    workerTrafficQuota = configService.getSystemConfigFromCache().getWorkerTrafficQuota();
+    for (Map.Entry<UserIdentifier, UserCongestionControlContext> entry :
+        userCongestionContextMap.entrySet()) {
+      UserIdentifier user = entry.getKey();
+      UserCongestionControlContext context = entry.getValue();
+      context.updateUserTrafficQuota(
+          configService
+              .getTenantUserConfigFromCache(user.tenantId(), user.name())
+              .getUserTrafficQuota());
+    }
+  }
 }
