@@ -45,6 +45,7 @@ import io.netty.buffer.CompositeByteBuf;
 import org.apache.commons.io.IOUtils;
 import org.apache.hadoop.fs.FSDataInputStream;
 import org.apache.hadoop.fs.FSDataOutputStream;
+import org.apache.hadoop.fs.FileSystem;
 import org.apache.hadoop.fs.Path;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -53,6 +54,7 @@ import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.common.identity.UserIdentifier;
 import org.apache.celeborn.common.meta.*;
 import org.apache.celeborn.common.metrics.source.AbstractSource;
+import org.apache.celeborn.common.protocol.StorageInfo;
 import org.apache.celeborn.common.unsafe.Platform;
 import org.apache.celeborn.common.util.*;
 import org.apache.celeborn.common.util.ShuffleBlockInfoUtils.ShuffleBlockInfo;
@@ -175,10 +177,14 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
         });
   }
 
-  public int getSortingCount() {
+  public int getPendingSortTaskCount() {
     return shuffleSortTaskDeque.size();
   }
 
+  public int getSortingCount() {
+    return sortingShuffleFiles.values().stream().map(Set::size).reduce(Integer::sum).orElse(0);
+  }
+
   public int getSortedCount() {
     return sortedFileCount.get();
   }
@@ -209,9 +215,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
                   startMapIndex, endMapIndex, shuffleChunkSize, indexesMap, true),
               shuffleChunkSize);
       CompositeByteBuf targetBuffer =
-          MemoryManager.instance()
-              .getStoragePooledByteBufAllocator()
-              .compositeBuffer(Integer.MAX_VALUE);
+          MemoryManager.instance().getStorageByteBufAllocator().compositeBuffer(Integer.MAX_VALUE);
       ShuffleBlockInfoUtils.sliceSortedBufferByMapRange(
           startMapIndex,
           endMapIndex,
@@ -252,8 +256,8 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
             throw new IOException(
                 "Sort scheduler thread is interrupted means worker is shutting down.", e);
           } catch (IOException e) {
-            logger.error("File sorter access HDFS failed.", e);
-            throw new IOException("File sorter access HDFS failed.", e);
+            logger.error("File sorter access DFS failed.", e);
+            throw new IOException("File sorter access DFS failed.", e);
           }
         }
       }
@@ -320,7 +324,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
           int compressedSize = Platform.getInt(batchHeader, Platform.BYTE_ARRAY_OFFSET + 12);
           ShuffleBlockInfo shuffleBlockInfo = new ShuffleBlockInfo();
           shuffleBlockInfo.offset = index;
-          shuffleBlockInfo.length = 16 + compressedSize;
+          shuffleBlockInfo.length = 16L + compressedSize;
           List<ShuffleBlockInfo> singleMapIdShuffleBlockList =
               blocksMap.computeIfAbsent(mapId, v -> new ArrayList<>());
           singleMapIdShuffleBlockList.add(shuffleBlockInfo);
@@ -332,7 +336,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
         // because this will affect origin buffer's reference count
         CompositeByteBuf sortedBuffer =
             MemoryManager.instance()
-                .getStoragePooledByteBufAllocator()
+                .getStorageByteBufAllocator()
                 .compositeBuffer(Integer.MAX_VALUE - 1);
         Map<Integer, List<ShuffleBlockInfo>> sortedBlocks = new TreeMap<>();
         int sortedBufferIndex = 0;
@@ -476,14 +480,21 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
   }
 
   protected void writeIndex(
-      Map<Integer, List<ShuffleBlockInfo>> indexMap, String indexFilePath, boolean isHdfs)
+      Map<Integer, List<ShuffleBlockInfo>> indexMap, String indexFilePath, boolean isDfs)
       throws IOException {
-    FSDataOutputStream hdfsIndexOutput = null;
+    FSDataOutputStream dfsIndexOutput = null;
     FileChannel indexFileChannel = null;
-    if (isHdfs) {
+    if (isDfs) {
+      StorageInfo.Type storageType = StorageInfo.Type.HDFS;
+      if (Utils.isS3Path(indexFilePath)) {
+        storageType = StorageInfo.Type.S3;
+      } else if (Utils.isOssPath(indexFilePath)) {
+        storageType = StorageInfo.Type.OSS;
+      }
+      FileSystem hadoopFs = StorageManager.hadoopFs().get(storageType);
       // If the index file exists, it will be overwritten.
       // So there is no need to check its existence.
-      hdfsIndexOutput = StorageManager.hadoopFs().create(new Path(indexFilePath));
+      dfsIndexOutput = hadoopFs.create(new Path(indexFilePath));
     } else {
       indexFileChannel = FileChannelUtils.createWritableFileChannel(indexFilePath);
     }
@@ -508,12 +519,12 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     }
 
     indexBuf.flip();
-    if (isHdfs) {
+    if (isDfs) {
       // Direct byte buffer has no array, so can not invoke indexBuf.array() here.
       byte[] tmpBuf = new byte[indexSize];
       indexBuf.get(tmpBuf);
-      hdfsIndexOutput.write(tmpBuf);
-      hdfsIndexOutput.close();
+      dfsIndexOutput.write(tmpBuf);
+      dfsIndexOutput.close();
     } else {
       while (indexBuf.hasRemaining()) {
         indexFileChannel.write(indexBuf);
@@ -591,25 +602,31 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
               fileId,
               () -> {
                 FileChannel indexChannel = null;
-                FSDataInputStream hdfsIndexStream = null;
-                boolean isHdfs = Utils.isHdfsPath(indexFilePath);
+                FSDataInputStream dfsIndexStream = null;
+                boolean isDfs =
+                    Utils.isHdfsPath(indexFilePath)
+                        || Utils.isS3Path(indexFilePath)
+                        || Utils.isOssPath(indexFilePath);
                 int indexSize;
                 try {
-                  if (isHdfs) {
-                    hdfsIndexStream = StorageManager.hadoopFs().open(new Path(indexFilePath));
-                    indexSize =
-                        (int)
-                            StorageManager.hadoopFs()
-                                .getFileStatus(new Path(indexFilePath))
-                                .getLen();
+                  if (isDfs) {
+                    StorageInfo.Type storageType = StorageInfo.Type.HDFS;
+                    if (Utils.isS3Path(indexFilePath)) {
+                      storageType = StorageInfo.Type.S3;
+                    } else if (Utils.isOssPath(indexFilePath)) {
+                      storageType = StorageInfo.Type.OSS;
+                    }
+                    FileSystem hadoopFs = StorageManager.hadoopFs().get(storageType);
+                    dfsIndexStream = hadoopFs.open(new Path(indexFilePath));
+                    indexSize = (int) hadoopFs.getFileStatus(new Path(indexFilePath)).getLen();
                   } else {
                     indexChannel = FileChannelUtils.openReadableFileChannel(indexFilePath);
                     File indexFile = new File(indexFilePath);
                     indexSize = (int) indexFile.length();
                   }
                   ByteBuffer indexBuf = ByteBuffer.allocate(indexSize);
-                  if (isHdfs) {
-                    readStreamFully(hdfsIndexStream, indexBuf, indexFilePath);
+                  if (isDfs) {
+                    readStreamFully(dfsIndexStream, indexBuf, indexFilePath);
                   } else {
                     readChannelFully(indexChannel, indexBuf, indexFilePath);
                   }
@@ -626,7 +643,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
                   throw new IOException("Read sorted shuffle file index failed.", e);
                 } finally {
                   IOUtils.closeQuietly(indexChannel, null);
-                  IOUtils.closeQuietly(hdfsIndexStream, null);
+                  IOUtils.closeQuietly(dfsIndexStream, null);
                 }
               });
     } catch (ExecutionException e) {
@@ -648,25 +665,32 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     private final String fileId;
     private final String shuffleKey;
     private final boolean isHdfs;
+    private final boolean isS3;
+    private final boolean isOss;
+    private final boolean isDfs;
     private final boolean isPrefetch;
     private final FileInfo originFileInfo;
 
-    private FSDataInputStream hdfsOriginInput = null;
-    private FSDataOutputStream hdfsSortedOutput = null;
+    private FSDataInputStream dfsOriginInput = null;
+    private FSDataOutputStream dfsSortedOutput = null;
     private FileChannel originFileChannel = null;
     private FileChannel sortedFileChannel = null;
+    private FileSystem hadoopFs;
 
     FileSorter(DiskFileInfo fileInfo, String fileId, String shuffleKey) throws IOException {
       this.originFileInfo = fileInfo;
       this.originFilePath = fileInfo.getFilePath();
       this.sortedFilePath = Utils.getSortedFilePath(originFilePath);
       this.isHdfs = fileInfo.isHdfs();
-      this.isPrefetch = !isHdfs && prefetchEnabled;
+      this.isS3 = fileInfo.isS3();
+      this.isOss = fileInfo.isOSS();
+      this.isDfs = isHdfs || isS3 || isOss;
+      this.isPrefetch = !isDfs && prefetchEnabled;
       this.originFileLen = fileInfo.getFileLength();
       this.fileId = fileId;
       this.shuffleKey = shuffleKey;
       this.indexFilePath = Utils.getIndexFilePath(originFilePath);
-      if (!isHdfs) {
+      if (!isDfs) {
         File sortedFile = new File(this.sortedFilePath);
         if (sortedFile.exists()) {
           sortedFile.delete();
@@ -676,11 +700,18 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
           indexFile.delete();
         }
       } else {
-        if (StorageManager.hadoopFs().exists(fileInfo.getHdfsSortedPath())) {
-          StorageManager.hadoopFs().delete(fileInfo.getHdfsSortedPath(), false);
+        StorageInfo.Type storageType = StorageInfo.Type.HDFS;
+        if (Utils.isS3Path(indexFilePath)) {
+          storageType = StorageInfo.Type.S3;
+        } else if (Utils.isOssPath(indexFilePath)) {
+          storageType = StorageInfo.Type.OSS;
         }
-        if (StorageManager.hadoopFs().exists(fileInfo.getHdfsIndexPath())) {
-          StorageManager.hadoopFs().delete(fileInfo.getHdfsIndexPath(), false);
+        this.hadoopFs = StorageManager.hadoopFs().get(storageType);
+        if (hadoopFs.exists(fileInfo.getDfsSortedPath())) {
+          hadoopFs.delete(fileInfo.getDfsSortedPath(), false);
+        }
+        if (hadoopFs.exists(fileInfo.getDfsIndexPath())) {
+          hadoopFs.delete(fileInfo.getDfsIndexPath(), false);
         }
       }
     }
@@ -751,7 +782,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
           sortedBlockInfoMap.put(mapId, sortedShuffleBlocks);
         }
 
-        writeIndex(sortedBlockInfoMap, indexFilePath, isHdfs);
+        writeIndex(sortedBlockInfoMap, indexFilePath, isDfs);
         updateSortedShuffleFiles(shuffleKey, fileId, originFileLen);
         originFileInfo.getReduceFileMeta().setSorted();
         cleaner.add(this);
@@ -778,10 +809,9 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     }
 
     private void initializeFiles() throws IOException {
-      if (isHdfs) {
-        hdfsOriginInput = StorageManager.hadoopFs().open(new Path(originFilePath));
-        hdfsSortedOutput =
-            StorageManager.hadoopFs().create(new Path(sortedFilePath), true, 256 * 1024);
+      if (isDfs) {
+        dfsOriginInput = hadoopFs.open(new Path(originFilePath));
+        dfsSortedOutput = hadoopFs.create(new Path(sortedFilePath), true, 256 * 1024);
       } else {
         originFileChannel = FileChannelUtils.openReadableFileChannel(originFilePath);
         sortedFileChannel = FileChannelUtils.createWritableFileChannel(sortedFilePath);
@@ -789,23 +819,23 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     }
 
     private void closeFiles() {
-      IOUtils.closeQuietly(hdfsOriginInput, null);
-      IOUtils.closeQuietly(hdfsSortedOutput, null);
+      IOUtils.closeQuietly(dfsOriginInput, null);
+      IOUtils.closeQuietly(dfsSortedOutput, null);
       IOUtils.closeQuietly(originFileChannel, null);
       IOUtils.closeQuietly(sortedFileChannel, null);
     }
 
     private void readBufferFully(ByteBuffer buffer) throws IOException {
-      if (isHdfs) {
-        readStreamFully(hdfsOriginInput, buffer, originFilePath);
+      if (isDfs) {
+        readStreamFully(dfsOriginInput, buffer, originFilePath);
       } else {
         readChannelFully(originFileChannel, buffer, originFilePath);
       }
     }
 
     private long transferBlock(long offset, long length) throws IOException {
-      if (isHdfs) {
-        return transferStreamFully(hdfsOriginInput, hdfsSortedOutput, offset, length);
+      if (isDfs) {
+        return transferStreamFully(dfsOriginInput, dfsSortedOutput, offset, length);
       } else {
         return transferChannelFully(originFileChannel, sortedFileChannel, offset, length);
       }
@@ -813,8 +843,8 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
 
     public void deleteOriginFiles() throws IOException {
       boolean deleteSuccess;
-      if (isHdfs) {
-        deleteSuccess = StorageManager.hadoopFs().delete(new Path(originFilePath), false);
+      if (isDfs) {
+        deleteSuccess = hadoopFs.delete(new Path(originFilePath), false);
       } else {
         deleteSuccess = new File(originFilePath).delete();
       }
@@ -852,9 +882,9 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     }
 
     private void readBufferBySize(ByteBuffer buffer, int toRead) throws IOException {
-      if (isHdfs) {
-        // HDFS does not need to prefetch.
-        hdfsOriginInput.seek(toRead + hdfsOriginInput.getPos());
+      if (isDfs) {
+        // DFS does not need to prefetch.
+        dfsOriginInput.seek(toRead + dfsOriginInput.getPos());
       } else if (prefetchEnabled) {
         buffer.clear();
         readChannelBySize(originFileChannel, buffer, originFilePath, toRead);
