@@ -18,9 +18,8 @@
 package org.apache.celeborn.service.deploy.worker
 
 import java.io.File
-import java.lang.{Long => JLong}
 import java.util
-import java.util.{HashMap => JHashMap, HashSet => JHashSet, Locale, Map => JMap, UUID}
+import java.util.{HashSet => JHashSet, Locale, Map => JMap, UUID}
 import java.util.concurrent._
 import java.util.concurrent.atomic.{AtomicBoolean, AtomicIntegerArray}
 
@@ -37,8 +36,9 @@ import org.apache.celeborn.common.identity.UserIdentifier
 import org.apache.celeborn.common.internal.Logging
 import org.apache.celeborn.common.meta.{DiskInfo, WorkerInfo, WorkerPartitionLocationInfo}
 import org.apache.celeborn.common.metrics.MetricsSystem
-import org.apache.celeborn.common.metrics.source.{JVMCPUSource, JVMSource, ResourceConsumptionSource, SystemMiscSource, ThreadPoolSource}
+import org.apache.celeborn.common.metrics.source.{JVMCPUSource, JVMSource, ResourceConsumptionSource, Role, SystemMiscSource, ThreadPoolSource}
 import org.apache.celeborn.common.network.{CelebornRackResolver, TransportContext}
+import org.apache.celeborn.common.network.protocol.TransportMessagesHelper
 import org.apache.celeborn.common.network.sasl.SaslServerBootstrap
 import org.apache.celeborn.common.network.server.TransportServerBootstrap
 import org.apache.celeborn.common.network.util.TransportConf
@@ -52,7 +52,6 @@ import org.apache.celeborn.common.util.{CelebornExitKind, CollectionUtils, JavaU
 // Can Remove this if celeborn don't support scala211 in future
 import org.apache.celeborn.common.util.FunctionConverter._
 import org.apache.celeborn.server.common.{HttpService, Service}
-import org.apache.celeborn.service.deploy.worker.WorkerSource.ACTIVE_CONNECTION_COUNT
 import org.apache.celeborn.service.deploy.worker.congestcontrol.CongestionController
 import org.apache.celeborn.service.deploy.worker.memory.{ChannelsLimiter, MemoryManager}
 import org.apache.celeborn.service.deploy.worker.memory.MemoryManager.ServingState
@@ -72,20 +71,28 @@ private[celeborn] class Worker(
   override val metricsSystem: MetricsSystem =
     MetricsSystem.createMetricsSystem(serviceName, conf)
   val workerSource = new WorkerSource(conf)
-  private val resourceConsumptionSource =
-    new ResourceConsumptionSource(conf, MetricsSystem.ROLE_WORKER)
-  private val threadPoolSource = ThreadPoolSource(conf, MetricsSystem.ROLE_WORKER)
+  val resourceConsumptionSource =
+    new ResourceConsumptionSource(conf, Role.WORKER)
+  private val threadPoolSource = ThreadPoolSource(conf, Role.WORKER)
   metricsSystem.registerSource(workerSource)
   metricsSystem.registerSource(threadPoolSource)
   metricsSystem.registerSource(resourceConsumptionSource)
-  metricsSystem.registerSource(new JVMSource(conf, MetricsSystem.ROLE_WORKER))
-  metricsSystem.registerSource(new JVMCPUSource(conf, MetricsSystem.ROLE_WORKER))
-  metricsSystem.registerSource(new SystemMiscSource(conf, MetricsSystem.ROLE_WORKER))
+  metricsSystem.registerSource(new JVMSource(conf, Role.WORKER))
+  metricsSystem.registerSource(new JVMCPUSource(conf, Role.WORKER))
+  metricsSystem.registerSource(new SystemMiscSource(conf, Role.WORKER))
+
+  private val topAppResourceConsumptionCount = conf.metricsWorkerAppTopResourceConsumptionCount
+  private val topAppResourceConsumptionBytesWrittenThreshold =
+    conf.metricsWorkerAppTopResourceConsumptionBytesWrittenThreshold
+  private val topApplicationUserIdentifiers =
+    JavaUtils.newConcurrentHashMap[String, UserIdentifier]()
 
   val workerStatusManager = new WorkerStatusManager(conf)
   private val authEnabled = conf.authEnabled
   private val secretRegistry = new WorkerSecretRegistryImpl(conf.workerApplicationRegistryCacheSize)
 
+  private val hasHDFSStorage = conf.hasHDFSStorage
+
   if (conf.logCelebornConfEnabled) {
     logInfo(getConf)
   }
@@ -96,10 +103,10 @@ private[celeborn] class Worker(
         RpcNameConstants.WORKER_SYS,
         TransportModuleConstants.RPC_SERVICE_MODULE,
         workerArgs.host,
-        workerArgs.host,
         workerArgs.port,
         conf,
         Math.min(64, Math.max(4, Runtime.getRuntime.availableProcessors())),
+        Role.WORKER,
         None,
         Some(workerSource))
     } else {
@@ -114,13 +121,14 @@ private[celeborn] class Worker(
         RpcNameConstants.WORKER_SYS,
         TransportModuleConstants.RPC_SERVICE_MODULE,
         workerArgs.host,
-        workerArgs.host,
         workerArgs.port,
         conf,
         Math.max(64, Runtime.getRuntime.availableProcessors()),
+        Role.WORKER,
         Some(externalSecurityContext),
         Some(workerSource))
     }
+  metricsSystem.registerSource(rpcEnv.rpcSource())
 
   private[worker] var internalRpcEnvInUse =
     if (!conf.internalPortEnabled) {
@@ -130,10 +138,10 @@ private[celeborn] class Worker(
         RpcNameConstants.WORKER_INTERNAL_SYS,
         TransportModuleConstants.RPC_SERVICE_MODULE,
         workerArgs.host,
-        workerArgs.host,
         workerArgs.internalPort,
         conf,
         Math.min(64, Math.max(4, Runtime.getRuntime.availableProcessors())),
+        Role.WORKER,
         None,
         Some(workerSource))
     }
@@ -174,24 +182,17 @@ private[celeborn] class Worker(
 
   val storageManager = new StorageManager(conf, workerSource)
 
-  val memoryManager: MemoryManager = MemoryManager.initialize(conf, storageManager)
+  val memoryManager: MemoryManager = MemoryManager.initialize(conf, storageManager, workerSource)
   memoryManager.registerMemoryListener(storageManager)
 
   val partitionsSorter = new PartitionFilesSorter(memoryManager, conf, workerSource)
 
   if (conf.workerCongestionControlEnabled) {
-    if (conf.workerCongestionControlLowWatermark.isEmpty || conf.workerCongestionControlHighWatermark.isEmpty) {
-      throw new IllegalArgumentException("High watermark and low watermark must be set" +
-        " when enabling rate limit")
-    }
-
     CongestionController.initialize(
       workerSource,
       conf.workerCongestionControlSampleTimeWindowSeconds.toInt,
-      conf.workerCongestionControlHighWatermark.get,
-      conf.workerCongestionControlLowWatermark.get,
-      conf.workerCongestionControlUserInactiveIntervalMs,
-      conf.workerCongestionControlCheckIntervalMs)
+      conf,
+      configService)
   }
 
   val controller = new Controller(rpcEnv, conf, metricsSystem, workerSource)
@@ -261,7 +262,8 @@ private[celeborn] class Worker(
         fetchHandler,
         closeIdleConnections,
         conf.workerFetchHeartbeatEnabled,
-        workerSource)
+        workerSource,
+        conf.metricsCollectCriticalEnabled)
     (
       transportContext,
       transportContext.createServer(conf.workerFetchPort, getServerBootstraps(transportConf)))
@@ -309,6 +311,10 @@ private[celeborn] class Worker(
   val shuffleCommitInfos: ConcurrentHashMap[String, ConcurrentHashMap[Long, CommitInfo]] =
     JavaUtils.newConcurrentHashMap[String, ConcurrentHashMap[Long, CommitInfo]]()
 
+  val shuffleCommitTime
+      : ConcurrentHashMap[String, ConcurrentHashMap[Long, (Long, RpcCallContext)]] =
+    JavaUtils.newConcurrentHashMap[String, ConcurrentHashMap[Long, (Long, RpcCallContext)]]()
+
   private val masterClient = new MasterClient(internalRpcEnvInUse, conf, true)
   secretRegistry.initialize(masterClient)
 
@@ -329,6 +335,8 @@ private[celeborn] class Worker(
     ThreadUtils.newDaemonCachedThreadPool("worker-data-replicator", conf.workerReplicateThreads)
   val commitThreadPool: ThreadPoolExecutor =
     ThreadUtils.newDaemonCachedThreadPool("worker-files-committer", conf.workerCommitThreads)
+  val commitFinishedChecker: ScheduledExecutorService =
+    ThreadUtils.newDaemonSingleThreadScheduledExecutor("worker-commit-checker")
   val cleanThreadPool: ThreadPoolExecutor =
     ThreadUtils.newDaemonCachedThreadPool(
       "worker-expired-shuffle-cleaner",
@@ -353,10 +361,12 @@ private[celeborn] class Worker(
 
   private var jvmQuake: JVMQuake = _
   if (conf.workerJvmQuakeEnabled) {
-    jvmQuake = JVMQuake.create(conf, workerInfo.toUniqueId().replace(":", "-"))
+    jvmQuake = JVMQuake.create(conf, workerInfo.toUniqueId.replace(":", "-"))
     jvmQuake.start()
   }
 
+  private val messagesHelper: TransportMessagesHelper = new TransportMessagesHelper()
+
   workerSource.addGauge(WorkerSource.REGISTERED_SHUFFLE_COUNT) { () =>
     workerInfo.getShuffleKeySet.size
   }
@@ -366,6 +376,9 @@ private[celeborn] class Worker(
   workerSource.addGauge(WorkerSource.SORT_MEMORY) { () =>
     memoryManager.getSortMemoryCounter.get()
   }
+  workerSource.addGauge(WorkerSource.PENDING_SORT_TASKS) { () =>
+    partitionsSorter.getPendingSortTaskCount
+  }
   workerSource.addGauge(WorkerSource.SORTING_FILES) { () =>
     partitionsSorter.getSortingCount
   }
@@ -432,37 +445,51 @@ private[celeborn] class Worker(
       0
     }
   }
+  // Unreleased shuffle count when worker is decommissioning
+  workerSource.addGauge(WorkerSource.UNRELEASED_SHUFFLE_COUNT) { () =>
+    if (shutdown.get() && (workerStatusManager.currentWorkerStatus.getState == State.InDecommission ||
+        workerStatusManager.currentWorkerStatus.getState == State.InDecommissionThenIdle)) {
+      storageManager.shuffleKeySet().size
+    } else {
+      0
+    }
+  }
+  // Unreleased partition location count when worker is restarting
+  workerSource.addGauge(WorkerSource.UNRELEASED_PARTITION_LOCATION_COUNT) { () =>
+    if (shutdown.get()) {
+      partitionLocationInfo.primaryPartitionLocations.size() +
+        partitionLocationInfo.replicaPartitionLocations.size()
+    } else {
+      0
+    }
+  }
   workerSource.addGauge(WorkerSource.CLEAN_TASK_QUEUE_SIZE) { () =>
     cleanTaskQueue.size()
   }
 
   private def highWorkload: Boolean = {
     (
-      memoryManager.currentServingState(),
+      memoryManager.currentServingState,
       Option(CongestionController.instance()),
       conf.workerActiveConnectionMax) match {
       case (_, Some(instance), _) if instance.isOverHighWatermark => true
       case (ServingState.PUSH_AND_REPLICATE_PAUSED, _, _) => true
       case (ServingState.PUSH_PAUSED, _, _) => true
       case (_, _, Some(activeConnectionMax)) =>
-        workerSource.getCounterCount(ACTIVE_CONNECTION_COUNT) >= activeConnectionMax
+        workerSource.getCounterCount(WorkerSource.ACTIVE_CONNECTION_COUNT) >= activeConnectionMax
       case _ => false
     }
   }
 
   private def heartbeatToMaster(): Unit = {
     val activeShuffleKeys = new JHashSet[String]()
-    val estimatedAppDiskUsage = new JHashMap[String, JLong]()
     activeShuffleKeys.addAll(partitionLocationInfo.shuffleKeySet)
     activeShuffleKeys.addAll(storageManager.shuffleKeySet())
-    storageManager.topAppDiskUsage.asScala.foreach { case (shuffleId, usage) =>
-      estimatedAppDiskUsage.put(shuffleId, usage)
-    }
     storageManager.updateDiskInfos()
     val diskInfos =
       workerInfo.updateThenGetDiskInfos(storageManager.disksSnapshot().map { disk =>
         disk.mountPoint -> disk
-      }.toMap.asJava).values().asScala.toSeq ++ storageManager.hdfsDiskInfo
+      }.toMap.asJava).values().asScala.toSeq ++ storageManager.hdfsDiskInfo ++ storageManager.s3DiskInfo ++ storageManager.ossDiskInfo
     workerStatusManager.checkIfNeedTransitionStatus()
     val response = masterClient.askSync[HeartbeatFromWorkerResponse](
       HeartbeatFromWorker(
@@ -474,7 +501,6 @@ private[celeborn] class Worker(
         diskInfos,
         handleResourceConsumption(),
         activeShuffleKeys,
-        estimatedAppDiskUsage,
         highWorkload,
         workerStatusManager.currentWorkerStatus),
       classOf[HeartbeatFromWorkerResponse])
@@ -584,11 +610,13 @@ private[celeborn] class Worker(
         forwardMessageScheduler.shutdown()
         replicateThreadPool.shutdown()
         commitThreadPool.shutdown()
+        commitFinishedChecker.shutdown();
         asyncReplyPool.shutdown()
       } else {
         forwardMessageScheduler.shutdownNow()
         replicateThreadPool.shutdownNow()
         commitThreadPool.shutdownNow()
+        commitFinishedChecker.shutdownNow();
         asyncReplyPool.shutdownNow()
       }
       workerSource.appActiveConnections.clear()
@@ -608,6 +636,7 @@ private[celeborn] class Worker(
       if (conf.internalPortEnabled) {
         internalRpcEnvInUse.stop(internalRpcEndpointRef)
       }
+      messagesHelper.close()
       super.stop(exitKind)
 
       logInfo("Worker is stopped.")
@@ -642,6 +671,7 @@ private[celeborn] class Worker(
             logWarning(
               s"Register worker to master failed, will retry after ${Utils.msDurationToString(interval)}",
               throwable)
+            workerSource.incCounter(WorkerSource.REGISTER_WITH_MASTER_FAIL_COUNT)
             exception = throwable
             null
         }
@@ -660,58 +690,87 @@ private[celeborn] class Worker(
   }
 
   private def handleResourceConsumption(): util.Map[UserIdentifier, ResourceConsumption] = {
-    val resourceConsumptionSnapshot = storageManager.userResourceConsumptionSnapshot()
-    val userResourceConsumptions =
-      workerInfo.updateThenGetUserResourceConsumption(resourceConsumptionSnapshot.asJava)
-    resourceConsumptionSnapshot.foreach { case (userIdentifier, userResourceConsumption) =>
+    val resourceConsumptionSnapshot = storageManager.userResourceConsumptionSnapshot().asJava
+    workerInfo.updateThenGetUserResourceConsumption(resourceConsumptionSnapshot)
+    resourceConsumptionSnapshot.asScala.foreach { case (userIdentifier, _) =>
       gaugeResourceConsumption(userIdentifier)
     }
-    userResourceConsumptions
+    if (topAppResourceConsumptionCount > 0) {
+      handleTopAppResourceConsumption(resourceConsumptionSnapshot)
+    }
+    resourceConsumptionSnapshot
+  }
+
+  def handleTopAppResourceConsumption(userResourceConsumptions: util.Map[
+    UserIdentifier,
+    ResourceConsumption]): Unit = {
+    // Remove application top resource consumption gauges to refresh top resource consumption metrics.
+    removeAppResourceConsumption(topApplicationUserIdentifiers.keySet().asScala)
+    // Top resource consumption is determined by diskBytesWritten+hdfsBytesWritten.
+    userResourceConsumptions.asScala.filter { case (_, resourceConsumption) =>
+      CollectionUtils.isNotEmpty(resourceConsumption.subResourceConsumptions)
+    }.flatMap { case (userIdentifier, resourceConsumption) =>
+      resourceConsumption.subResourceConsumptions.asScala.map { case (appId, appConsumption) =>
+        (appId, userIdentifier, appConsumption)
+      }
+    }.toSeq
+      .sortBy { case (_, _, appConsumption) =>
+        appConsumption.diskBytesWritten + appConsumption.hdfsBytesWritten
+      }
+      .reverse
+      .take(topAppResourceConsumptionCount).foreach {
+        case (appId, userIdentifier, appConsumption) =>
+          if (appConsumption.diskBytesWritten + appConsumption.hdfsBytesWritten >=
+              topAppResourceConsumptionBytesWrittenThreshold) {
+            topApplicationUserIdentifiers.put(appId, userIdentifier)
+            gaugeResourceConsumption(userIdentifier, appId, appConsumption)
+          } else {
+            return
+          }
+      }
   }
 
   private def gaugeResourceConsumption(
       userIdentifier: UserIdentifier,
-      applicationId: String = null): Unit = {
+      applicationId: String = null,
+      resourceConsumption: ResourceConsumption = null): Unit = {
     var resourceConsumptionLabel = userIdentifier.toMap
     if (applicationId != null)
       resourceConsumptionLabel += (resourceConsumptionSource.applicationLabel -> applicationId)
     resourceConsumptionSource.addGauge(
       ResourceConsumptionSource.DISK_FILE_COUNT,
       resourceConsumptionLabel) { () =>
-      computeResourceConsumption(userIdentifier, applicationId).diskFileCount
+      computeResourceConsumption(userIdentifier, resourceConsumption).diskFileCount
     }
     resourceConsumptionSource.addGauge(
       ResourceConsumptionSource.DISK_BYTES_WRITTEN,
       resourceConsumptionLabel) { () =>
-      computeResourceConsumption(userIdentifier, applicationId).diskBytesWritten
+      computeResourceConsumption(userIdentifier, resourceConsumption).diskBytesWritten
     }
-    resourceConsumptionSource.addGauge(
-      ResourceConsumptionSource.HDFS_FILE_COUNT,
-      resourceConsumptionLabel) { () =>
-      computeResourceConsumption(userIdentifier, applicationId).hdfsFileCount
-    }
-    resourceConsumptionSource.addGauge(
-      ResourceConsumptionSource.HDFS_BYTES_WRITTEN,
-      resourceConsumptionLabel) { () =>
-      computeResourceConsumption(userIdentifier, applicationId).hdfsBytesWritten
+    if (hasHDFSStorage) {
+      resourceConsumptionSource.addGauge(
+        ResourceConsumptionSource.HDFS_FILE_COUNT,
+        resourceConsumptionLabel) { () =>
+        computeResourceConsumption(userIdentifier, resourceConsumption).hdfsFileCount
+      }
+      resourceConsumptionSource.addGauge(
+        ResourceConsumptionSource.HDFS_BYTES_WRITTEN,
+        resourceConsumptionLabel) { () =>
+        computeResourceConsumption(userIdentifier, resourceConsumption).hdfsBytesWritten
+      }
     }
   }
 
   private def computeResourceConsumption(
       userIdentifier: UserIdentifier,
-      applicationId: String = null): ResourceConsumption = {
-    var resourceConsumption =
+      resourceConsumption: ResourceConsumption = null): ResourceConsumption = {
+    if (resourceConsumption == null) {
       workerInfo.userResourceConsumption.getOrDefault(
         userIdentifier,
         ResourceConsumption(0, 0, 0, 0))
-    if (applicationId != null) {
-      val subResourceConsumptions = resourceConsumption.subResourceConsumptions
-      if (CollectionUtils.isNotEmpty(subResourceConsumptions)) {
-        resourceConsumption =
-          subResourceConsumptions.getOrDefault(applicationId, ResourceConsumption(0, 0, 0, 0))
-      }
+    } else {
+      resourceConsumption
     }
-    resourceConsumption
   }
 
   @VisibleForTesting
@@ -724,6 +783,7 @@ private[celeborn] class Worker(
         shufflePushDataTimeout.remove(shuffleKey)
         shuffleMapperAttempts.remove(shuffleKey)
         shuffleCommitInfos.remove(shuffleKey)
+        shuffleCommitTime.remove(shuffleKey)
         workerInfo.releaseSlots(shuffleKey)
         val applicationId = Utils.splitShuffleKey(shuffleKey)._1
         if (!workerInfo.getApplicationIdSet.contains(applicationId)) {
@@ -736,6 +796,7 @@ private[celeborn] class Worker(
       fetchHandler.cleanupExpiredShuffleKey(expiredShuffleKeys)
       threadPool.execute(new Runnable {
         override def run(): Unit = {
+          removeAppResourceConsumption(expiredApplicationIds.asScala)
           removeAppActiveConnection(expiredApplicationIds)
           workerSource.sample(
             WorkerSource.CLEAN_EXPIRED_SHUFFLE_KEYS_TIME,
@@ -746,6 +807,35 @@ private[celeborn] class Worker(
       })
     }
 
+  private def removeAppResourceConsumption(applicationIds: Iterable[String]): Unit = {
+    applicationIds.foreach { applicationId => removeAppResourceConsumption(applicationId) }
+  }
+
+  private def removeAppResourceConsumption(applicationId: String): Unit = {
+    val userIdentifier = topApplicationUserIdentifiers.remove(applicationId)
+    if (userIdentifier != null) {
+      removeAppResourceConsumption(
+        userIdentifier.toMap + (resourceConsumptionSource.applicationLabel -> applicationId))
+    }
+  }
+
+  private def removeAppResourceConsumption(resourceConsumptionLabel: Map[String, String]): Unit = {
+    resourceConsumptionSource.removeGauge(
+      ResourceConsumptionSource.DISK_FILE_COUNT,
+      resourceConsumptionLabel)
+    resourceConsumptionSource.removeGauge(
+      ResourceConsumptionSource.DISK_BYTES_WRITTEN,
+      resourceConsumptionLabel)
+    if (hasHDFSStorage) {
+      resourceConsumptionSource.removeGauge(
+        ResourceConsumptionSource.HDFS_FILE_COUNT,
+        resourceConsumptionLabel)
+      resourceConsumptionSource.removeGauge(
+        ResourceConsumptionSource.HDFS_BYTES_WRITTEN,
+        resourceConsumptionLabel)
+    }
+  }
+
   private def removeAppActiveConnection(applicationIds: JHashSet[String]): Unit = {
     workerSource.removeAppActiveConnection(applicationIds)
   }
@@ -782,6 +872,16 @@ private[celeborn] class Worker(
     sb.toString()
   }
 
+  override def isDecommissioning: String = {
+    val sb = new StringBuilder
+    sb.append("========================= Worker Decommission ==========================\n")
+    sb.append(
+      shutdown.get() && (workerStatusManager.currentWorkerStatus.getState == State.InDecommission ||
+        workerStatusManager.currentWorkerStatus.getState == State.InDecommissionThenIdle))
+      .append("\n")
+    sb.toString()
+  }
+
   override def isRegistered: String = {
     val sb = new StringBuilder
     sb.append("========================= Worker Registered ==========================\n")
@@ -789,15 +889,6 @@ private[celeborn] class Worker(
     sb.toString()
   }
 
-  override def listTopDiskUseApps: String = {
-    val sb = new StringBuilder
-    sb.append("================== Top Disk Usage Applications =======================\n")
-    storageManager.topAppDiskUsage.asScala.foreach { case (appId, usage) =>
-      sb.append(s"Application $appId used ${Utils.bytesToString(usage)}\n")
-    }
-    sb.toString()
-  }
-
   override def listPartitionLocationInfo: String = {
     val sb = new StringBuilder
     sb.append("==================== Partition Location Info =========================\n")
@@ -809,7 +900,7 @@ private[celeborn] class Worker(
     val sb = new StringBuilder
     sb.append("==================== Unavailable Peers of Worker =====================\n")
     unavailablePeers.asScala.foreach { case (peer, time) =>
-      sb.append(s"${peer.toUniqueId().padTo(50, " ").mkString}${Utils.formatTimestamp(time)}\n")
+      sb.append(s"${peer.toUniqueId.padTo(50, " ").mkString}${Utils.formatTimestamp(time)}\n")
     }
     sb.toString()
   }
@@ -874,10 +965,10 @@ private[celeborn] class Worker(
     workerStatusManager.transitionState(State.Exit)
   }
 
-  def sendWorkerUnavailableToMaster(): Unit = {
+  def sendWorkerDecommissionToMaster(): Unit = {
     try {
       masterClient.askSync(
-        ReportWorkerUnavailable(List(workerInfo).asJava),
+        ReportWorkerDecommission(List(workerInfo).asJava),
         OneWayMessageResponse.getClass)
     } catch {
       case e: Throwable =>
@@ -891,7 +982,7 @@ private[celeborn] class Worker(
   def decommissionWorker(): Unit = {
     logInfo("Worker start to decommission")
     workerStatusManager.transitionState(State.InDecommission)
-    sendWorkerUnavailableToMaster()
+    sendWorkerDecommissionToMaster()
     shutdown.set(true)
     val interval = conf.workerDecommissionCheckInterval
     val timeout = conf.workerDecommissionForceExitTimeout
@@ -903,11 +994,13 @@ private[celeborn] class Worker(
       Thread.sleep(interval)
       waitTimes += 1
     }
-    if (storageManager.shuffleKeySet().isEmpty) {
+
+    val unreleasedShuffleKeys = storageManager.shuffleKeySet()
+    if (unreleasedShuffleKeys.isEmpty) {
       logInfo(s"Waiting for all shuffle expired cost ${waitTime}ms.")
     } else {
       logWarning(s"Waiting for all shuffle expired cost ${waitTime}ms, " +
-        s"unreleased shuffle: \n${storageManager.shuffleKeySet().asScala.mkString("[", ", ", "]")}")
+        s"unreleased shuffle: \n${unreleasedShuffleKeys.asScala.mkString("[", ", ", "]")}")
     }
     workerStatusManager.transitionState(State.Exit)
   }
