@@ -25,16 +25,16 @@ import scala.collection.JavaConverters._
 import com.google.common.collect.Sets
 
 import org.apache.celeborn.common.CelebornConf
-import org.apache.celeborn.common.metrics.MetricsSystem
-import org.apache.celeborn.common.metrics.source.AbstractSource
+import org.apache.celeborn.common.metrics.source.{AbstractSource, Role}
 import org.apache.celeborn.common.network.client.TransportClient
 import org.apache.celeborn.common.util.{JavaUtils, Utils}
 
-class WorkerSource(conf: CelebornConf) extends AbstractSource(conf, MetricsSystem.ROLE_WORKER) {
+class WorkerSource(conf: CelebornConf) extends AbstractSource(conf, Role.WORKER) {
   override val sourceName = "worker"
 
   val appActiveConnections: ConcurrentHashMap[String, util.Set[String]] =
     JavaUtils.newConcurrentHashMap[String, util.Set[String]]
+  private val metricsAppLevelEnabled = conf.metricsWorkerAppLevelEnabled
 
   import WorkerSource._
   // add counters
@@ -56,8 +56,12 @@ class WorkerSource(conf: CelebornConf) extends AbstractSource(conf, MetricsSyste
   addCounter(REGION_START_FAIL_COUNT)
   addCounter(REGION_FINISH_FAIL_COUNT)
   addCounter(ACTIVE_CONNECTION_COUNT)
+  addCounter(SEGMENT_START_FAIL_COUNT)
 
   addCounter(SLOTS_ALLOCATED)
+  addCounter(REGISTER_WITH_MASTER_FAIL_COUNT)
+
+  addCounter(COMMIT_FILES_FAIL_COUNT)
 
   // add timers
   addTimer(COMMIT_FILES_TIME)
@@ -72,50 +76,64 @@ class WorkerSource(conf: CelebornConf) extends AbstractSource(conf, MetricsSyste
   addTimer(REPLICA_REGION_START_TIME)
   addTimer(PRIMARY_REGION_FINISH_TIME)
   addTimer(REPLICA_REGION_FINISH_TIME)
+  addTimer(PRIMARY_SEGMENT_START_TIME)
+  addTimer(REPLICA_SEGMENT_START_TIME)
 
   addTimer(FETCH_CHUNK_TIME)
   addTimer(OPEN_STREAM_TIME)
   addTimer(TAKE_BUFFER_TIME)
   addTimer(SORT_TIME)
+  addTimer(FETCH_CHUNK_TRANSFER_TIME)
 
   addTimer(CLEAN_EXPIRED_SHUFFLE_KEYS_TIME)
 
+  addHistogram(FETCH_CHUNK_TRANSFER_SIZE)
+  addHistogram(PARTITION_FILE_SIZE)
+
   def getCounterCount(metricsName: String): Long = {
     val metricNameWithLabel = metricNameWithCustomizedLabels(metricsName, Map.empty)
     namedCounters.get(metricNameWithLabel).counter.getCount
   }
 
   def connectionActive(client: TransportClient): Unit = {
-    appActiveConnections.putIfAbsent(
-      client.getChannel.id().asLongText(),
-      Sets.newConcurrentHashSet[String]())
+    if (metricsAppLevelEnabled) {
+      appActiveConnections.putIfAbsent(
+        client.getChannel.id().asLongText(),
+        Sets.newConcurrentHashSet[String]())
+    }
     incCounter(ACTIVE_CONNECTION_COUNT, 1)
   }
 
   def connectionInactive(client: TransportClient): Unit = {
-    val applicationIds = appActiveConnections.remove(client.getChannel.id().asLongText())
     incCounter(ACTIVE_CONNECTION_COUNT, -1)
-    if (null != applicationIds) {
-      applicationIds.asScala.foreach(applicationId =>
-        incCounter(ACTIVE_CONNECTION_COUNT, -1, Map(applicationLabel -> applicationId)))
+    if (metricsAppLevelEnabled) {
+      val applicationIds = appActiveConnections.remove(client.getChannel.id().asLongText())
+      if (null != applicationIds) {
+        applicationIds.asScala.foreach(applicationId =>
+          incCounter(ACTIVE_CONNECTION_COUNT, -1, Map(applicationLabel -> applicationId)))
+      }
     }
   }
 
   def recordAppActiveConnection(client: TransportClient, shuffleKey: String): Unit = {
-    val applicationIds = appActiveConnections.get(client.getChannel.id().asLongText())
-    val applicationId = Utils.splitShuffleKey(shuffleKey)._1
-    if (applicationIds != null && !applicationIds.contains(applicationId)) {
-      addCounter(ACTIVE_CONNECTION_COUNT, Map(applicationLabel -> applicationId))
-      incCounter(ACTIVE_CONNECTION_COUNT, 1, Map(applicationLabel -> applicationId))
-      applicationIds.add(applicationId)
+    if (metricsAppLevelEnabled) {
+      val applicationIds = appActiveConnections.get(client.getChannel.id().asLongText())
+      val applicationId = Utils.splitShuffleKey(shuffleKey)._1
+      if (applicationIds != null && !applicationIds.contains(applicationId)) {
+        addCounter(ACTIVE_CONNECTION_COUNT, Map(applicationLabel -> applicationId))
+        incCounter(ACTIVE_CONNECTION_COUNT, 1, Map(applicationLabel -> applicationId))
+        applicationIds.add(applicationId)
+      }
     }
   }
 
   def removeAppActiveConnection(applicationIds: util.Set[String]): Unit = {
-    applicationIds.asScala.foreach(applicationId =>
-      removeCounter(ACTIVE_CONNECTION_COUNT, Map(applicationLabel -> applicationId)))
-    appActiveConnections.values().asScala.foreach(connectionAppIds =>
-      applicationIds.asScala.foreach(applicationId => connectionAppIds.remove(applicationId)))
+    if (metricsAppLevelEnabled) {
+      applicationIds.asScala.foreach(applicationId =>
+        removeCounter(ACTIVE_CONNECTION_COUNT, Map(applicationLabel -> applicationId)))
+      appActiveConnections.values().asScala.foreach(connectionAppIds =>
+        applicationIds.asScala.foreach(applicationId => connectionAppIds.remove(applicationId)))
+    }
   }
 
   // start cleaner thread
@@ -127,6 +145,8 @@ object WorkerSource {
 
   val RUNNING_APPLICATION_COUNT = "RunningApplicationCount"
 
+  val REGISTER_WITH_MASTER_FAIL_COUNT = "RegisterWithMasterFailCount"
+
   // fetch data
   val OPEN_STREAM_TIME = "OpenStreamTime"
   val FETCH_CHUNK_TIME = "FetchChunkTime"
@@ -135,6 +155,8 @@ object WorkerSource {
   val OPEN_STREAM_FAIL_COUNT = "OpenStreamFailCount"
   val FETCH_CHUNK_SUCCESS_COUNT = "FetchChunkSuccessCount"
   val FETCH_CHUNK_FAIL_COUNT = "FetchChunkFailCount"
+  val FETCH_CHUNK_TRANSFER_SIZE = "FetchChunkTransferSize"
+  val FETCH_CHUNK_TRANSFER_TIME = "FetchChunkTransferTime"
 
   // push data
   val PRIMARY_PUSH_DATA_TIME = "PrimaryPushDataTime"
@@ -151,12 +173,15 @@ object WorkerSource {
   val PUSH_DATA_HANDSHAKE_FAIL_COUNT = "PushDataHandshakeFailCount"
   val REGION_START_FAIL_COUNT = "RegionStartFailCount"
   val REGION_FINISH_FAIL_COUNT = "RegionFinishFailCount"
+  val SEGMENT_START_FAIL_COUNT = "SegmentStartFailCount"
   val PRIMARY_PUSH_DATA_HANDSHAKE_TIME = "PrimaryPushDataHandshakeTime"
   val REPLICA_PUSH_DATA_HANDSHAKE_TIME = "ReplicaPushDataHandshakeTime"
   val PRIMARY_REGION_START_TIME = "PrimaryRegionStartTime"
   val REPLICA_REGION_START_TIME = "ReplicaRegionStartTime"
   val PRIMARY_REGION_FINISH_TIME = "PrimaryRegionFinishTime"
   val REPLICA_REGION_FINISH_TIME = "ReplicaRegionFinishTime"
+  val PRIMARY_SEGMENT_START_TIME = "PrimarySegmentStartTime"
+  val REPLICA_SEGMENT_START_TIME = "ReplicaSegmentStartTime"
 
   // pause push data
   val PAUSE_PUSH_DATA_TIME = "PausePushDataTime"
@@ -168,6 +193,7 @@ object WorkerSource {
   val TAKE_BUFFER_TIME = "TakeBufferTime"
   val FLUSH_DATA_TIME = "FlushDataTime"
   val COMMIT_FILES_TIME = "CommitFilesTime"
+  val COMMIT_FILES_FAIL_COUNT = "CommitFilesFailCount"
   val FLUSH_WORKING_QUEUE_SIZE = "FlushWorkingQueueSize"
 
   // slots
@@ -183,6 +209,7 @@ object WorkerSource {
   val SORT_TIME = "SortTime"
   val SORT_MEMORY = "SortMemory"
   val SORTING_FILES = "SortingFiles"
+  val PENDING_SORT_TASKS = "PendingSortTasks"
   val SORTED_FILES = "SortedFiles"
   val SORTED_FILE_SIZE = "SortedFileSize"
   val DISK_BUFFER = "DiskBuffer"
@@ -204,6 +231,7 @@ object WorkerSource {
   val DEVICE_OS_TOTAL_CAPACITY = "DeviceOSTotalBytes"
   val DEVICE_CELEBORN_FREE_CAPACITY = "DeviceCelebornFreeBytes"
   val DEVICE_CELEBORN_TOTAL_CAPACITY = "DeviceCelebornTotalBytes"
+  val PARTITION_FILE_SIZE = "PartitionFileSizeBytes"
 
   // congestion control
   val POTENTIAL_CONSUME_SPEED = "PotentialConsumeSpeed"
@@ -216,6 +244,10 @@ object WorkerSource {
 
   // decommission
   val IS_DECOMMISSIONING_WORKER = "IsDecommissioningWorker"
+  val UNRELEASED_SHUFFLE_COUNT = "UnreleasedShuffleCount"
+
+  // graceful
+  val UNRELEASED_PARTITION_LOCATION_COUNT = "UnreleasedPartitionLocationCount"
 
   // clean
   val CLEAN_TASK_QUEUE_SIZE = "CleanTaskQueueSize"
