@@ -29,7 +29,7 @@ import scala.collection.JavaConverters._
 import scala.concurrent.duration._
 
 import com.google.common.annotations.VisibleForTesting
-import io.netty.buffer.PooledByteBufAllocator
+import io.netty.buffer.{ByteBufAllocator, PooledByteBufAllocator}
 import org.apache.commons.io.FileUtils
 import org.apache.hadoop.fs.{FileSystem, Path}
 import org.apache.hadoop.fs.permission.FsPermission
@@ -57,6 +57,8 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   val workingDirWriters =
     JavaUtils.newConcurrentHashMap[File, ConcurrentHashMap[String, PartitionDataWriter]]()
   val hdfsWriters = JavaUtils.newConcurrentHashMap[String, PartitionDataWriter]()
+  val s3Writers = JavaUtils.newConcurrentHashMap[String, PartitionDataWriter]()
+  val ossWriters = JavaUtils.newConcurrentHashMap[String, PartitionDataWriter]()
   val memoryWriters = JavaUtils.newConcurrentHashMap[MemoryFileInfo, PartitionDataWriter]()
   // (shuffleKey->(filename->DiskFileInfo))
   private val diskFileInfos =
@@ -67,7 +69,12 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
 
   val hasHDFSStorage = conf.hasHDFSStorage
 
+  val hasS3Storage = conf.hasS3Storage
+
+  val hasOssStorage = conf.hasOssStorage
+
   val storageExpireDirTimeout = conf.workerStorageExpireDirTimeout
+  val storagePolicy = new StoragePolicy(conf, this, workerSource)
 
   val diskReserveSize = conf.workerDiskReserveSize
   val diskReserveRatio = conf.workerDiskReserveRatio
@@ -79,7 +86,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         (new File(workdir, conf.workerWorkingDir), maxSpace, flusherThread, storageType)
       }
 
-    if (workingDirInfos.size <= 0 && !hasHDFSStorage) {
+    if (workingDirInfos.size <= 0 && !hasHDFSStorage && !hasS3Storage && !hasOssStorage) {
       throw new IOException("Empty working directory configuration!")
     }
 
@@ -91,6 +98,16 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       Option(new DiskInfo("HDFS", Long.MaxValue, 999999, 999999, 0, StorageInfo.Type.HDFS))
     else None
 
+  val s3DiskInfo =
+    if (conf.hasS3Storage)
+      Option(new DiskInfo("S3", Long.MaxValue, 999999, 999999, 0, StorageInfo.Type.S3))
+    else None
+
+  val ossDiskInfo =
+    if (conf.hasOssStorage)
+      Option(new DiskInfo("OSS", Long.MaxValue, 999999, 999999, 0, StorageInfo.Type.OSS))
+    else None
+
   def disksSnapshot(): List[DiskInfo] = {
     diskInfos.synchronized {
       val disks = new util.ArrayList[DiskInfo](diskInfos.values())
@@ -121,8 +138,8 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   private val deviceMonitor =
     DeviceMonitor.createDeviceMonitor(conf, this, deviceInfos, tmpDiskInfos, workerSource)
 
-  val storageBufferAllocator: PooledByteBufAllocator =
-    NettyUtils.getPooledByteBufAllocator(new TransportConf("StorageManager", conf), null, true)
+  val storageBufferAllocator: ByteBufAllocator =
+    NettyUtils.getByteBufAllocator(new TransportConf("StorageManager", conf), workerSource, true)
 
   // (mountPoint -> LocalFlusher)
   private val (
@@ -149,6 +166,8 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   }
 
   val hdfsDir = conf.hdfsDir
+  val s3Dir = conf.s3Dir
+  val ossDir = conf.ossDir
   val hdfsPermission = new FsPermission("755")
   val (hdfsFlusher, _totalHdfsFlusherThread) =
     if (hasHDFSStorage) {
@@ -171,12 +190,57 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       (None, 0)
     }
 
-  def totalFlusherThread: Int = _totalLocalFlusherThread + _totalHdfsFlusherThread
+  val (s3Flusher, _totalS3FlusherThread) =
+    if (hasS3Storage) {
+      logInfo(s"Initialize S3 support with path $s3Dir")
+      try {
+        StorageManager.hadoopFs = CelebornHadoopUtils.getHadoopFS(conf)
+      } catch {
+        case e: Exception =>
+          logError("Celeborn initialize S3 failed.", e)
+          throw e
+      }
+      (
+        Some(new S3Flusher(
+          workerSource,
+          conf.workerS3FlusherThreads,
+          storageBufferAllocator,
+          conf.workerPushMaxComponents)),
+        conf.workerS3FlusherThreads)
+    } else {
+      (None, 0)
+    }
+
+  val (ossFlusher, _totalOssFlusherThread) =
+    if (hasOssStorage) {
+      logInfo(s"Initialize OSS support with path $ossDir")
+      try {
+        StorageManager.hadoopFs = CelebornHadoopUtils.getHadoopFS(conf)
+      } catch {
+        case e: Exception =>
+          logError("Celeborn initialize OSS failed.", e)
+          throw e
+      }
+      (
+        Some(new OssFlusher(
+          workerSource,
+          conf.workerOssFlusherThreads,
+          storageBufferAllocator,
+          conf.workerPushMaxComponents)),
+        conf.workerOssFlusherThreads)
+    } else {
+      (None, 0)
+    }
+
+  def totalFlusherThread: Int =
+    _totalLocalFlusherThread + _totalHdfsFlusherThread + _totalS3FlusherThread + _totalOssFlusherThread
+
   val activeTypes = conf.availableStorageTypes
 
-  def localOrHdfsStorageAvailable(): Boolean = {
-    StorageInfo.HDFSAvailable(activeTypes) || StorageInfo.localDiskAvailable(
-      activeTypes) || hdfsDir.nonEmpty || !diskInfos.isEmpty
+  lazy val localOrDfsStorageAvailable: Boolean = {
+    StorageInfo.OSSAvailable(activeTypes) || StorageInfo.S3Available(activeTypes) ||
+    StorageInfo.HDFSAvailable(activeTypes) || StorageInfo.localDiskAvailable(activeTypes) ||
+    hdfsDir.nonEmpty || !diskInfos.isEmpty || s3Dir.nonEmpty || ossDir.nonEmpty
   }
 
   override def notifyError(mountPoint: String, diskStatus: DiskStatus): Unit = this.synchronized {
@@ -369,7 +433,8 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       partitionType,
       rangeReadFilter,
       userIdentifier,
-      true)
+      true,
+      isSegmentGranularityVisible = false)
   }
 
   @throws[IOException]
@@ -382,8 +447,9 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       partitionType: PartitionType,
       rangeReadFilter: Boolean,
       userIdentifier: UserIdentifier,
-      partitionSplitEnabled: Boolean): PartitionDataWriter = {
-    if (healthyWorkingDirs().size <= 0 && !hasHDFSStorage) {
+      partitionSplitEnabled: Boolean,
+      isSegmentGranularityVisible: Boolean): PartitionDataWriter = {
+    if (healthyWorkingDirs().size <= 0 && !hasHDFSStorage && !hasS3Storage && !hasOssStorage) {
       throw new IOException("No available working dirs!")
     }
     val partitionDataWriterContext = new PartitionDataWriterContext(
@@ -395,25 +461,17 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       shuffleId,
       userIdentifier,
       partitionType,
-      partitionSplitEnabled)
+      partitionSplitEnabled,
+      isSegmentGranularityVisible)
 
     val writer =
       try {
-        partitionType match {
-          case PartitionType.MAP => new MapPartitionDataWriter(
-              this,
-              workerSource,
-              conf,
-              deviceMonitor,
-              partitionDataWriterContext)
-          case PartitionType.REDUCE => new ReducePartitionDataWriter(
-              this,
-              workerSource,
-              conf,
-              deviceMonitor,
-              partitionDataWriterContext)
-          case _ => throw new UnsupportedOperationException(s"Not support $partitionType yet")
-        }
+        new PartitionDataWriter(
+          this,
+          conf,
+          deviceMonitor,
+          partitionDataWriterContext,
+          partitionType)
       } catch {
         case e: Exception =>
           logError("Create partition data writer failed", e)
@@ -441,11 +499,19 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       writer: PartitionDataWriter,
       workingDir: File,
       fileInfo: DiskFileInfo): Unit = {
-    if (writer.getDiskFileInfo.isHdfs) {
+    if (fileInfo.isHdfs) {
       hdfsWriters.put(fileInfo.getFilePath, writer)
       return
     }
-    deviceMonitor.registerFileWriter(writer)
+    if (fileInfo.isS3) {
+      s3Writers.put(fileInfo.getFilePath, writer)
+      return
+    }
+    if (fileInfo.isOSS) {
+      ossWriters.put(fileInfo.getFilePath, writer)
+      return
+    }
+    deviceMonitor.registerFileWriter(writer, fileInfo.getFilePath)
     workingDirWriters.computeIfAbsent(workingDir, workingDirWriterListFunc).put(
       fileInfo.getFilePath,
       writer)
@@ -508,35 +574,39 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
     hashSet
   }
 
-  def topAppDiskUsage: util.Map[String, Long] = {
-    diskFileInfos.asScala.map { keyedWriters =>
-      {
-        keyedWriters._1 -> keyedWriters._2.values().asScala.map(_.getFileLength).sum
-      }
-    }.toList.map { case (shuffleKey, usage) =>
-      Utils.splitShuffleKey(shuffleKey)._1 -> usage
-    }.groupBy(_._1).map { case (key, values) =>
-      key -> values.map(_._2).sum
-    }.toSeq.sortBy(_._2).reverse.take(conf.metricsAppTopDiskUsageCount * 2).toMap.asJava
-  }
-
   def cleanFile(shuffleKey: String, fileName: String): Unit = {
     cleanFileInternal(shuffleKey, removeFileInfo(shuffleKey, fileName))
   }
 
   def cleanFileInternal(shuffleKey: String, fileInfo: FileInfo): Boolean = {
     if (fileInfo == null) return false
-    var isHdfsExpired = false
+    var isDfsExpired = false
     fileInfo match {
       case info: DiskFileInfo =>
         if (info.isHdfs) {
-          isHdfsExpired = true
+          isDfsExpired = true
           val hdfsFileWriter = hdfsWriters.get(info.getFilePath)
           if (hdfsFileWriter != null) {
             hdfsFileWriter.destroy(new IOException(
               s"Destroy FileWriter $hdfsFileWriter caused by shuffle $shuffleKey expired."))
             hdfsWriters.remove(info.getFilePath)
           }
+        } else if (info.isS3) {
+          isDfsExpired = true
+          val s3FileWriter = s3Writers.get(info.getFilePath)
+          if (s3FileWriter != null) {
+            s3FileWriter.destroy(new IOException(
+              s"Destroy FileWriter $s3FileWriter caused by shuffle $shuffleKey expired."))
+            s3Writers.remove(info.getFilePath)
+          }
+        } else if (info.isOSS) {
+          isDfsExpired = true
+          val ossFileWriter = ossWriters.get(info.getFilePath)
+          if (ossFileWriter != null) {
+            ossFileWriter.destroy(new IOException(
+              s"Destroy FileWriter $ossFileWriter caused by shuffle $shuffleKey expired."))
+            ossWriters.remove(info.getFilePath)
+          }
         } else {
           val workingDir =
             info.getFile.getParentFile.getParentFile.getParentFile
@@ -557,7 +627,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       case _ =>
     }
 
-    isHdfsExpired
+    isDfsExpired
   }
 
   def cleanupExpiredShuffleKey(
@@ -567,12 +637,16 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       logInfo(s"Cleanup expired shuffle $shuffleKey.")
       if (diskFileInfos.containsKey(shuffleKey)) {
         val removedFileInfos = diskFileInfos.remove(shuffleKey)
-        var isHdfsExpired = false
+        var isDfsExpired = false
+        var isHdfs = false
+        var isOss = false
         if (removedFileInfos != null) {
           removedFileInfos.asScala.foreach {
             case (_, fileInfo) =>
               if (cleanFileInternal(shuffleKey, fileInfo)) {
-                isHdfsExpired = true
+                isDfsExpired = true
+                isHdfs = fileInfo.isHdfs
+                isOss = fileInfo.isOSS
               }
           }
         }
@@ -585,13 +659,21 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
             deleteDirectory(file, diskOperators.get(diskInfo.mountPoint))
           }
         }
-        if (isHdfsExpired) {
+        if (isDfsExpired) {
           try {
-            StorageManager.hadoopFs.delete(
-              new Path(new Path(hdfsDir, conf.workerWorkingDir), s"$appId/$shuffleId"),
+            val dir =
+              if (hasHDFSStorage && isHdfs) hdfsDir
+              else if (hasOssStorage && isOss) ossDir
+              else s3Dir
+            val storageInfo =
+              if (hasHDFSStorage && isHdfs) StorageInfo.Type.HDFS
+              else if (hasOssStorage && isOss) StorageInfo.Type.OSS
+              else StorageInfo.Type.S3
+            StorageManager.hadoopFs.get(storageInfo).delete(
+              new Path(new Path(dir, conf.workerWorkingDir), s"$appId/$shuffleId"),
               true)
           } catch {
-            case e: Exception => logWarning("Clean expired HDFS shuffle failed.", e)
+            case e: Exception => logWarning("Clean expired DFS shuffle failed.", e)
           }
         }
         if (workerGracefulShutdown) {
@@ -691,12 +773,13 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
           }
         }
 
-      val hdfsCleaned = hadoopFs match {
-        case hdfs: FileSystem =>
-          val hdfsWorkPath = new Path(hdfsDir, conf.workerWorkingDir)
-          // HDFS path not exist when first time initialize
-          if (hdfs.exists(hdfsWorkPath)) {
-            !hdfs.listFiles(hdfsWorkPath, false).hasNext
+      val dfsCleaned = hadoopFs match {
+        case dfs: FileSystem =>
+          val dfsDir = if (hasHDFSStorage) hdfsDir else if (hasOssStorage) ossDir else s3Dir
+          val dfsWorkPath = new Path(dfsDir, conf.workerWorkingDir)
+          // DFS path not exist when first time initialize
+          if (dfs.exists(dfsWorkPath)) {
+            !dfs.listFiles(dfsWorkPath, false).hasNext
           } else {
             true
           }
@@ -704,7 +787,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
           true
       }
 
-      if (localCleaned && hdfsCleaned) {
+      if (localCleaned && dfsCleaned) {
         return true
       }
       retryTimes += 1
@@ -759,28 +842,31 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       override def accept(
           t: File,
           writers: ConcurrentHashMap[String, PartitionDataWriter]): Unit = {
-        writers.forEach(new BiConsumer[String, PartitionDataWriter] {
-          override def accept(file: String, writer: PartitionDataWriter): Unit = {
-            if (writer.getException == null) {
-              try {
-                writer.flushOnMemoryPressure()
-              } catch {
-                case t: Throwable =>
-                  logError(
-                    s"FileWrite of $writer faces unexpected exception when flush on memory pressure.",
-                    t)
-              }
-            } else {
-              logWarning(s"Skip flushOnMemoryPressure because ${writer.flusher} " +
-                s"has error: ${writer.getException.getMessage}")
-            }
-          }
-        })
+        flushOnMemoryPressure(writers)
       }
     })
-    hdfsWriters.forEach(new BiConsumer[String, PartitionDataWriter] {
-      override def accept(t: String, u: PartitionDataWriter): Unit = {
-        u.flushOnMemoryPressure()
+    flushOnMemoryPressure(hdfsWriters)
+    flushOnMemoryPressure(s3Writers)
+    flushOnMemoryPressure(ossWriters)
+  }
+
+  private def flushOnMemoryPressure(writers: ConcurrentHashMap[String, PartitionDataWriter])
+      : Unit = {
+    writers.forEach(new BiConsumer[String, PartitionDataWriter] {
+      override def accept(file: String, writer: PartitionDataWriter): Unit = {
+        if (writer.getException == null) {
+          try {
+            writer.flushOnMemoryPressure()
+          } catch {
+            case t: Throwable =>
+              logError(
+                s"FileWrite of $writer faces unexpected exception when flush on memory pressure.",
+                t)
+          }
+        } else {
+          logWarning(s"Skip flushOnMemoryPressure because ${writer.getFlusher} " +
+            s"has error: ${writer.getException.getMessage}")
+        }
       }
     })
   }
@@ -864,18 +950,15 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         }
         // userIdentifier -> List((userIdentifier, (applicationId, fileInfo))))
         .groupBy(_._1)
-        .map { case (userIdentifier, userWithFileInfoList) =>
+        .mapValues { userWithFileInfoList =>
           // collect resource consumed by each user on this worker
-          val userFileInfos = userWithFileInfoList.map(_._2)
-          (
-            userIdentifier,
-            resourceConsumption(
-              userFileInfos.map(_._2),
-              userFileInfos.groupBy(_._1).map {
-                case (applicationId, appWithFileInfoList) =>
-                  (applicationId, resourceConsumption(appWithFileInfoList.map(_._2)))
-              }.asJava))
-        }
+          val subResourceConsumption = userWithFileInfoList.map(_._2).groupBy(_._1).map {
+            case (applicationId, appWithFileInfoList) =>
+              (applicationId, resourceConsumption(appWithFileInfoList.map(_._2)))
+          }
+          subResourceConsumption.values.foldLeft(ResourceConsumption(0, 0, 0, 0))(_ add _)
+            .withSubResourceConsumptions(subResourceConsumption.asJava)
+        }.toMap
     }
   }
 
@@ -883,7 +966,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       fileInfos: List[DiskFileInfo],
       subResourceConsumptions: util.Map[String, ResourceConsumption] = null)
       : ResourceConsumption = {
-    val diskFileInfos = fileInfos.filter(!_.isHdfs)
+    val diskFileInfos = fileInfos.filter(!_.isDFS)
     val hdfsFileInfos = fileInfos.filter(_.isHdfs)
     ResourceConsumption(
       diskFileInfos.map(_.getFileLength).sum,
@@ -919,7 +1002,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       && MemoryManager.instance().memoryFileStorageAvailable()) {
       logDebug(s"Create memory file for ${partitionDataWriterContext.getShuffleKey} ${partitionDataWriterContext.getPartitionLocation.getFileName}")
       (
-        createMemoryFile(
+        createMemoryFileInfo(
           partitionDataWriterContext.getAppId,
           partitionDataWriterContext.getShuffleId,
           location.getFileName,
@@ -929,7 +1012,8 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         null,
         null,
         null)
-    } else if (location.getStorageInfo.localDiskAvailable() || location.getStorageInfo.HDFSAvailable()) {
+    } else if (location.getStorageInfo.localDiskAvailable() || location.getStorageInfo.HDFSAvailable()
+      || location.getStorageInfo.S3Available() || location.getStorageInfo.OSSAvailable()) {
       logDebug(s"create non-memory file for ${partitionDataWriterContext.getShuffleKey} ${partitionDataWriterContext.getPartitionLocation.getFileName}")
       val createDiskFileResult = createDiskFile(
         location,
@@ -945,7 +1029,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
     }
   }
 
-  def createMemoryFile(
+  def createMemoryFileInfo(
       appId: String,
       shuffleId: Int,
       fileName: String,
@@ -994,18 +1078,25 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         if (diskInfo != null && diskInfo.status.equals(DiskStatus.HEALTHY)) {
           diskInfo.dirs
         } else {
-          logDebug(s"Disk unavailable for $suggestedMountPoint, return all healthy" +
-            s" working dirs. diskInfo $diskInfo")
+          if (suggestedMountPoint.isEmpty) {
+            logDebug(s"Location suggestedMountPoint is not set, return all healthy working dirs.")
+          } else {
+            logInfo(s"Disk(${diskInfo.mountPoint}) unavailable for $suggestedMountPoint, return all healthy" +
+              s" working dirs.")
+          }
           healthyWorkingDirs()
         }
-      if (dirs.isEmpty && hdfsFlusher.isEmpty) {
+      if (dirs.isEmpty && hdfsFlusher.isEmpty && s3Flusher.isEmpty && ossFlusher.isEmpty) {
         throw new IOException(s"No available disks! suggested mountPoint $suggestedMountPoint")
       }
 
       if (dirs.isEmpty && location.getStorageInfo.HDFSAvailable()) {
         val shuffleDir =
           new Path(new Path(hdfsDir, conf.workerWorkingDir), s"$appId/$shuffleId")
-        FileSystem.mkdirs(StorageManager.hadoopFs, shuffleDir, hdfsPermission)
+        FileSystem.mkdirs(
+          StorageManager.hadoopFs.get(StorageInfo.Type.HDFS),
+          shuffleDir,
+          hdfsPermission)
         val hdfsFilePath = new Path(shuffleDir, fileName).toString
         val hdfsFileInfo = new DiskFileInfo(
           userIdentifier,
@@ -1017,6 +1108,42 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
           fileName,
           hdfsFileInfo)
         return (hdfsFlusher.get, hdfsFileInfo, null)
+      } else if (dirs.isEmpty && location.getStorageInfo.S3Available()) {
+        val shuffleDir =
+          new Path(new Path(s3Dir, conf.workerWorkingDir), s"$appId/$shuffleId")
+        FileSystem.mkdirs(
+          StorageManager.hadoopFs.get(StorageInfo.Type.S3),
+          shuffleDir,
+          hdfsPermission)
+        val s3FilePath = new Path(shuffleDir, fileName).toString
+        val s3FileInfo = new DiskFileInfo(
+          userIdentifier,
+          partitionSplitEnabled,
+          new ReduceFileMeta(conf.shuffleChunkSize),
+          s3FilePath,
+          StorageInfo.Type.S3)
+        diskFileInfos.computeIfAbsent(shuffleKey, diskFileInfoMapFunc).put(
+          fileName,
+          s3FileInfo)
+        return (s3Flusher.get, s3FileInfo, null)
+      } else if (hasOssStorage && location.getStorageInfo.OSSAvailable()) {
+        val shuffleDir =
+          new Path(new Path(ossDir, conf.workerWorkingDir), s"$appId/$shuffleId")
+        FileSystem.mkdirs(
+          StorageManager.hadoopFs.get(StorageInfo.Type.OSS),
+          shuffleDir,
+          hdfsPermission)
+        val ossFilePath = new Path(shuffleDir, fileName).toString
+        val ossFileInfo = new DiskFileInfo(
+          userIdentifier,
+          partitionSplitEnabled,
+          new ReduceFileMeta(conf.shuffleChunkSize),
+          ossFilePath,
+          StorageInfo.Type.OSS)
+        diskFileInfos.computeIfAbsent(shuffleKey, diskFileInfoMapFunc).put(
+          fileName,
+          ossFileInfo)
+        return (ossFlusher.get, ossFileInfo, null)
       } else if (dirs.nonEmpty && location.getStorageInfo.localDiskAvailable()) {
         val dir = dirs(getNextIndex % dirs.size)
         val mountPoint = DeviceInfo.getMountPoint(dir.getAbsolutePath, mountPoints)
@@ -1089,5 +1216,5 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
 }
 
 object StorageManager {
-  var hadoopFs: FileSystem = _
+  var hadoopFs: util.Map[StorageInfo.Type, FileSystem] = _
 }
